
#include "CreateAcrossBoundaries.h"

using namespace std;
using namespace tlp;


//================================================================
Graph *makeMergeGraph (BooleanProperty &splitSelection,
		       Graph *displayGraph) {
  string errMsg;
  DataSet splitData;
  splitData.set ("Nodes", &splitSelection);
  BooleanProperty nodeAndEdgeSelection (displayGraph);
  bool success = 
    displayGraph->computeProperty ("Induced sub-graph",
				   &nodeAndEdgeSelection, errMsg,
				   NULL, &splitData);
  if (success)
    return displayGraph->getRoot ()->addSubGraph (&nodeAndEdgeSelection);
  else {
    cerr << "Induced sub-graph returned the following error:\n"
	 << errMsg << "\n(Is the plugin properly installed?)" << endl;
    return NULL;
  }//end else
}//end makeSplitGraph

//==================================================================
CreateAcrossBoundaries::
CreateAcrossBoundaries (GrouseInterfaceData *grouseMainWidget,
			Graph *displayGraph,
			Graph *hierarchy,
			Graph *hierarchySubGraphs,
			OpenCloseHierarchy *openCloseHierarchy,
			const string &metaGraphPropertyName) :
displayGraph (displayGraph),
hierarchy (hierarchy),
hierarchySubGraphs (hierarchySubGraphs),
openCloseHierarchy (openCloseHierarchy) {
  split = new Split (displayGraph, hierarchy, hierarchySubGraphs,
		     openCloseHierarchy, metaGraphPropertyName);
  createWithSelection = new CreateWithSelection (grouseMainWidget,
						 hierarchy,
						 hierarchySubGraphs,
						 metaGraphPropertyName,
						 openCloseHierarchy);
  hierarchyMetaGraphs =
    hierarchy->getProperty<GraphProperty> (metaGraphPropertyName);
}//end constructor

//==================================================================
CreateAcrossBoundaries::
~CreateAcrossBoundaries () {
  delete split;
  delete createWithSelection;
}//end ~CreateAcrossBoundaries

//================================================================
node CreateAcrossBoundaries::componentLCA (Graph *curComp) {
  //get the paths through the hierarchy
  vector<node> paths [curComp->numberOfNodes()];
  unsigned int counter = 0;
  node curDisplayNode;
  forEach (curDisplayNode, curComp->getNodes ()) {
    node curHierarchyNode = 
      openCloseHierarchy->getHierarchyNode (curDisplayNode);
    openCloseHierarchy->getPath (curHierarchyNode, paths[counter]);
    ++counter;
  }//end if

  //find the LCA
  vector<node> lcaPath = paths[0];
  for (unsigned int i = 1; i < curComp->numberOfNodes (); ++i) {
    unsigned int pathCounter = 0;
    while (lcaPath[pathCounter] == paths[i][pathCounter]) 
      ++pathCounter;
    if (pathCounter < lcaPath.size())
      lcaPath.resize (pathCounter);
  }//end for
  return lcaPath[lcaPath.size() - 1];
}//end componentLCA

//================================================================
bool CreateAcrossBoundaries::findLCAs (Graph *mergeGraph,
				       vector<node> &lcas,
				       vector<Graph *> 
				       &newDisplaySubGraphs) {
  //decompose subgraphs into connected components
  DoubleProperty connectedComponents (mergeGraph);
  string errMsg;
  bool success = mergeGraph->computeProperty ("Connected Component", 
					      &connectedComponents, errMsg);
  if (!success) {
    cerr << "Connected components returned the following error " << endl;
    cerr << errMsg << endl << " is the plugin installed." << endl;
    return false;
  }//end if
  
  DataSet localData;
  localData.set("Property", &connectedComponents);
  success = tlp::applyAlgorithm (mergeGraph, errMsg, &localData, 
				 "Equal Value");
  if (!success) {
    cerr << "Equal Value clustering returned the following error:\n"
	 << errMsg << "\n(Is the plugin properly installed?)" << endl;
    return false;
  }//end if

  StableIterator<Graph *> concomps (mergeGraph->getSubGraphs());
  while (concomps.hasNext()) {
    Graph *curComp = concomps.next ();
    node lca = componentLCA (curComp);
    lcas.push_back (lca);
    newDisplaySubGraphs.push_back (curComp);
  }//end while
  return true;
}//end findLCAs

//==================================================================
bool CreateAcrossBoundaries::splitAndMerge (vector<node> &lcaHierarchyNodes,
					    vector<Graph *> 
					    &newDisplaySubGraphs,
					    vector<node> &newHierarchyNodes) {
  BooleanProperty mergeSelection (hierarchy);
  for (unsigned int i = 0; i < lcaHierarchyNodes.size (); ++i) {
    Graph *curStopGraph = 
      hierarchyMetaGraphs->getNodeValue (lcaHierarchyNodes[i]);
    node curNewDisplayNode;
    forEach (curNewDisplayNode, newDisplaySubGraphs[i]->getNodes ()) {
      node curNewHierarchyNode =
	openCloseHierarchy->getHierarchyNode (curNewDisplayNode);
      split->splitToGraph (curNewHierarchyNode, curStopGraph);
      mergeSelection.setNodeValue (curNewHierarchyNode, true);
    }//end forEach
  }//end for
  createWithSelection->mergeSelections (mergeSelection,
					newHierarchyNodes);
}//end splitAndMerge

//==================================================================
bool CreateAcrossBoundaries::
mergeAcrossBoundaries (BooleanProperty &splitSelection,
		       vector<node> &newHierarchyNodes) {
  //backup entire hierarchy and destroy hierarchy below affected nodes
  vector<node> lcaHierarchyNodes;
  vector<Graph *> newDisplaySubGraphs;
  BooleanProperty displaySelection (displayGraph);
  node hierarchyNodeInDisplay;
  forEach (hierarchyNodeInDisplay, splitSelection.getNodesEqualTo (true)) {
    node displayNode = 
      openCloseHierarchy->getDisplayNode (hierarchyNodeInDisplay);
    displaySelection.setNodeValue (displayNode, true);
  }//end forEach

  Graph *mergeGraph = makeMergeGraph (displaySelection, displayGraph);
  findLCAs (mergeGraph, lcaHierarchyNodes, newDisplaySubGraphs);
  splitAndMerge (lcaHierarchyNodes, newDisplaySubGraphs, newHierarchyNodes);
  displayGraph->getRoot ()->delAllSubGraphs (mergeGraph); 
  return true;
}//end performSplit
