#include <cstdio>
#include <qmessagebox.h>
#include <tulip/StableIterator.h>
#include <tulip/StringCollection.h>
#include <tulip/DoubleProperty.h>
#include "ColourShapeDefaults.h"
#include "HierarchyNodeCreation.h"
#include "CreateHierarchy.h"

using namespace std;
using namespace tlp;

//pattern match or category colours
#define ATTRIBUTE_MATCH (string ("attribute match "))
#define ATTRIBUTE_MATCH_0 (ATTRIBUTE_MATCH + "0")
#define ATTRIBUTE_MATCH_1 (ATTRIBUTE_MATCH + "1")
#define ATTRIBUTE_MATCH_2 (ATTRIBUTE_MATCH + "2")

//================================================================
CreateHierarchy::CreateHierarchy (GrouseInterfaceData *grouseMainWidget,
				  Graph *hierarchy, 
				  Graph *hierarchySubGraphs,
				  const std::string &metaGraphPropertyName,
				  OpenCloseHierarchy *openCloseHierarchy,
				  RotatingColours *rotatingColours,
				  PluginProgress *pluginProgress) :
grouseMainWidget(grouseMainWidget), 
hierarchy(hierarchy), 
hierarchySubGraphs(hierarchySubGraphs), 
openCloseHierarchy(openCloseHierarchy),
rotatingColours(rotatingColours),
pluginProgress (pluginProgress), 
colourIteration(0),
onRotatingColours(false)
{
  StringCollection featureTypes (FEATURE_LIST);
  StringCollection algoTypes (ALGORITHM_LIST);
  for (unsigned int i = 0; i < featureTypes.size(); ++i)
    featureTypeToMetricAlgo[featureTypes[i]] = algoTypes[i];
  hierarchyMetaGraphs = 
    hierarchy->getProperty<GraphProperty> (metaGraphPropertyName);
  matchColour[ATTRIBUTE_MATCH_0] = GROUSE_PINK;
  notMatchColour[ATTRIBUTE_MATCH_0] = GROUSE_GOLD;
  matchColour[ATTRIBUTE_MATCH_1] = GROUSE_GREEN;
  notMatchColour[ATTRIBUTE_MATCH_1] = GROUSE_TEAL;
  matchColour[ATTRIBUTE_MATCH_2] = GROUSE_ORANGE;
  notMatchColour[ATTRIBUTE_MATCH_2] = GROUSE_BLUE;
  matchColour[CATEGORY] = GROUSE_MAGENTA;
  notMatchColour[CATEGORY] = GROUSE_MAGENTA;
  matchColour[TREE] = GROUSE_BLUE;
  notMatchColour[TREE] = GROUSE_GOLD;
  matchColour[BICON_COMPONENTS] = GROUSE_BROWN;
  notMatchColour[BICON_COMPONENTS] = GROUSE_BROWN;
  matchColour[CLUSTER] = GROUSE_PURPLE;
  notMatchColour[CLUSTER] = GROUSE_PURPLE;
}//end constructor

//================================================================
inline string getStringAlgoNumber (unsigned int number) {
  char charAlgoNumber[10];
  sprintf (charAlgoNumber, "%d", number);
  number++;
  return charAlgoNumber;
} 

//================================================================
void CreateHierarchy::setColours (Graph *curHierarchy,
				  string featureType,
				  BooleanProperty &match) {
  ColorProperty *viewColor = 
    curHierarchy->getProperty<ColorProperty> ("viewColor");
  StringProperty *catMatch = 
    curHierarchy->getProperty<StringProperty> ("view Bracket Match Property");
  string matchType;

  if ((featureType == CATEGORY) || (featureType == PATTERN_MATCH)) {
    matchType =  ATTRIBUTE_MATCH + getStringAlgoNumber (colourIteration % 3);
  }

  node curNode;
  forEach (curNode, curHierarchy->getNodes()) {
    if (match.getNodeValue (curNode)) {
      if (!onRotatingColours)
	viewColor->setNodeValue (curNode, matchColour[matchType]);
      else if ((rotatingColours->getGreyScale ()) &&
	       (featureType == CATEGORY)) {
	string bracketValue = catMatch->getNodeValue (curNode);
	viewColor->setNodeValue (curNode, 
				 rotatingColours
				 ->getCategoryColour (bracketValue));
      }//end else if
      else 
	viewColor->setNodeValue (curNode, rotatingColours
				 ->getCurMatchColour ());
    }
    else { 
      if (!onRotatingColours)
	viewColor->setNodeValue (curNode, notMatchColour[matchType]);
      else if ((rotatingColours->getGreyScale ()) &&
	       (featureType == CATEGORY)) {
	string bracketValue = catMatch->getNodeValue (curNode);
	viewColor->setNodeValue (curNode, 
				 rotatingColours
				 ->getCategoryColour (bracketValue));
      }//end else if
      else 
	viewColor->setNodeValue (curNode, rotatingColours
				 ->getCurUnmatchColour ());
    }
  }//end forEach
  viewColor->setAllEdgeValue (GROUSE_LIGHT_GREY);
}//end setColours

//================================================================
bool CreateHierarchy::applyAlgorithm (Graph *curHierarchy, 
				      const string &featureType,
				      DataSet *localData) {
  const string algorithm = featureTypeToMetricAlgo[featureType];
  string algoNumber = getStringAlgoNumber (clusterCounter[algorithm]);
  curHierarchy->setAttribute <string> ("name", 
				       featureType + " base " + 
				       algoNumber);

  //clone graph for clustering
  Graph *cloneHierarchy = curHierarchy->getRoot()->addSubGraph();
  copy (curHierarchy, cloneHierarchy);
  
  //try the passed algorithm on the working localGraph
  bool success;
  string errMsg;
  localData->set ("Non-Tree Roots", true);
  BooleanProperty matches (curHierarchy);
  matches.setAllNodeValue (false);
  localData->set ("matches", &matches);
  DoubleProperty metricClusters (cloneHierarchy);
  success = 
    cloneHierarchy->computeProperty (algorithm, 
				     &metricClusters, errMsg,
				     NULL, localData);

  //make sure the plugin ran.
  if (!success) {
    cerr << "The metric plugin " << algorithm 
	 << " encountered the following error:\n" << errMsg 
	 << "\n(Is the plugin installed?)" << endl;
    return false;
  }//end if

  //cluster them
  DataSet equalValueData;
  equalValueData.set("Property", &metricClusters);
  success = tlp::applyAlgorithm (cloneHierarchy, errMsg, &equalValueData, 
				 "Equal Value");
  if (!success) {
    cerr << "Equal Value clustering returned the following error:\n"
	 << errMsg << "\n(Is the plugin properly installed?)" << endl;
    return false;
  }//end if
  
  //update curHierarchy so it is the same as cloneHierarchy
  StableIterator <node> workingNodes (curHierarchy->getNodes());
  while (workingNodes.hasNext())
    curHierarchy->delNode (workingNodes.next());
  StableIterator <edge> workingEdges (curHierarchy->getEdges());
  while (workingEdges.hasNext())
    curHierarchy->delEdge (workingEdges.next());
  copy (cloneHierarchy, curHierarchy);
  
  Iterator<Graph *> *subgraphs = cloneHierarchy->getSubGraphs();
  unsigned int numSubGraphs = 0;
  unsigned int numSubAndSingle = 0;
  while (subgraphs->hasNext()) {
    ++numSubAndSingle;
    Graph *curSub = subgraphs->next();
    if (curSub->numberOfNodes () > 1)
      ++numSubGraphs;
  } delete subgraphs;

  //restructure metagraph hierarchy tree
  StableIterator <Graph *> clusters (cloneHierarchy->getSubGraphs());
  while (clusters.hasNext()) {
    Graph *curCluster = clusters.next();
   
    if ((curCluster->numberOfNodes() > 1) && (numSubAndSingle > 1)) {
      Graph *newHierarchyGraph = hierarchySubGraphs->addSubGraph();

      string label;
      string bracketMatchValue;
      StringProperty *catMatch;
      if (featureType == CATEGORY) {
        catMatch = 
	  curHierarchy->getRoot()
	  ->getLocalProperty<StringProperty> ("view Bracket Match Property");
	bracketMatchValue = 
	  catMatch->getNodeValue (curCluster->getOneNode());
	label = featureType + " " + bracketMatchValue + " ";
      }//end if

      else if (featureType == PATTERN_MATCH) {
	bool doesMatch = matches.getNodeValue (curCluster->getOneNode());
	if (doesMatch) label = "In Pattern Match ";
	else label = "Out of Pattern Match ";
	string searchString;
	localData->get<string> ("searchString", searchString);
	label += searchString + " ";
      }
      else label = featureType;
      
      algoNumber = getStringAlgoNumber (clusterCounter[algorithm]);
      ++clusterCounter[algorithm];
      label = algoNumber + " " + label;
      newHierarchyGraph->setAttribute<string> ("name", label);
      copy (curCluster, newHierarchyGraph);

      //set colours here
      node newClusterNode = curHierarchy->addNode();
      hierarchyMetaGraphs->setNodeValue (newClusterNode, newHierarchyGraph);
      matches.setNodeValue (newClusterNode, 
			    matches.getNodeValue (newHierarchyGraph->getOneNode()));
      if (featureType == CATEGORY)
	catMatch->setNodeValue (newClusterNode,
				catMatch
				->getNodeValue (curCluster->getOneNode()));
      setNewConnections (curHierarchy, newHierarchyGraph, newClusterNode);

      delFeature (newHierarchyGraph, curHierarchy);
      nodesClustered += newHierarchyGraph->numberOfNodes();
      pluginProgress->progress (nodesClustered, nodesToCluster);
    }//end if
    else {
      pluginProgress->progress (nodesClustered++, nodesToCluster);
    }
  }//end for clusters
  setColours (curHierarchy, featureType, matches);
  curHierarchy->getRoot()->delAllSubGraphs (cloneHierarchy);
  if (onRotatingColours) rotatingColours->incrementColour ();
  return true;
}//end applyAlgorithm

//================================================================
unsigned int CreateHierarchy::numNodesBelowSelection (BooleanProperty &selection) {
  unsigned int numNodes = 0;
  Iterator<node> *newNodeSelection = selection.getNodesEqualTo (true);
  while (newNodeSelection->hasNext()) {
    node curNode = newNodeSelection->next();
    Graph *subGraph = hierarchyMetaGraphs->getNodeValue (curNode);
    if (subGraph != NULL) 
      numNodes += subGraph->numberOfNodes ();
  } delete newNodeSelection;
  return numNodes;
}

//================================================================
bool CreateHierarchy::newHierarchyLevel (tlp::BooleanProperty &selection,
					 const std::string &featureType,
					 DataSet *localData) {
  bool success = true;
  nodesClustered = 0;
  nodesToCluster = numNodesBelowSelection (selection);
  Iterator<node> *newNodeSelection = selection.getNodesEqualTo (true);
  while (newNodeSelection->hasNext()) {
    node curHierarchyNode = newNodeSelection->next();
    Graph *hierarchyMetaGraph = 
      hierarchyMetaGraphs->getNodeValue (curHierarchyNode);
    if (hierarchyMetaGraph == NULL) continue;
    if (localData != NULL)
      success = applyAlgorithm (hierarchyMetaGraph, featureType,
				localData);
    else {
      localData = new DataSet();
      success = applyAlgorithm (hierarchyMetaGraph, featureType,
				localData);
      delete localData;
    }
    if (!success) return false;
  } delete newNodeSelection; //end while
  ++colourIteration;
  return success;
}//end newHierarchyLabel
