/**
 * The following class provides functions to create and delete 
 * nodes of a meta graph hierarchy.  
 */
#ifndef __CREATE_HIERARCHY_H__
#define __CREATE_HIERARCHY_H__

#if (__GNUC__ < 3)
#  include <hash_map>
#else
#  include <ext/hash_map>
#endif
#include <string>
#include <tulip/BooleanProperty.h>
#include <tulip/GraphProperty.h>
#include <tulip/PluginProgress.h>
#include <tulip/QtProgressWidget.h>
#include "GrouseInterfaceData.h"
#include "OpenCloseHierarchy.h"
#include "GrouseFeatures.h"
#include "RotatingColours.h"

class CreateHierarchy {
 public:
  /**
   * Constructor which requires a graph hierarchy and a openCloseHierarchy
   * object.
   */
  CreateHierarchy (GrouseInterfaceData *grouseMainWidget,
		   tlp::Graph *hierarchy, tlp::Graph *hierarchySubGraphs,
		   const std::string &metaGraphPropertyName,
		   OpenCloseHierarchy *openCloseHierarchy,
		   RotatingColours *rotatingColours,
		   PluginProgress *pluginProgress);

  /**
   * Destructor which does nothing for the time being.
   */
  inline ~CreateHierarchy () {}

 public:
   /**
    * This function creates a new level of the graph hierarchy.  The
    * new level is created below each of the selected nodes, if they
    * are all leaf nodes.  Returns true if features are found.
    * \param selection The selection below which the new level is created.
    * \param algorithm the algorithm used for hierarchy creation.
    */
   bool newHierarchyLevel (tlp::BooleanProperty &selection,
			   const std::string &featureType,
			   tlp::DataSet *localData);

   /**
    * An inline function to set whether or not rotating colours is on.
    * \param turnOn If true, rotating colours is on.  Otherwise, it is off.
    */
   inline void turnOnRotatingColours (bool turnOn) {
     onRotatingColours = turnOn;
   }//end turnOnRotatingColours
   

 private:
  GrouseInterfaceData *grouseMainWidget;
  tlp::Graph *hierarchy;
  tlp::Graph *hierarchySubGraphs;
  tlp::GraphProperty *hierarchyMetaGraphs;
  OpenCloseHierarchy *openCloseHierarchy;
  RotatingColours *rotatingColours;
  stdext::hash_map<std::string, unsigned int> clusterCounter;
  stdext::hash_map<std::string, std::string> featureTypeToMetricAlgo;
  stdext::hash_map<std::string, tlp::Color> notMatchColour;
  stdext::hash_map<std::string, tlp::Color> matchColour;
  PluginProgress *pluginProgress;
  unsigned int nodesToCluster;
  unsigned int nodesClustered;
  unsigned int colourIteration;
  bool onRotatingColours;

  /**
    * This function creates a new hierarchy level by applying the passed
    * algorithm.  Returns true if features are found.
    * \param curHierarchy Graph to which the algorithm is applied.
    * \param featureType The feature sought.
    */
  bool applyAlgorithm (tlp::Graph *curHierarchy, 
		       const std::string &featureType,
		       tlp::DataSet *localData);

  /**
   * This function sets the colours of the nodes in the graph hierarchy
   * based on if they match or do not match the decomposition feature.
   * \param curHierarchy Graph to which the algorithm is applied.
   * \param featureType The type of feature being sought.
   * \param match Values to indicate if this matched the requested feature
   * type.  True matches and false does not.
   */
  void setColours (tlp::Graph *curHierarchy, std::string featureType,
		   tlp::BooleanProperty &match);

  /**
   * Returns the number of nodes one level below the hierarchy node selection.
   * \param selection The selection of hierarchy nodes
   */
  unsigned int numNodesBelowSelection (BooleanProperty &selection);
};
#endif
