#include <cstdio>
#include <qmessagebox.h>
#include <tulip/StableIterator.h>
#include <tulip/DrawingTools.h>
#include <tulip/DoubleProperty.h>
#include "ColourShapeDefaults.h"
#include "HierarchyNodeCreation.h"
#include "CreateWithSelection.h"

using namespace std;
using namespace tlp;

//================================================================
CreateWithSelection::CreateWithSelection (GrouseInterfaceData *grouseMainWidget,
				  Graph *hierarchy, 
				  Graph *hierarchySubGraphs,
				  const std::string &metaGraphPropertyName,
				  OpenCloseHierarchy *openCloseHierarchy) :
grouseMainWidget(grouseMainWidget), hierarchy(hierarchy), 
hierarchySubGraphs(hierarchySubGraphs), openCloseHierarchy(openCloseHierarchy),
selectionNumber(0) 
{
  hierarchyMetaGraphs = 
    hierarchy->getProperty<GraphProperty> (metaGraphPropertyName);
}//end constructor

//================================================================
bool CreateWithSelection::getInducedSubGraphs (Graph *containedIn,
					       BooleanProperty &selection,
					       BooleanProperty &result) {
  result.setAllNodeValue (false);
  result.setAllEdgeValue (false);
  DataSet inducedData;
  inducedData.set ("Nodes", &selection);
  string errMsg;
  bool success = containedIn->computeProperty ("Induced sub-graph",
					       &result, errMsg,
					       NULL, &inducedData);
  if (!success) {
    QString msg = "Induced sub-graph returned the following error\n" + errMsg 
      + "\n(Is the plugin installed?)";
    QMessageBox::critical (grouseMainWidget, "Plugin not installed", msg,
			   QMessageBox::Ok, QMessageBox::NoButton,
			   QMessageBox::NoButton);
    return false;
  }//end if
  return true;
}//end getInducedSubGraphs

//================================================================
inline string getStringAlgoNumber (unsigned int &number) {
  char charAlgoNumber[10];
  sprintf (charAlgoNumber, "%d", number);
  number++;
  return charAlgoNumber;
} 

//================================================================
inline Coord getSubgraphAveCoord (Graph *subGraph, Graph *referenceGraph) {
  LayoutProperty *refLayout =
    referenceGraph->getProperty<LayoutProperty> ("viewLayout");
  Coord result (0,0,0);
  node curNode;
  forEach (curNode, subGraph->getNodes ())
    result += refLayout->getNodeValue (curNode);
  return result / static_cast<double> (subGraph->numberOfNodes ());
}//end getSubgraphAveCoord

//================================================================
void copyLayoutAndCentre (LayoutProperty *startLayout, 
			  Graph *newHierarchyGraph,
			  const Coord &aveCoord) {
  LayoutProperty *newHierarchyLayout = 
    newHierarchyGraph->getLocalProperty<LayoutProperty> ("viewLayout");
  node curNode;
  forEach (curNode, newHierarchyGraph->getNodes ())
    newHierarchyLayout->setNodeValue (curNode,
				      startLayout->getNodeValue (curNode) -
				      aveCoord);
  edge curEdge;
  forEach (curEdge, newHierarchyGraph->getEdges ()) {
    vector<Coord> bends = startLayout->getEdgeValue (curEdge);
    for (unsigned int i = 0; i < bends.size (); ++i)
      bends[i] -= aveCoord;
    newHierarchyLayout->setEdgeValue (curEdge, bends);
  }//end forEach
}
//================================================================
inline void setNodeDone (Graph *doneGraph, MutableContainer<bool> &nodeDone) {
  node curNode;
  forEach (curNode, doneGraph->getNodes ()) nodeDone.set (curNode.id, true);
}//end setNodeDone

//================================================================
bool CreateWithSelection::mergeConnectedComponents (Graph *containedIn,
						    BooleanProperty &induced,
						    vector<node> &mergeNodes,
						    MutableContainer<bool>  
						    &nodeDone) {
  Graph *connectedSub = containedIn->getRoot()->addSubGraph (&induced);
  string errMsg;
  DataSet localData;
  DoubleProperty connectedMetric (connectedSub);
  bool success = connectedSub->computeProperty ("Connected Component",
						&connectedMetric,
						errMsg);
  if (!success) {
    cerr << "Connected components returned the following error " << endl 
	 << errMsg << endl << "(Is the plugin installed?)" << endl;
    containedIn->getRoot()->delAllSubGraphs (connectedSub);
    return false;
  }//end if

  localData.set("Property", &connectedMetric);
  success = tlp::applyAlgorithm (connectedSub, errMsg, &localData, 
				 "Equal Value");
  if (!success) {
    cerr << "Equal Value clustering returned the following error:\n"
	 << errMsg << "\n(Is the plugin properly installed?)" << endl;
    containedIn->getRoot()->delAllSubGraphs (connectedSub);
    return false;
  }//end if
  
  LayoutProperty *containedInLayout =
    containedIn->getProperty<LayoutProperty> ("viewLayout");
  ColorProperty *mergeColour =
    containedIn->getProperty<ColorProperty> ("viewColor");
  StableIterator <Graph *> components (connectedSub->getSubGraphs());
  while (components.hasNext()) {
    Graph *curComponent = components.next();
    if ((curComponent->numberOfNodes () > 1) && 
	(curComponent->numberOfNodes () < containedIn->numberOfNodes ())) {
      Graph *newHierarchyGraph = hierarchySubGraphs->addSubGraph();
      string label = "Merge Selection " + getStringAlgoNumber (selectionNumber);
      ++selectionNumber;
      newHierarchyGraph->setAttribute <string> ("name", label);
      setNodeDone (curComponent, nodeDone);
      copy (curComponent, newHierarchyGraph);
      Coord aveCoord = getSubgraphAveCoord (newHierarchyGraph, containedIn);
      copyLayoutAndCentre (containedInLayout, newHierarchyGraph, aveCoord);
      node newComponentNode = containedIn->addNode();
      containedInLayout->setNodeValue (newComponentNode, aveCoord);
      mergeNodes.push_back (newComponentNode);
      hierarchyMetaGraphs->setNodeValue (newComponentNode, newHierarchyGraph);
      mergeColour->setNodeValue (newComponentNode, GROUSE_GREEN);
      setNewConnections (containedIn, newHierarchyGraph, newComponentNode);
      delFeature (newHierarchyGraph, containedIn);
      delFeature (newHierarchyGraph, connectedSub);
    }//end if
  }//end while
  containedIn->getRoot()->delAllSubGraphs (connectedSub);
  return true;
}

//================================================================
bool CreateWithSelection::mergeInducedSubGraphs (Graph *containedIn,
						 BooleanProperty &selection,
						 vector<node> &mergeNodes,
						 MutableContainer<bool> 
						 &nodeDone) {
  BooleanProperty inducedSubGraphs (containedIn);
  getInducedSubGraphs (containedIn, selection, inducedSubGraphs);
  mergeConnectedComponents (containedIn, inducedSubGraphs, mergeNodes,
			    nodeDone);
  return true;
}//end mergeInducedSubGraphs

//================================================================
bool CreateWithSelection::mergeSelections (BooleanProperty &selection,
					   vector<node> &newHierarchyNodes) {
  bool success = true;
  MutableContainer<bool> nodeDone;
  nodeDone.setAll (false);
  Iterator<node> *newNodeSelection = selection.getNodesEqualTo (true);
  while (newNodeSelection->hasNext()) {
    node curHierarchyNode = newNodeSelection->next();
    node containingHierarchyNode = 
      openCloseHierarchy->getContainingHierarchyNode (curHierarchyNode);
    Graph *containedIn = 
      openCloseHierarchy->getContainingHierarchyGraph (curHierarchyNode);
    if (!nodeDone.get (curHierarchyNode.id)) {
      vector<node> mergeNodes;
      mergeInducedSubGraphs (containedIn, selection, mergeNodes, nodeDone);
      for (unsigned int i = 0; i < mergeNodes.size(); ++i) {
	openCloseHierarchy->addOneLevel (mergeNodes[i], 
					 containingHierarchyNode);
	newHierarchyNodes.push_back (mergeNodes[i]);
      }//end for
    }//end if
  } delete newNodeSelection; //end while
  return success;
}//end mergeSelections
