/**
 * The following class provides functions to create and delete 
 * nodes of a meta graph hierarchy.  
 */
#ifndef __CREATE_WITH_SELECTION_H__
#define __CREATE_WITH_SELECTION_H__

#if (__GNUC__ < 3)
#  include <hash_map>
#else
#  include <ext/hash_map>
#endif
#include <string>
#include <tulip/BooleanProperty.h>
#include <tulip/GraphProperty.h>
#include "GrouseInterfaceData.h"
#include "OpenCloseHierarchy.h"
#include "GrouseFeatures.h"

class CreateWithSelection {
 public:
  /**
   * Constructor which requires a graph hierarchy and a openCloseHierarchy
   * object.
   */
  CreateWithSelection (GrouseInterfaceData *grouseMainWidget,
		       tlp::Graph *hierarchy, tlp::Graph *hierarchySubGraphs,
		       const std::string &metaGraphPropertyName,
		       OpenCloseHierarchy *openCloseHierarchy);

  /**
   * Destructor which does nothing for the time being.
   */
  inline ~CreateWithSelection () {}

 public:
   /**
    * Merge selected connected components at the same level.
    * \param selection The visible selection.
    * \param newHierarchyNodes The set of nodes generated by the selection
    * creation operation.
    */
   bool mergeSelections (tlp::BooleanProperty &selection,
			 std::vector<tlp::node> &newHierarchyNodes);

 private:
  GrouseInterfaceData *grouseMainWidget;
  tlp::Graph *hierarchy;
  tlp::Graph *hierarchySubGraphs;
  tlp::GraphProperty *hierarchyMetaGraphs;
  OpenCloseHierarchy *openCloseHierarchy;
  unsigned int selectionNumber;

  
 /**
   * Function to get induced subgraphs from the nodes selected in selection.
   * \param containedIn Hierarchy subgraph containing selected node.
   * \param selection The hierarchy node selection
   * \param result The selected subgraphs produced by the induced subgraph.
   */
  bool getInducedSubGraphs (tlp::Graph *containedIn,
			    tlp::BooleanProperty &selection,
			    tlp::BooleanProperty &result);

  /**
   * Merges each induced subgraph down to a single node.
   * \param containedIn The containing hierarchy subgraph.
   * \param induced The selected components.
   * \param mergNodes The nodes added for the merged levels.
   * \param nodeDone Keeps track of nodes that have been merged once
   */
  bool mergeConnectedComponents (Graph *containedIn,
				 BooleanProperty &induced,
				 std::vector<tlp::node> &mergeNodes,
				 tlp::MutableContainer<bool> &nodeDone);

  /**
   * Function to do entire merging procedure (getInducedSubGraphs,
   * followed by mergeConnectedComponents).
   * \param containedIn The containing hierarchy subgraph.
   * \param selection The selected components.
   * \param mergeNodes The nodes added for the merged levels
   * \param nodeDone Keeps track of nodes that have been merged once
   */
  bool mergeInducedSubGraphs (Graph *containedIn,
			      BooleanProperty &selection,
			      std::vector<tlp::node> &mergeNodes,
			      tlp::MutableContainer<bool> &nodeDone);
};
#endif
