#include <qmessagebox.h>
#include <tulip/StableIterator.h>
#include "SteeredLayout.h"
#include "DestroyHierarchy.h"

using namespace std;
using namespace tlp;

//================================================================
DestroyHierarchy::DestroyHierarchy (GrouseInterfaceData *grouseMainWidget,
				    tlp::Graph *hierarchy, 
				    tlp::Graph *hierarchySubGraphs,
				    tlp::Graph *inputGraph,
				    const std::string &metaGraphPropertyName,
				    OpenCloseHierarchy *openCloseHierarchy,
				    MetaNodeBoundingCircles *boundingCircles,
				    PluginProgress *pluginProgress) :
grouseMainWidget (grouseMainWidget), 
hierarchy (hierarchy), 
hierarchySubGraphs (hierarchySubGraphs), 
inputGraph (inputGraph),
openCloseHierarchy (openCloseHierarchy),
boundingCircles (boundingCircles),
pluginProgress (pluginProgress) {
  hierarchyMetaGraphs = 
    hierarchy->getProperty<GraphProperty> (metaGraphPropertyName);
}

//========================================================================
//a function that copies the nodes and edges from src to dest if they
//are selected. 
inline void
copySelection (BooleanProperty &selection, Graph *dest) {
  StableIterator <node> nodes (selection.getNodesEqualTo (true));
  while (nodes.hasNext()) dest->addNode (nodes.next());
  StableIterator <edge> edges (selection.getEdgesEqualTo (true));
  while (edges.hasNext()) dest->addEdge (edges.next());
}//end copySelection

//========================================================================
void DestroyHierarchy::deleteHierarchy (Graph *deleteGraph) {
  StableIterator <node> nodes (deleteGraph->getNodes());
  while (nodes.hasNext ()) {
    node curDelNode = nodes.next();
    Graph *curSubGraph = hierarchyMetaGraphs->getNodeValue (curDelNode);
    if (curSubGraph != NULL) {
      boundingCircles->delBoundingCircle (curDelNode);
      deleteHierarchy (curSubGraph);
      hierarchyMetaGraphs->setNodeValue (curDelNode, NULL);
      clearLayoutParameters (curSubGraph);
      hierarchySubGraphs->delSubGraph (curSubGraph);
      pluginProgress->progress (nodesDestroyed++, nodesToDestroy);
    }//end if
    deleteGraph->delNode (curDelNode);
    pluginProgress->progress (nodesDestroyed++, nodesToDestroy);
  }//end while
}//end delGraph

//================================================================
void DestroyHierarchy::extractLeafSubGraph (Graph *hierarchyMetaGraph,
					    BooleanProperty &leaves) {
  node curNode;
  forEach (curNode, hierarchyMetaGraph->getNodes()) {
    Graph *curNodeMetaGraph = hierarchyMetaGraphs->getNodeValue (curNode);
    if (curNodeMetaGraph == NULL) {
      if (inputGraph->isElement (curNode))
	leaves.setNodeValue (curNode, true);
    }//end if
    else 
      extractLeafSubGraph (curNodeMetaGraph, leaves);
  }//end forEach
}//end extractLeafSubGraph

//================================================================
bool DestroyHierarchy::deleteLevels (Graph *hierarchyMetaGraph) {
  BooleanProperty leafSelection (inputGraph);
  leafSelection.setAllNodeValue (false);
  extractLeafSubGraph (hierarchyMetaGraph, leafSelection);
  BooleanProperty inducedSubGraph (inputGraph);
  DataSet leafData;
  leafData.set ("Nodes", &leafSelection);
  string errMsg;
  bool success = inputGraph->computeProperty ("Induced sub-graph",
					      &inducedSubGraph, errMsg,
					      NULL, &leafData);
  if (!success) {
    QString msg = "Induced sub-graph returned the following error\n" + errMsg 
      + "\n(Is the plugin installed?)";
    QMessageBox::critical (grouseMainWidget, "Plugin not installed", msg,
			   QMessageBox::Ok, QMessageBox::NoButton,
			   QMessageBox::NoButton);
    return false;
  }//end if
  deleteHierarchy (hierarchyMetaGraph);
  pluginProgress->progress (nodesDestroyed++, nodesToDestroy);
  clearLayoutParameters (hierarchyMetaGraph);
  copySelection (inducedSubGraph, hierarchyMetaGraph);
  return true;
}//end deleteLevels

//================================================================
unsigned int DestroyHierarchy::nodesInHierarchy (const node &curHierarchyNode,
						 MutableContainer<bool> 
						 &destroyable) {
  unsigned int numNodes = 0;
  Graph *curSubGraph = hierarchyMetaGraphs->getNodeValue (curHierarchyNode);
  if (curSubGraph == NULL) return 1;
  node curSubNode;
  forEach (curSubNode, curSubGraph->getNodes ()) {
    if (hierarchyMetaGraphs->getNodeValue (curSubNode) != NULL)
      destroyable.set (curHierarchyNode.id, true);
    numNodes += nodesInHierarchy (curSubNode, destroyable);
  }
  return numNodes + 1;
}//end nodesInHierarchy

//================================================================
unsigned int 
DestroyHierarchy::getNumberNodesToDestroy (BooleanProperty &selection,
					   MutableContainer<bool> 
					   &destroyable) {
  unsigned int toDestroy = 0;
  Iterator<node> *newNodeSelection = selection.getNodesEqualTo (true);
  while (newNodeSelection->hasNext()) {
    toDestroy += nodesInHierarchy (newNodeSelection->next(), destroyable);
  } delete newNodeSelection;
  return toDestroy;
}//end getNumberNodesToDestroy

//================================================================
bool DestroyHierarchy::destroyLevelsBelow (tlp::BooleanProperty &selection) {
  bool success = true;
  MutableContainer<bool> destroyable;
  destroyable.setAll (false);
  nodesDestroyed = 0;
  nodesToDestroy = getNumberNodesToDestroy (selection, destroyable);
  vector<node> destroyedMetaNodes;
  Iterator<node> *newNodeSelection = selection.getNodesEqualTo (true);
  while (newNodeSelection->hasNext()) {
    node curHierarchyNode = newNodeSelection->next();
    Graph *hierarchyMetaGraph = 
      hierarchyMetaGraphs->getNodeValue (curHierarchyNode);
    if ((hierarchyMetaGraph != NULL) && 
	destroyable.get (curHierarchyNode.id)) {
      success = deleteLevels (hierarchyMetaGraph);
      if (!success) return false;
    }//end if
    else {
      if (hierarchyMetaGraph == NULL) ++nodesDestroyed;
      else nodesDestroyed += hierarchyMetaGraph->numberOfNodes (); 
      pluginProgress->progress (nodesDestroyed, nodesToDestroy);
    }//end else
  } delete newNodeSelection; //end while
  return success;
}//end destroyLevelsBelow
