/**
 * The following class provides functions to destroy hierarchical structure
 * below a comunity and make new structure.
 */
#ifndef __DESTROY_HIERARCHY_H__
#define __DESTROY_HIERARCHY_H__

#if (__GNUC__ < 3)
#  include <hash_map>
#else
#  include <ext/hash_map>
#endif
#include <string>
#include <tulip/BooleanProperty.h>
#include <tulip/GraphProperty.h>
#include <tulip/PluginProgress.h>
#include <tulip/QtProgressWidget.h>
#include <tulip/MutableContainer.h>
#include "MetaNodeBoundingCircle.h"
#include "GrouseInterfaceData.h"
#include "OpenCloseHierarchy.h"

class DestroyHierarchy {
 public:
  /**
   * Constructor which requires a graph hierarchy and a openCloseHierarchy
   * object.
   */
  DestroyHierarchy (GrouseInterfaceData *grouseMainWidget,
		    tlp::Graph *hierarchy, 
		    tlp::Graph *hierarchySubGraphs,
		    tlp::Graph *inputGraph,
		    const std::string &metaGraphPropertyName,
		    OpenCloseHierarchy *openCloseHierarchy,
		    MetaNodeBoundingCircles *boundingCircles,
		    PluginProgress *pluginProgress);

  /**
   * Destructor which does nothing for the time being.
   */
  inline ~DestroyHierarchy () {}

 public:
   /**
    * This function destroys all the levels of the graph hierarchy below
    * the nodes of the selection. 
    * \param selection The selection below which the levels are destroyed
    */
   bool destroyLevelsBelow (tlp::BooleanProperty &selection);

 private:

   /**
    * Deletes all hierarchical structure below deleteGraph.
    * \param deleteGraph Graph where structure is deleted below.
    */
   void deleteHierarchy (Graph *deleteGraph);

   /**
    * Function which deletes all hierarchical structure at this
    * level in the graph hierarchy.
    * \param hierarchyMetaGraph The graph which all hierarchical structure
    * is deleted.
    */
   bool deleteLevels (tlp::Graph *hierarchyMetaGraph);

   /**
    * Extracts all leaf nodes of the hierarchy below the passed
    * hierarchyMetaGraph.
    * \param hierarchyMetaGraph The graph to extract the leaves from.
    * \param leaves A selection of the leaves below this point.
    */
   void extractLeafSubGraph (tlp::Graph *hierarchyMetaGraph,
			     tlp::BooleanProperty &leaves);

   /**
    * Counts the number of nodes below the passed hierarchy node.
    * Includes both leaves and internal.
    * \param curHierarchyNode node which we want to count levels below
    * \param destroyable This is a mutable container set by the function
    * to indicate metanodes with levels below.
    */
   unsigned int nodesInHierarchy (const node &curHierarchyNode,
				  tlp::MutableContainer<bool> &destroyable);

   /**
    * Counts the number of all hierarchy nodes in the passed selection.
    * Includes both leaves and internal.
    * \param selection the selection we for which we want to know the 
    * hierarchy nodes below
    * \param destroyable This is a mutable container set by the function
    * to indicate metanodes with levels below.
    */
   unsigned int getNumberNodesToDestroy (tlp::BooleanProperty &selection,
					 tlp::MutableContainer<bool>
					 &destroyable);

  GrouseInterfaceData *grouseMainWidget;
  tlp::Graph *hierarchy;
  tlp::Graph *hierarchySubGraphs;
  tlp::Graph *inputGraph;
  tlp::GraphProperty *hierarchyMetaGraphs;
  OpenCloseHierarchy *openCloseHierarchy;
  MetaNodeBoundingCircles *boundingCircles;
  PluginProgress *pluginProgress;
  unsigned int nodesToDestroy;
  unsigned int nodesDestroyed;
};
#endif
