/**
 * Implemlementations for the grouse back end functions.  These
 * functions are all of the users actions with the QTwidgets.
 */
#if (__GNUC__ < 3)
#  include <hash_map>
#else
#  include <ext/hash_map>
#endif
#include <qlayout.h>
#include <qfiledialog.h>
#include <qinputdialog.h>
#include <qmessagebox.h>
#include <qlistbox.h>
#include <qpushbutton.h>
#include <qsplitter.h>
#include <qaction.h>
#include <tulip/TlpTools.h>
#include <tulip/GlGraph.h>
#include <tulip/GlGraphWidget.h>
#include <tulip/ForEach.h>
#include <tulip/StableIterator.h>
#include <tulip/IntegerProperty.h>
#include <tulip/StringCollection.h>
#include "FindSelection.h"
#include "InitializeHierarchyStructure.h"
#include "GrouseInterfaceBackEnd.h"
#include "HierarchyMap.h"

using namespace std;
using namespace tlp;
using namespace stdext;

const string finalLayoutDisplayLabel ("1__Final Layout Display");
const string hierarchyMetaGraphProperty ("viewMetaGraph");
const string viewLabel ("viewLabel");

//================================================================
inline void copy (Graph *src, Graph *dest) {
  node curNode;
  forEach (curNode, src->getNodes()) dest->addNode (curNode);
  edge curEdge;
  forEach (curEdge, src->getEdges()) dest->addEdge (curEdge);
}

//===============================================================
inline void startTime (clock_t &start, time_t &startLongTime) {
  time (&startLongTime);
  start = clock ();
}//end startPullOutTime

//===============================================================
inline void endTime (clock_t &start, time_t &startLongTime) {
  clock_t end = clock();
  time_t endLongTime;
  time (&endLongTime);
  double diff, diffLong;
  diff = ((double) (end - start) / (double) CLOCKS_PER_SEC);
  diffLong = difftime (endLongTime, startLongTime);
  cout << "-------------------------------------------------------" << endl;
  cout << "Action took:" << endl;
  cout << "In short time " << diff << endl;
  cout << "In long time " <<  diffLong << endl; 
  cout << "-------------------------------------------------------" << endl;
}//end endPullOutTime

//=======================================================================
void GrouseInterfaceBackEnd::initDisplayAttributeProperties () {
  displayAttributes->clear();
  Iterator<string> * propIt = hierarchy->getProperties();
  while (propIt->hasNext()) {
    string curPropName = propIt->next();
    PropertyInterface *curProperty = hierarchy->getProperty (curPropName);
    if ((curPropName.find ("view") == string::npos) && 
	(curPropName.find ("Original Edge ID") == string::npos) &&
	((PropertyInterface::getTypename (curProperty) == "string") ||
	 (PropertyInterface::getTypename (curProperty) == "double") ||
	 (PropertyInterface::getTypename (curProperty) == "boolean") ||
	 (PropertyInterface::getTypename (curProperty) == "int"))) 
      displayAttributes->insertItem (curPropName);
  } delete propIt; //end while
  displayAttributes->setSelected (0, true);
}//end getListViewProperties

//================================================================
void setGraphView (GlGraph *glGraph, Graph *newGraph) {
  GlGraphRenderingParameters param = glGraph->getRenderingParameters();
  param.setEdgeColorInterpolate (false);
  param.setEdgeSizeInterpolate (false);
  param.setViewArrow (false);
  param.setBackgroundColor (BACKGROUND_COLOUR);
  param.setGraph (newGraph);
  param.setLabelsBorder (0);
  glGraph->setRenderingParameters(param);
}//end setGraphView

//===============================================================
void GrouseInterfaceBackEnd::qtConnections () {
  connect (metaGraphHierarchyWidget1, 
	   SIGNAL (selectionChanged (const std::vector<tlp::node> &,
				     const std::vector<tlp::node> &,
				     const std::vector<tlp::node> &)),
	   this,
	   SLOT (metaGraphHierarchyWidget1_selectionChanged (const std::vector<tlp::node> &, const std::vector<tlp::node> &, const std::vector<tlp::node> &)));
  connect (metaGraphHierarchyWidget1,
	   SIGNAL (newPath (const std::vector<tlp::node> &)),
	   this,
	   SLOT (metaGraphHierarchyWidget1_newRightButtonPath (const std::vector<tlp::node> &)));
  connect (grouseSelectionProperty,
	   SIGNAL (selectionChanged (tlp::BooleanProperty &, const tlp::node &)),
	   	   this,
	   SLOT (graphWidgetSelectionChanged (tlp::BooleanProperty &,
					      const tlp::node &)));
  connect (this,
	   SIGNAL (graphWidgetSelectionChangedSignal (tlp::BooleanProperty &,
						      const tlp::node &)),
	   metaGraphHierarchyWidget1,
	   SLOT (selectionChanged (tlp::BooleanProperty &,
				   const tlp::node &)));
  connect (metaGraphHierarchyWidget1,
	   SIGNAL (labelDataChanged ()),
	   this,
	   SLOT (updateLabels ()));
}
//================================================================
void GrouseInterfaceBackEnd::initInterface (Graph *graph) {
  rootGraph = graph;
  //set up the metagraph hierarchy structure
  initFinalLayout (finalLayout, rootGraph);
  initHierarchy (hierarchy, hierarchySubGraphs, 
		 hierarchyRootNode, hierarchyMetaGraphProperty,
		 rootGraph, disconnected);
  initDisplayAttributeProperties();
  unsigned int depth = computeDepth (hierarchy);
  boundingCircles = new MetaNodeBoundingCircles (renderingWidget, hierarchy,
						 depth);
  rotatingColours = new RotatingColours ();
  openCloseHierarchy = 
    new OpenCloseHierarchy (renderingWidget,
			    hierarchy, hierarchyRootNode, 
			    hierarchySubGraphs, NULL,
			    finalLayout, boundingCircles,
			    metaGraphHierarchyWidget1,
			    graphLayoutProgress,
			    layoutScale_ModeAction->isOn (), 
			    layoutDefault_AlgorithmsAction->isOn ());
  
  finalLayoutDisplay = openCloseHierarchy->getDisplayGraph (); 
  finalLayoutDisplay->setAttribute<string> ("name", finalLayoutDisplayLabel);
  metaGraphHierarchyWidget1
    ->setHierarchyProperties (hierarchyMetaGraphProperty, hierarchy,
			      finalLayout);
  grouseSelectionProperty = new GrouseSelectionProperty (renderingWidget,
							 finalLayoutDisplay,
							 hierarchy,
							 openCloseHierarchy);
 
  createHierarchy = new CreateHierarchy (this, hierarchy, hierarchySubGraphs,
					 hierarchyMetaGraphProperty,
					 openCloseHierarchy,
					 rotatingColours,
					 graphLayoutProgress);
  destroyHierarchy = new DestroyHierarchy (this, hierarchy, hierarchySubGraphs,
					   finalLayout,
					   hierarchyMetaGraphProperty,
					   openCloseHierarchy,
					   boundingCircles,
					   graphLayoutProgress);
  createWithSelection = 
    new CreateWithSelection (this, hierarchy, hierarchySubGraphs,
			     hierarchyMetaGraphProperty,
			     openCloseHierarchy);
  createAcrossBoundaries =
    new CreateAcrossBoundaries (this, finalLayoutDisplay,
				hierarchy, hierarchySubGraphs,
				openCloseHierarchy, 
				hierarchyMetaGraphProperty);
  pullOutMode = new PullOut (hierarchy, hierarchySubGraphs, finalLayout, 
			     hierarchyMetaGraphProperty,
			     openCloseHierarchy, rotatingColours,
			     destroyHierarchy,
			     metaGraphHierarchyWidget1);
  grouseMorph = new GrouseMorphing (renderingWidget,
				    openCloseHierarchy,
				    pullOutMode);
  baseMouse = new MouseGrouseBaseNavigate (openCloseHierarchy, 
					   grouseSelectionProperty);
  hierarchyMouse = new MouseGrouseHierarchyNavigate (openCloseHierarchy,
						     metaGraphHierarchyWidget1,
						     grouseMorph);
  pullOutMouse = new MouseGrousePullOut (grouseMorph);
  flowMouse = new MouseFlowSelection (grouseSelectionProperty,
				      openCloseHierarchy);

  renderingWidget->resetInteractors (baseMouse);
  renderingWidget->pushInteractor (hierarchyMouse);
  renderingWidget->pushInteractor (pullOutMouse);
  renderingWidget->pushInteractor (flowMouse);
  setGraphView (renderingWidget, finalLayoutDisplay);
  qtConnections();
  optionsHueForMatchesToggled (optionsUse_Hue_for_MatchesAction->isOn ());
  renderingWidget->centerScene();
  openCloseHierarchy->updateDisplayGraphRendering();
  renderingWidget->draw();
  if (disconnected) {
    openCloseHierarchy->setHierarchyChange ();
    node displayNode = openCloseHierarchy->getDisplayNode (hierarchyRootNode);
    grouseMorph->morphOpen (displayNode, false);
  }//end if
}

//================================================================
/**
 * This implements the action taken when the the open functions of
 * the file menu is selected.
 */
void GrouseInterfaceBackEnd::fileOpen() {
  QString s = QFileDialog::getOpenFileName("",
					   "Tlp (*.tlp *.tlp.gz)",
					   this,
					   "open file dialog",
					   "Choose a tulip file" );
  if (s != QString::null) {
    Graph *graph = tlp::loadGraph(s.latin1());
    initInterface (graph);
  }//end if
}
//================================================================
void GrouseInterfaceBackEnd::layoutScale_Mode() {
  openCloseHierarchy->setScaleMode (layoutScale_ModeAction->isOn ());
}

//================================================================
void GrouseInterfaceBackEnd::optionsLabels() {
  GlGraphRenderingParameters param = renderingWidget->getRenderingParameters();
  if (optionsLabelsAction->isOn ()) {
    param.setViewNodeLabel (true);
    param.setFontsType (1);
  }
  else {
    param.setViewNodeLabel (false);
  }
  this->updateLabels();
  renderingWidget->setRenderingParameters (param);
  renderingWidget->draw();
}

//================================================================
void GrouseInterfaceBackEnd::optionsNumAnimationFrames() {
  int curNumFrames = GrouseMorphing::getNumAnimationFrames();
  bool ok = false;
  int newNumFrames = QInputDialog::getInteger ("Grouse",
					       "Please enter the number of iterations you would like during animations.",
					       curNumFrames,
					       0, 0x7fffffff, 5, &ok, this);
  if (ok) GrouseMorphing::setNumAnimationFrames (newNumFrames);    
}
//================================================================
void GrouseInterfaceBackEnd::optionsMaximumPathBoundAction_activated() {
  int curMaxPath = MouseFlowSelection::getMaxPath();
  bool ok = false;
  int newMaxPath = QInputDialog::getInteger ("Grouse",
					     "Please enter the maximum path you are willing to search for.",
					       curMaxPath,
					       0, 0x7fffffff, 5, &ok, this);
  if (ok) MouseFlowSelection::setMaxPath (newMaxPath);    
}

//================================================================
void GrouseInterfaceBackEnd::optionsMinimumSubgraphSize() {
  int minSize = openCloseHierarchy->getMaxNodeNum ();
  bool ok = false;
  int newSize = QInputDialog::getInteger ("Grouse",
					  "Please enter a new minimum subgraph size.  Anything greater will be coarsened automatically.",
					       minSize,
					       0, 0x7fffffff, 5, &ok, this);
  if (ok) openCloseHierarchy->setMaxNodeNum (newSize);  
}//end optionsMinimumSubgraphSize

//================================================================
void GrouseInterfaceBackEnd::optionsHueForMatchesToggled (bool turnOn) {
  createHierarchy->turnOnRotatingColours (turnOn);
  pullOutMode->turnOnRotatingColours (turnOn);
}//end optionsHueForMatchesToggled

//================================================================
void GrouseInterfaceBackEnd::optionsLogSizeNodesToggled (bool turnOn) {
  openCloseHierarchy->setLogSize (turnOn);
}//end optionsHueForMatchesToggled

//================================================================
void GrouseInterfaceBackEnd::layoutUseDefaultAlgorithms () {
  openCloseHierarchy->setUseDefaultAlgorithms (layoutDefault_AlgorithmsAction->isOn ());
}

//================================================================
void GrouseInterfaceBackEnd::layoutEdgeBundlingAction_activated () {
  openCloseHierarchy->setEdgeBundling (layoutEdgeBundlingAction->isOn ());
  openCloseHierarchy->updateDisplayGraphRendering ();
  renderingWidget->draw ();
}

//================================================================
void GrouseInterfaceBackEnd::optionsGrey_ScaleAction_toggled (bool turnOn) {
  rotatingColours->setGreyScale (turnOn);
}

//================================================================

/**
 * Function which closes grouse
 */
void GrouseInterfaceBackEnd::fileExit() {
  exit(EXIT_SUCCESS);
}

//================================================================
void GrouseInterfaceBackEnd::fileSaveAs() {
  QString saveFile = QFileDialog::getSaveFileName("",
						  "Tlp (*.tlp *.tlp.gz)",
						  this,
						  "Save As dialog",
						  "Save As File name" );
   tlp::saveGraph (rootGraph, saveFile); 
}

//================================================================
void GrouseInterfaceBackEnd::windowActivationChange (bool oldActivation) {
  if (!oldActivation) {
    this->repaint();
    renderingWidget->draw();
  }//end if
}//end showEvent

//================================================================
unsigned int GrouseInterfaceBackEnd::computeDepth (Graph *hierarchy) {
  GraphProperty *hierarchyMetaGraphs = 
    hierarchy->getProperty<GraphProperty> (hierarchyMetaGraphProperty);
  unsigned int maxDepth = 0;
  node curNode;
  forEach (curNode, hierarchy->getNodes()) {
    Graph *curSubGraph = hierarchyMetaGraphs->getNodeValue (curNode);
    if (curSubGraph != NULL) {
      unsigned int curDepth = computeDepth (curSubGraph);
      if (curDepth > maxDepth) maxDepth = curDepth;
    }//end if
  }
  return maxDepth + 1;
}

//================================================================
void GrouseInterfaceBackEnd::
metaGraphHierarchyWidget1_newRightButtonPath (const vector<node> &newPath)
{
  StringProperty *labels = hierarchy->getProperty<StringProperty> (viewLabel);
  metaGraphHierarchyWidget1->clearSelection();
  for (int i = newPath.size() - 1; i >= 0; --i) {
    node hierNode = newPath[i];
    node displayNode = openCloseHierarchy->getDisplayNode (hierNode);
    if (displayNode.isValid()) {
      if (displayNode == hierNode) continue;
      grouseMorph->morphOpen (displayNode, false);
      metaGraphHierarchyWidget1->synchListViewColour (hierNode);
    }
    else
      cerr << "could not open " << labels->getNodeValue (hierNode) << endl;
  }//end for
}//end rightButton

//================================================================
void GrouseInterfaceBackEnd::
metaGraphHierarchyWidget1_selectionChanged (const vector<node> &hierNodes,
					    const vector<node> &hierRemoved,
					    const vector<node> &hierAdded) {
  vector<node> newSelNodes;
  vector<GlABoundingCircle *> newSelBoundCircles;
  for (unsigned int i = 0; i < hierNodes.size(); ++i) {
    GlABoundingCircle *circle = 
      boundingCircles->getBoundingCircle (hierNodes[i]);
    if (circle != NULL)
      newSelBoundCircles.push_back (circle);
    else if (hierNodes[i].isValid()) {
      node displayNode = openCloseHierarchy->getDisplayNode (hierNodes[i]);
      newSelNodes.push_back (displayNode);
    }//end else if
  }//end for
  
  if (!hierRemoved.empty()) 
    grouseMorph->morphChangedSelection (hierRemoved, false);
  if (!hierAdded.empty()) 
    grouseMorph->morphChangedSelection (hierAdded, true);
  grouseSelectionProperty->setSelection (newSelNodes, newSelBoundCircles);
  renderingWidget->draw();
}//end selectionChanged

//================================================================
void GrouseInterfaceBackEnd::enableDecompositionGUI (bool enable) {
  createCommunitiesBelow->setEnabled(enable);
  createCommunitiesWithSelection->setEnabled(enable);
}
//================================================================
void GrouseInterfaceBackEnd::
graphWidgetSelectionChanged (tlp::BooleanProperty &selection, 
			     const tlp::node &lastSelected) {
  node oneNode;
  Iterator<node> *nodes = selection.getNodesEqualTo (true);
  if (nodes->hasNext()) oneNode = nodes->next();
  enableDecompositionGUI (oneNode.isValid());
  emit graphWidgetSelectionChangedSignal (selection, lastSelected);
}//end graphWidgetSelectionChanged

//================================================================
bool 
GrouseInterfaceBackEnd::findInHierarchySubGraphs (const node &hierarchyNode,
						  const string &searchString,
						  const string &attributeName,
						  vector<node> &nodesSelected,
						  unsigned int &nodesDone) {
  Graph *subGraph = openCloseHierarchy->getHierarchyGraph (hierarchyNode);
  if (subGraph == NULL) {
    Graph *tmpSubGraph = hierarchy->getRoot()->addSubGraph ();
    tmpSubGraph->addNode (hierarchyNode);
    BooleanProperty subGraphSelection (tmpSubGraph);
    PropertyInterface *property = tmpSubGraph->getProperty (attributeName);
    FindSelection *curFindSelection = new FindSelection (tmpSubGraph, property);
    curFindSelection->find (searchString, &subGraphSelection);
    delete curFindSelection;
    hierarchy->getRoot()->delSubGraph (tmpSubGraph);
    graphLayoutProgress->progress (nodesDone++, numberOfNodesInHierarchy);
    if (subGraphSelection.getNodeValue (hierarchyNode)) {
      nodesSelected.push_back (hierarchyNode); //this is a leaf
    }
    return subGraphSelection.getNodeValue (hierarchyNode);
  }
  if (graphLayoutProgress->state () != TLP_CONTINUE)
    return false;

  BooleanProperty subGraphSelection (subGraph);
  PropertyInterface *property = subGraph->getProperty (attributeName);
  FindSelection *curFindSelection = new FindSelection (subGraph, property);
  curFindSelection->find (searchString, &subGraphSelection);
  delete curFindSelection;
  
  bool selectMe= false;
  node subHierarchyNode;
  forEach (subHierarchyNode, subGraph->getNodes()) {
    if (graphLayoutProgress->state () != TLP_CONTINUE)
      returnForEach (false);
    if (openCloseHierarchy->getHierarchyGraph (subHierarchyNode) == NULL) {
      graphLayoutProgress->progress (nodesDone++, numberOfNodesInHierarchy);
      if (subGraphSelection.getNodeValue (subHierarchyNode)) {
	node displayNode = 
	  openCloseHierarchy->getDisplayNode (subHierarchyNode);
	nodesSelected.push_back (displayNode);
	selectMe = true;
      }//end if
    }//end if
  }//end forEach
  
  forEach (subHierarchyNode, subGraph->getNodes()) {
    bool foundSomething = false;
    if (graphLayoutProgress->state () != TLP_CONTINUE)
      returnForEach (false);
    if (openCloseHierarchy->getHierarchyGraph (subHierarchyNode) != NULL) {
      foundSomething =
	findInHierarchySubGraphs (subHierarchyNode, searchString,
				  attributeName, nodesSelected,
				  nodesDone);
      graphLayoutProgress->progress (nodesDone++, numberOfNodesInHierarchy);
    }//end if
    selectMe = foundSomething || selectMe;
  }//end forEach
  
  if (selectMe) {
    node displayNode = 
      openCloseHierarchy->getDisplayNode (hierarchyNode);
    nodesSelected.push_back (displayNode);
  }//end if
}//end findInHierarchySubGraphs

//================================================================
void GrouseInterfaceBackEnd::fixGUIAfterDecomp () {
  createCommunitiesBelow->setEnabled(false);
  createCommunitiesWithSelection->setEnabled(false);
  findExpression->clear();
  grouseSelectionProperty->clearSelection();
}//end fixGUIAfterDecomp

//================================================================
bool GrouseInterfaceBackEnd::catagorizeBelow (const node &hierarchyNode,
					      const string &searchString,
					      const string &attributeName,
					      StringProperty 
					      *bracketMatchProperty,
					      unsigned int &nodesDone) {
  if (graphLayoutProgress->state () != TLP_CONTINUE)
    return false;
  Graph *subGraph = openCloseHierarchy->getHierarchyGraph (hierarchyNode);
  if (subGraph == NULL) {
    Graph *tmpSubGraph = hierarchy->getRoot()->addSubGraph ();
    tmpSubGraph->addNode (hierarchyNode);
    PropertyInterface *property = tmpSubGraph->getProperty (attributeName);
    FindSelection *curFindSelection = new FindSelection (tmpSubGraph, property);
    curFindSelection->findCaptureText (searchString, bracketMatchProperty);
    delete curFindSelection;
    hierarchy->getRoot()->delSubGraph (tmpSubGraph);
    graphLayoutProgress->progress (nodesDone++, numberOfNodesInHierarchy);
    return false;
  }

  //set bracket match property for this subgraph
  PropertyInterface *property = subGraph->getProperty (attributeName);
  FindSelection *curFindSelection = new FindSelection (subGraph, property);
  curFindSelection->findCaptureText (searchString, bracketMatchProperty);
  delete curFindSelection;

  bool selectMe = false;
  hash_map<string, bool> seenCatagory;
  node curNode;
  forEach (curNode, subGraph->getNodes()) {
    if (graphLayoutProgress->state () != TLP_CONTINUE)
      returnForEach (false);
    if (openCloseHierarchy->getHierarchyGraph (curNode) == NULL) {
      graphLayoutProgress->progress (nodesDone++, numberOfNodesInHierarchy);
      string nodeCatagory = bracketMatchProperty->getNodeValue (curNode);
      if (seenCatagory.find (nodeCatagory) == seenCatagory.end())
	seenCatagory[nodeCatagory] = true;
    }//end if
  }//end forEach
  if (seenCatagory.size() > 1) selectMe = true;

  //proceed to remaining.
  node subHierarchyNode;
  forEach (subHierarchyNode, subGraph->getNodes()) {
    bool foundSomething = false;
    if (graphLayoutProgress->state () != TLP_CONTINUE)
      returnForEach (false);
    if (openCloseHierarchy->getHierarchyGraph (hierarchyNode) != NULL) {
      foundSomething = catagorizeBelow (subHierarchyNode, searchString,
					attributeName, bracketMatchProperty,
					nodesDone);
      graphLayoutProgress->progress (nodesDone++, numberOfNodesInHierarchy);
    }//end if
    selectMe = selectMe || foundSomething;
  }//end forEach
  return selectMe;
}//end catagorize below

//================================================================
void GrouseInterfaceBackEnd::findExpression_returnPressed() {
  vector<node> nodesSelected;
  vector<GlABoundingCircle *> augsSelected;
  node displayNode;
  graphLayoutProgress->reset_continue ();
  graphLayoutProgress->enable_buttons (false, true);
  numberOfNodesInHierarchy = hierarchySubGraphs->numberOfNodes () + 1;
  unsigned int nodesDone = 0;
  graphLayoutProgress->progress (0, numberOfNodesInHierarchy);
  searchString = findExpression->text().latin1();
  if (searchString[0] != '^') 
    searchString = ".*" + searchString;
  if (searchString[searchString.length() - 1] != '$')
    searchString = searchString + ".*";
  string attributeName = displayAttributes->currentText();
  clock_t start;
  time_t startLong;
  cout << "Timing selection..." << endl;
  startTime (start, startLong);
  if ((searchString.find ("(") == string::npos) &&
      (searchString.find (")") == string::npos)) {
    runDecompAlgorithm = PATTERN_MATCH;
    forEach (displayNode, finalLayoutDisplay->getNodes()) {
      node hierarchyNode = openCloseHierarchy->getHierarchyNode (displayNode);
      findInHierarchySubGraphs (hierarchyNode, searchString,
				attributeName, nodesSelected,
				nodesDone);
      if (graphLayoutProgress->state () != TLP_CONTINUE)
	breakForEach;
    }//end forEach
  }//end if no brackets
  else {
    runDecompAlgorithm = CATEGORY;
    StringProperty *bracketMatchProperty =
      hierarchy->getRoot()->getLocalProperty<StringProperty> ("view Bracket Match Property");
    bracketMatchProperty->setAllNodeValue ("");
    bracketMatchProperty->setAllEdgeValue ("");
    forEach (displayNode, finalLayoutDisplay->getNodes()) {
      node hierarchyNode = openCloseHierarchy->getHierarchyNode (displayNode);
      bool select = catagorizeBelow (hierarchyNode, searchString,
				     attributeName, bracketMatchProperty,
				     nodesDone);
      if (graphLayoutProgress->state () != TLP_CONTINUE)
	breakForEach;
      if (select)
	nodesSelected.push_back (displayNode);
    }//end forEach
  }//end with brackets
  endTime (start, startLong);

  if (graphLayoutProgress->state () == TLP_CANCEL)
    grouseSelectionProperty->clearSelection ();
  else {
    node selNode;
    if (!nodesSelected.empty()) selNode = nodesSelected[0]; 
    grouseSelectionProperty->setMouseSelection (nodesSelected, augsSelected,
						selNode);
  }//end else
}//end findExpression_returnPressed

//================================================================
void GrouseInterfaceBackEnd::createCommunitiesBelow_clicked() {
  BooleanProperty &grouseSelection = 
    grouseSelectionProperty->getHierarchySelection();
  BooleanProperty &visibleSelection =
    grouseSelectionProperty->getVisibleSelection();
  StringProperty *bracketMatchProperty =
      hierarchy->getRoot()->getLocalProperty<StringProperty> ("view Bracket Match Property");
  graphLayoutProgress->reset_continue ();
  graphLayoutProgress->enable_buttons (false, false);
  graphLayoutProgress->repaint();
  
  clock_t start;
  time_t startLong;
  cout << "Timing decomposition..." << endl;
  startTime (start, startLong);
  destroyHierarchy->destroyLevelsBelow (visibleSelection);
  DataSet localData;
  localData.set ("selection", &grouseSelection);
  localData.set ("Property", bracketMatchProperty);
  localData.set ("searchString", searchString);
  graphLayoutProgress->reset_continue ();
  createHierarchy->newHierarchyLevel (visibleSelection, runDecompAlgorithm,
				      &localData);
  fixGUIAfterDecomp ();
  graphLayoutProgress->reset_continue ();
  graphLayoutProgress->enable_buttons (true, false);
  graphLayoutProgress->repaint();
  Iterator<node> *visibleSelectionIt = visibleSelection.getNodesEqualTo (true);
  while (visibleSelectionIt->hasNext()) {
    if (graphLayoutProgress->state () != TLP_CONTINUE)
      break;
    node hierarchyNode = visibleSelectionIt->next();
    node displayNode = openCloseHierarchy->getDisplayNode (hierarchyNode);
    openCloseHierarchy->setHierarchyChange ();
    grouseMorph->morphOpen (displayNode, false);
  } delete visibleSelectionIt;
  endTime (start, startLong);
  graphLayoutProgress->reset_continue ();
  graphLayoutProgress->enable_buttons (true, true); 
  graphLayoutProgress->repaint();
}

//================================================================
void GrouseInterfaceBackEnd::clearSelection_clicked () {
  grouseSelectionProperty->clearSelection ();
}

//================================================================
void GrouseInterfaceBackEnd::createCommunitiesWithSelection_clicked () {
  vector<node> newHierarchyNodes;
  BooleanProperty &visibleSelection =
    grouseSelectionProperty->getVisibleSelection();
  Graph *displayGraph = openCloseHierarchy->getDisplayGraph ();
  if (!optionsAcross_Metanode_GroupingAction->isOn ())
    createWithSelection->mergeSelections (visibleSelection, newHierarchyNodes);
  else
    createAcrossBoundaries->mergeAcrossBoundaries (visibleSelection,
						   newHierarchyNodes);
  fixGUIAfterDecomp ();
  for (unsigned int i = 0; i < newHierarchyNodes.size(); ++i) {
    node newNode = newHierarchyNodes[i];
    node superNode = openCloseHierarchy->getContainingHierarchyNode (newNode);
    metaGraphHierarchyWidget1->updateTree (superNode);
    Coord globalPos = openCloseHierarchy->getGlobalPos (newNode);
    grouseMorph->morphClosed (newNode, globalPos, false);
    grouseMorph->morphLayoutChangeSetup ();
    openCloseHierarchy->redrawPath (newNode);
    grouseMorph->morphLayoutChangeExecute ();
  }//end for
}//end createCommunitiesWithSelection_clicked

//================================================================
void GrouseInterfaceBackEnd::updateLabels () {
  Graph *displayGraph = openCloseHierarchy->getDisplayGraph ();
  StringProperty *labels = 
    displayGraph->getProperty<StringProperty> ("viewLabel");
  node curDisplayNode;
  forEach (curDisplayNode, displayGraph->getNodes()) {
    node hierNode = openCloseHierarchy->getHierarchyNode (curDisplayNode);
    labels->setNodeValue (curDisplayNode,
			  metaGraphHierarchyWidget1->getLabel (hierNode));
  }//end forEach
  renderingWidget->draw();
}//end updateLabels
