/**
 * Implements the back end interface of grouse.  This class essentially
 * takes the QT callbacks and links it to the grouse code.
 */
#if (__GNUC__ < 3)
#  include <hash_map>
#else
#  include <ext/hash_map>
#endif
#include <string>
#include <tulip/Graph.h>
#include <tulip/GWInteractor.h>
#include "GrouseInterfaceData.h"
#include "MetaNodeBoundingCircle.h"
#include "GrouseMorphing.h"
#include "MouseGrouseBaseNavigate.h"
#include "MouseGrouseHierarchyNavigate.h"
#include "MouseGrousePullOut.h"
#include "MouseFlowSelection.h"
#include "OpenCloseHierarchy.h"
#include "GrouseSelectionProperty.h"
#include "CreateHierarchy.h"
#include "DestroyHierarchy.h"
#include "CreateWithSelection.h"
#include "CreateAcrossBoundaries.h"
#include "PullOut.h"
#include "RotatingColours.h"

class GrouseInterfaceBackEnd : public GrouseInterfaceData {
  Q_OBJECT;
 public:
  GrouseInterfaceBackEnd (QWidget *parent = 0, const char *name = NULL,
			  const char *WFlags = NULL) :
    boundingCircles (NULL) {}
  ~GrouseInterfaceBackEnd () {
    delete grouseMorph;
    delete baseMouse;
    delete hierarchyMouse;
    delete pullOutMouse;
    delete flowMouse;
    delete boundingCircles;
    delete grouseSelectionProperty;
    delete createHierarchy;
    delete createWithSelection;
    delete destroyHierarchy;
    delete rotatingColours;
  };

  //================================================================
  /**
   * finds a graph with the passed name.  If the name is not found
   * returns a null pointer.
   */
  Graph *findGraph (const std::string &graphName);

  //================================================================
  /**
   * Computes the depth of the graph hierarchy.
   */
  unsigned int computeDepth (tlp::Graph *hierarchy);
  
  //================================================================
  /**
   * The function initializes the rendering widgets to their initial
   * startup state after a file has been loaded.  The graph just
   * loaded is passed as the parameter.
   */
  void initInterface (tlp::Graph *graph);

  //================================================================
  /**
   * Loads all numeric and string attributes into the display attribute
   * property.  The function also excludes all "view" attributes.
   */
  void initDisplayAttributeProperties ();

 signals:
  void graphWidgetSelectionChangedSignal (tlp::BooleanProperty &selection,
					  const tlp::node &lastSelected);

 private slots:
  void metaGraphHierarchyWidget1_newRightButtonPath (const std::vector<tlp::node> &newPath);
  void metaGraphHierarchyWidget1_selectionChanged (const std::vector<tlp::node> &newHierNodes, const std::vector<tlp::node> &hierRemoved, const std::vector<tlp::node> &hierAdded);
  void graphWidgetSelectionChanged (tlp::BooleanProperty &selection,
				    const tlp::node &lastSelected);
  void findExpression_returnPressed();
  void createCommunitiesBelow_clicked();
  void createCommunitiesWithSelection_clicked();
  void clearSelection_clicked();
  void updateLabels ();

 protected:
  void windowActivationChange (bool oldActivation);

 private:
  std::string runDecompAlgorithm; 
  std::string searchString;
  bool disconnected;
  unsigned int numberOfNodesInHierarchy;
  MetaNodeBoundingCircles *boundingCircles;
  GrouseMorphing *grouseMorph;
  MouseGrouseBaseNavigate *baseMouse;
  MouseGrouseHierarchyNavigate *hierarchyMouse;
  MouseGrousePullOut *pullOutMouse;
  MouseFlowSelection *flowMouse;
  CreateHierarchy *createHierarchy;
  DestroyHierarchy *destroyHierarchy;
  RotatingColours *rotatingColours;
  CreateWithSelection *createWithSelection;
  CreateAcrossBoundaries *createAcrossBoundaries;
  PullOut *pullOutMode;
  tlp::Graph *rootGraph;
  tlp::node hierarchyRootNode;
  OpenCloseHierarchy *openCloseHierarchy;
  GrouseSelectionProperty *grouseSelectionProperty;
  tlp::Graph *finalLayout;
  tlp::Graph *hierarchy;
  tlp::Graph *hierarchySubGraphs;
  tlp::Graph *finalLayoutDisplay;
  void fileOpen();
  void fileExit();
  void fileSaveAs();
  void layoutScale_Mode();
  void layoutUseDefaultAlgorithms();
  void layoutEdgeBundlingAction_activated();
  void optionsLabels();
  void optionsNumAnimationFrames();
  void optionsMinimumSubgraphSize();
  void optionsMaximumPathBoundAction_activated();
  void optionsHueForMatchesToggled (bool turnOn);
  void optionsLogSizeNodesToggled (bool turnOn);
  void optionsGrey_ScaleAction_toggled (bool turnOn);
  void qtConnections();
  void enableDecompositionGUI (bool enable);
  void fixGUIAfterDecomp ();
  bool findInHierarchySubGraphs (const tlp::node &hierarchyNode,
				 const std::string &searchString,
				 const std::string &attributeName,
				 std::vector<tlp::node> &nodesSelected,
				 unsigned int &nodesDone);
  bool catagorizeBelow (const node &hierarchyNode,
			const string &searchString,
			const string &attributeName,
			StringProperty *bracketMatchProperty,
			unsigned int &nodesDone);
};
