#include <queue>
#include <tulip/ColorProperty.h>
#include <tulip/SizeProperty.h>
#include "GrouseMorphing.h"

using namespace tlp;
using namespace std;
using namespace stdext;

//===============================================================
#define DEFAULT_NUM_ANIMATION_FRAMES (10)
int GrouseMorphing::numAnimationFrames = DEFAULT_NUM_ANIMATION_FRAMES;

//===============================================================
GrouseMorphing::GrouseMorphing (GlGraphWidget *graphWidget, 
				OpenCloseHierarchy *openCloseHierarchy,
				PullOut *pullOut) :
  graphWidget (graphWidget),
  gStart (NULL),
  lChangeOrder (NULL),
  openCloseHierarchy (openCloseHierarchy),
  pullOut(pullOut) {}

//===============================================================
GrouseMorphing::~GrouseMorphing () {
  if (gStart != NULL) delete gStart;
  if (lChangeOrder != NULL) delete lChangeOrder;
}

//===============================================================
//The following helper function fixes the proxies in the graph state
//so that they can handle the addition of augmented displays, nodes
//and edges.
void fixForAdditions (const vector<node> &nodesAdded,
		      const vector<edge> &edgesAdded,
		      const vector<GlABoundingCircle *> &augsAdded,
		      stdext::hash_map<node, unsigned int> &order,
		      const Coord &startPos,
		      GraphState *gi, GraphState *gf,
		      GlGraphWidget *glW) {
  ColorProperty *colours = 
    glW->getGraph()->getProperty<ColorProperty> ("viewColor");
  SizeProperty *sizes = 
    glW->getGraph()->getProperty<SizeProperty> ("viewSize");
  for (unsigned int i = 0; i < nodesAdded.size(); ++i) {
    gi->layout->setNodeValue (nodesAdded[i], startPos);
    gi->size->setNodeValue (nodesAdded[i],
			    sizes->getNodeValue (nodesAdded[i]));
    gi->color->setNodeValue (nodesAdded[i],
			     colours->getNodeValue (nodesAdded[i]));
  }//end for nodesAdded
  for (unsigned int i = 0; i < edgesAdded.size(); ++i) {
    gi->layout->setEdgeValue (edgesAdded[i], vector<Coord> ());
    gi->color->setEdgeValue (edgesAdded[i],
			     colours->getEdgeValue (edgesAdded[i]));
  }//end for edgesAdded

  //Adjust for aug display insertions or deletions
  if (gi->augPoints.size() < gf->augPoints.size()) {
    gi->augPoints.resize (gf->augPoints.size());
    gi->augCenters.resize (gf->augPoints.size());
  }//end if
  for (unsigned int i = 0; i < augsAdded.size(); ++i) {
    vector<Coord> addedDisplay;
    node hierarchyNode = augsAdded[i]->getHierarchyNode();
    const vector<Coord> &points = augsAdded[i]->getPoints(); 
    Coord augStartPositions = startPos;
    augStartPositions.setZ (points[0].getZ());
    for (unsigned int j = 0; j < points.size(); ++j)
      addedDisplay.push_back (augStartPositions); 
    unsigned int place = order[hierarchyNode];
    gi->augPoints[place] = addedDisplay;
    gi->augCenters[place] = augStartPositions;
  }
}

//===============================================================
//The following helper function updates the positions of the vertices involved
//in an augmented display during a morph event.
void updateAugs (GlGraphWidget *graphWidget, GraphState *g0, 
		 stdext::hash_map<node, unsigned int> &order) {
  const stdext::hash_map<std::string, GlAugmentedDisplay*> *augDisplays =
    graphWidget->getAugmentedDisplays();
  if ((augDisplays != NULL) && !g0->interpolateCenters.empty()
       && !g0->curInterpolation.empty()) {
    stdext::hash_map<string, GlAugmentedDisplay *>::const_iterator augIt = 
      augDisplays->begin();
    for (; augIt != augDisplays->end(); ++augIt) {
      GlABoundingCircle *curAug = 
	static_cast<GlABoundingCircle *> (augIt->second);
      unsigned int place = order[curAug->getHierarchyNode()];
      curAug->setPoints (g0->interpolateCenters[place],
			 g0->curInterpolation[place]);
    }
  }
}

//===============================================================
void GrouseMorphing::morphGraph (GraphState *g0, 
				 GraphState *g1,
				 stdext::hash_map<node, unsigned int> &order) {
  Morphing morph;
  bool morphable = morph.init (graphWidget, g0, g1);
  if (!morphable) return;
  float t = 0;
  for (int k = 0; k < numAnimationFrames; ++k) {
    morph.interpolate (graphWidget, t);
    if (!order.empty())
      updateAugs (graphWidget, g0, order);
    graphWidget->draw();
    t += 1.0/numAnimationFrames;
  }//end for
  graphWidget->draw();
}//end morphGraph

//===============================================================
void setupAugs (GraphState *gs,
		GlGraphWidget *graphWidget,
		stdext::hash_map<node, unsigned int> &order) {
  bool firstTime = order.empty();
  if (!firstTime) {
    gs->augPoints.resize (order.size());
    gs->augCenters.resize (order.size());
  }
  unsigned int augCounter = 0;
  const stdext::hash_map<string, GlAugmentedDisplay *> *augDisplays =
    graphWidget->getAugmentedDisplays();
  if (augDisplays != NULL) {
    stdext::hash_map<string, GlAugmentedDisplay *>::const_iterator augIt = 
      augDisplays->begin();
    for (; augIt != augDisplays->end(); ++augIt) {
      GlABoundingCircle *curAug = 
	static_cast<GlABoundingCircle *> (augIt->second);
      node hierarchyNode = curAug->getHierarchyNode();
      const vector<Coord> points = curAug->getPoints();
      if (firstTime) {
	gs->augPoints.push_back (points);
	gs->augCenters.push_back (curAug->getPos());
	order[hierarchyNode] = augCounter;
	augCounter++;
      }
      else {
	if (order.find (hierarchyNode) != order.end()) {
	  unsigned int place = order[hierarchyNode];
	  gs->augPoints[place] = points;
	  gs->augCenters[place] = curAug->getPos();
	}
	else {
	  order[hierarchyNode] = gs->augPoints.size();
	  gs->augPoints.push_back (points);
	  gs->augCenters.push_back (curAug->getPos());
	}
      }
    }
  }
}

//===============================================================
//The following function determines the zoom factor given information
//about a node and the camera.
inline double checkNewZoom (double nodeCoord, double camCoord,
			    double halfCameraSceneRadius,
			    double newZoomFactor) {
  double newZoom = halfCameraSceneRadius/fabs (nodeCoord - camCoord);
  if (newZoom < newZoomFactor) return newZoom;
  else return newZoomFactor;
}

//===============================================================
void GrouseMorphing::morphCamera (const Coord &nodePos,
				  const double nodeRadius) {
  GlGraphRenderingParameters parameters = 
    graphWidget->getRenderingParameters();
  Camera camera = parameters.getCamera();
  double newZoomFactor = camera.zoomFactor;

  newZoomFactor = checkNewZoom (nodePos.getX() + nodeRadius, 
				camera.center.getX(),
				camera.sceneRadius/2.0,
				newZoomFactor);
  newZoomFactor = checkNewZoom (nodePos.getX() - nodeRadius, 
				camera.center.getX(),
				camera.sceneRadius/2.0, 
				newZoomFactor);
  newZoomFactor = checkNewZoom (nodePos.getY() + nodeRadius, 
				camera.center.getY(),
				camera.sceneRadius/2.0,
				newZoomFactor);
  newZoomFactor = checkNewZoom (nodePos.getY() - nodeRadius, 
				camera.center.getY(),
				camera.sceneRadius/2.0, 
				newZoomFactor);
  camera.zoomFactor = newZoomFactor;
  parameters.setCamera (camera);
  graphWidget->setRenderingParameters (parameters);
}

//===============================================================
void GrouseMorphing::nodesAndEdgesToRemove (BooleanProperty *closeSelection,
						 vector<node> &nodesRemoved,
						 vector<edge> &edgesRemoved) {
  Graph *displayGraph = openCloseHierarchy->getDisplayGraph ();
  string errMsg;

  BooleanProperty inducedSubGraph (displayGraph);
  DataSet localData;
  localData.set ("Nodes", closeSelection);
  bool success = displayGraph->computeProperty ("Induced sub-graph",
						&inducedSubGraph, errMsg, NULL,
						&localData);
  if (!success) {
    cerr << "Induced sub-graph returned the following error " << endl 
	 << errMsg << endl << "(Is the plugin installed?)" << endl;
    return;
  }//end if

  node curNode;
  forEach (curNode, inducedSubGraph.getNodesEqualTo (true))
    nodesRemoved.push_back (curNode);
  
  edge curEdge;
  forEach (curEdge, inducedSubGraph.getEdgesEqualTo (true))
    edgesRemoved.push_back (curEdge);
}//end nodesAndEdgesToRemove

//===============================================================
//The following helper function reinstates the layout inside a node
//after a close event.
void layoutReinstate (LayoutProperty *layout, GlGraphWidget *graphWidget) {
  Graph *curGraph = graphWidget->getGraph();
  LayoutProperty *curLayout = 
    curGraph->getProperty<LayoutProperty> ("viewLayout");
  node curNode;
  forEach (curNode, curGraph->getNodes())
    curLayout->setNodeValue (curNode, layout->getNodeValue (curNode));
  edge curEdge;
  forEach (curEdge, curGraph->getEdges())
    curLayout->setEdgeValue (curEdge, layout->getEdgeValue (curEdge));
}//end layoutReinstate

//===============================================================
void GrouseMorphing::morphLayoutChangeSetup () {
  openCloseHierarchy->updateDisplayGraphRendering ();
  graphWidget->draw ();
  gStart = new GraphState (graphWidget);
  lChangeOrder = new stdext::hash_map<node, unsigned int> ();
  setupAugs (gStart, graphWidget, *lChangeOrder);
}

//===============================================================
void GrouseMorphing::morphLayoutChangeExecute () {
  if ((gStart == NULL) || (lChangeOrder == NULL)) {
    cerr << "gStart or lChangeOrder has not been initialized " 
	 << "Coding bug:  Call morphLayoutChangeSetup." << endl;
    return;
  }//end if
 
  GraphState *gEnd = new GraphState (graphWidget);
  setupAugs (gEnd, graphWidget, *lChangeOrder);
  morphGraph (gStart, gEnd, *lChangeOrder);
  openCloseHierarchy->updateDisplayGraphRendering ();
  graphWidget->draw();
  delete gStart;
  delete gEnd;
  delete lChangeOrder;
  gStart = NULL;
  lChangeOrder = NULL;
}

//===============================================================
void GrouseMorphing::morphOpen (const node &pickedNode,
				const bool layoutReset) {
  node hierarchyNode = openCloseHierarchy->getHierarchyNode (pickedNode);

  vector<node> nodesAdded;
  vector<edge> edgesAdded;
  vector<GlABoundingCircle *> augsAdded;
  stdext::hash_map<node, unsigned int> order;
  GraphState *g0 = new GraphState (graphWidget);
  setupAugs (g0, graphWidget, order);
  LayoutProperty *viewLayout =
    graphWidget->getGraph()->getProperty<LayoutProperty> ("viewLayout");
  const Coord curCoord = viewLayout->getNodeValue (pickedNode);
  ColorProperty *viewColor =
    graphWidget->getGraph()->getProperty<ColorProperty> ("viewColor");
  Color prevColor = viewColor->getNodeValue (pickedNode);
  viewColor->setNodeValue (pickedNode, PROCESSING_COLOUR);
  graphWidget->draw();
  bool success = 
    openCloseHierarchy->open (pickedNode, nodesAdded, edgesAdded, augsAdded,
			      layoutReset);
  if (!success) {
    delete g0;
    viewColor->setNodeValue (pickedNode, prevColor);
    graphWidget->draw();
    return;
  }//end if

  double openNodeRadius = augsAdded[0]->getRadius();
  Coord openNodePosition = augsAdded[0]->getPos();
  morphCamera (openNodePosition, openNodeRadius);
  openCloseHierarchy->updateDisplayGraphRendering ();
  GraphState *g1 = new GraphState (graphWidget);
  setupAugs (g1, graphWidget, order);
  fixForAdditions (nodesAdded, edgesAdded, augsAdded, order,
		   curCoord, g0, g1, graphWidget);
  morphGraph (g0, g1, order);
  graphWidget->draw();
  delete g0;
  delete g1;
}

//===============================================================
void GrouseMorphing::morphClosed(const node &closeHierarchyNode,
				 const Coord &closeCentre,
				 const bool layoutReset) {
  vector<node> nodesRemoved;
  vector<edge> edgesRemoved;
  vector<GlABoundingCircle *> augsRemoved;
  openCloseHierarchy->findAugsBelow (closeHierarchyNode, augsRemoved);
  BooleanProperty closeSelection (graphWidget->getGraph ()); 
  openCloseHierarchy->getDisplaySelection (closeHierarchyNode,
					   closeSelection);
  stdext::hash_map<node, unsigned int> order;
  graphWidget->draw();
  GraphState *g0 = new GraphState (graphWidget);
  setupAugs (g0, graphWidget, order);
  
  GraphState *g1 = new GraphState (graphWidget);
  setupAugs (g1, graphWidget, order);

  nodesAndEdgesToRemove (&closeSelection, nodesRemoved, edgesRemoved);
  fixForAdditions (nodesRemoved, edgesRemoved, augsRemoved, order,
		   closeCentre, g1, g0, graphWidget);
  morphGraph (g0, g1, order);
  layoutReinstate (g0->layout, graphWidget);
  openCloseHierarchy->close (closeHierarchyNode, layoutReset);
  openCloseHierarchy->updateDisplayGraphRendering ();
  graphWidget->draw();
  delete g0;
  delete g1;
}//end morphClosed

//===============================================================
void GrouseMorphing::powerOpenNodes (const node &displayNode) {
  queue<node> toOpen;
  toOpen.push (displayNode);
  while (!toOpen.empty()) {
    node displayNode = toOpen.front();
    toOpen.pop();
    node hierarchyNode = openCloseHierarchy->getHierarchyNode (displayNode);
    Graph *hierarchySubGraph =
      openCloseHierarchy->getHierarchyGraph (hierarchyNode);
    if (hierarchySubGraph != NULL) {
      morphOpen (displayNode, false);
      node curHierNode;
      forEach (curHierNode, hierarchySubGraph->getNodes()) {
	node curDisplayNode = openCloseHierarchy->getDisplayNode (curHierNode);
	//display nodes are not equal to hier nodes for non leafs
	if (curDisplayNode.isValid() && (curDisplayNode != curHierNode))
	  toOpen.push(curDisplayNode);
      }//end forEach
    }//end if
  }//end while
}//end powerOpenNodes

//===============================================================
void GrouseMorphing::morphPullOut (const node &displayNode, 
				   unsigned int numHops) {
  vector<node> pullOutDisplays;
  Graph *displayGraph = openCloseHierarchy->getDisplayGraph ();
  pullOut->performPullOut (displayNode, numHops, pullOutDisplays);
  for (unsigned int i = 0; i < pullOutDisplays.size(); ++i) {
    node curHierarchyNode = 
      openCloseHierarchy->getHierarchyNode (pullOutDisplays[i]);
    Graph *curHierarchyGraph = 
      openCloseHierarchy->getHierarchyGraph (curHierarchyNode);
    if (curHierarchyGraph != NULL) {
      openCloseHierarchy->setHierarchyChange ();
      morphOpen (pullOutDisplays[i], false);
    }//end if
  }//end for
}//end pullOut

//===============================================================
void GrouseMorphing::morphChangedSelection (const vector<node> &diffHierSel,
					    bool added) {
  vector<Coord> positions;
  GraphState *g0, *g1;
  stdext::hash_map<node, unsigned int> order;
  Graph *displayGraph = openCloseHierarchy->getDisplayGraph ();
  SizeProperty *displaySize =
    displayGraph->getProperty<SizeProperty> ("viewSize");
  LayoutProperty *displayLayout =
    displayGraph->getProperty<LayoutProperty> ("viewLayout");
  g0 = new GraphState (graphWidget);
  for (unsigned int i = 0; i < diffHierSel.size(); ++i) {
    node displayNode = openCloseHierarchy->getDisplayNode (diffHierSel[i]);
    if (!displayNode.isValid () || 
	!displayGraph->isElement (displayNode)) continue;
    double radius;
    Coord pos;
    if (displayNode.isValid ()) {
      Size size = displaySize->getNodeValue (displayNode);
      radius = sqrt (size.getW()*size.getW() + size.getH()*size.getH());
      pos = displayLayout->getNodeValue (displayNode);
    }//end if
    else {
      GlABoundingCircle * circle =
	openCloseHierarchy->getBoundingCircle (diffHierSel[i]);
      radius = circle->getRadius();
      pos = circle->getPos();
    }//end else
    //morphCamera (pos, radius);
    positions.push_back (pos);
  }//end for
  g1 = new GraphState (graphWidget);
  morphGraph (g0, g1, order);
  delete g0;
  delete g1;

  //morph along the set of selected nodes
  for (unsigned int i = 0; i < positions.size(); ++i) {
    g0 = new GraphState (graphWidget);
    GlGraphRenderingParameters parameters = 
      graphWidget->getRenderingParameters();
    Camera camera = parameters.getCamera();
    camera.center.setX (positions[i].getX());
    camera.center.setY (positions[i].getY());
    camera.eyes.setX (positions[i].getX());
    camera.eyes.setY (positions[i].getY());
    parameters.setCamera (camera);
    graphWidget->setRenderingParameters (parameters);
    g1 = new GraphState (graphWidget);
    morphGraph (g0, g1, order);
    delete g0;
    delete g1;
  }//end for i
}//end morphChangedSelection
