#ifndef __GROUSE_MORPHING_H__
#define __GROUSE_MORPHING_H__
#if (__GNUC__ < 3)
#  include <hash_map>
#else
#  include <ext/hash_map>
#endif
#include <vector>
#include <tulip/GlGraphWidget.h>
#include <tulip/LayoutProperty.h>
#include <tulip/Morphing.h>
#include "OpenCloseHierarchy.h"
#include "MetaNodeBoundingCircle.h"
#include "PullOut.h"

//================================================================
/**
 * The following class controls the various morph open, close, and
 * powerdown animations in grouse.  It also keeps track of the framerate
 * for all animations in the program.
 */
class GrouseMorphing {
 private:
  GlGraphWidget *graphWidget;
  tlp::GraphState *gStart;
  stdext::hash_map<tlp::node, unsigned int> *lChangeOrder;
  static int numAnimationFrames;
  OpenCloseHierarchy *openCloseHierarchy;
  PullOut *pullOut;

  /**
   * Function to morph the camera given a node has changed its position
   * or size and may have gone offscreen.
   * \param nodePos The new position of the node.
   * \param nodeRadius The new radius of the node.
   */
  void morphCamera (const Coord &nodePos, const double nodeRadius);

  /**
   * Function to morph a graph given that it has undergone some change.
   * Primarily a control loop for the frames of the animation.
   * \param g0 The initial graph state we morph from.
   * \param g1 The final graph state we morph to.
   * \param order The order of the augmented displays in the graph state.
   */
  void morphGraph (tlp::GraphState *g0, 
		   tlp::GraphState *g1,
		   stdext::hash_map<tlp::node, unsigned int> &order);

  /**
   * The following function returns a list of nodes and edges that
   * will be removed from the graph during a close event.
   * \param closeSubGraph A selection proxy corresponding to the nodes
   * and edges that will be removed.
   * \param nodesRemoved List of nodes that are removed.
   * \param edgesRemoved List of edges that are removed.
   */
  void nodesAndEdgesToRemove (tlp::BooleanProperty *closeSubGraph,
			      std::vector<node> &nodesRemoved,
			      std::vector<edge> &edgesRemoved);
 public:
  /**
   * Constructor for the morphing object.
   * \param graphWidget The widget where these morphs will be displayed.
   * \param openCloseHiearchy The associated openCloseHiearchy object with
   * this graph hierarchy.
   * \param pullOut The associated pullOut object with this hierarchy.
   */
  GrouseMorphing (GlGraphWidget *graphWidget, 
		  OpenCloseHierarchy *openCloseHierarchy,
		  PullOut *pullOut);

  /**
   * Destructor for the morphing object.
   */
  ~GrouseMorphing ();

  /**
   * Function to morph open a node in the cut.
   * \param pickedNode The display cut metanode to open.
   * \param layoutReset Resets the layout after the event
   */
  void morphOpen (const tlp::node &pickedNode, const bool layoutReset);
  
  /**
   * Function to close a node in the cut.
   * \param hierarchyNode The hierarchy node corresponding to the bounding
   * circle in the cut.
   * \param closeCenter A location to close the node to.
   */
  void morphClosed (const tlp::node &hierarchyNode,
		    const tlp::Coord &closeCentre,
		    const bool layoutReset);

  /**
   * Function to set up an arbitrary layout change morph.  This function
   * cannot handle the addition or removal of nodes from the cut.
   */
  void morphLayoutChangeSetup ();

  /**
   * Function to execute the arbitrary layout morph.  This function
   * requires as a precondition that you call morphLayoutChangeSetup ().
   */
  void morphLayoutChangeExecute ();

  /**
   * Function to open a series of nodes to the leaves below the passed
   * display node.
   * \param displayNode Node below which to open to the leaves.
   */
  void powerOpenNodes (const tlp::node &displayNode);

  /**
   * Function to pull out adjacent nodes from the passed display node
   * a specified number of hops away.
   * \param displayNode The node from which pullout occurs.
   * \param numHops The number of hops away from which pull out occurs.
   */
  void morphPullOut (const node &displayNode, unsigned int numHops);

  /**
   * Function to morph along the set of nodes to be selected or 
   * deselected.
   * \param diffHierSel The difference in selection.
   * \param added True the selection is nodes added, otherwise nodes removed.
   */
  void morphChangedSelection (const std::vector<tlp::node> &diffHierSel,
			      bool added);
  
  /**
   * The function returns the number of frames in an amimation.
   */
  inline static int getNumAnimationFrames () { return numAnimationFrames; }
  
  /**
   * The function sets the number of frames in an animation.
   */
  inline static void setNumAnimationFrames (int newNumAnimationFrames) {
    numAnimationFrames = newNumAnimationFrames;
  }
};
#endif
