#include <string>
#include <tulip/DoubleProperty.h>
#include <tulip/ColorProperty.h>
#include "GrouseSelectionProperty.h" 

using namespace tlp;
using namespace std;

const string viewBorderWidth ("viewBorderWidth");
const string viewBorderColor ("viewBorderColor");
const string viewColor ("viewColor");

//========================================================================
GrouseSelectionProperty::
GrouseSelectionProperty (GlGraphWidget *displayWidget,
			 Graph *displayGraph,
			 Graph *hierarchy, 
			 OpenCloseHierarchy *openCloseHierarchy) :
  displayWidget(displayWidget),
  displayGraph(displayGraph),
  hierarchy(hierarchy),
  openCloseHierarchy(openCloseHierarchy)
{
  nodeSelection = new BooleanProperty (hierarchy);
  nodeSelection->setAllNodeValue (false);
  nodeSelection->setAllEdgeValue (false);
  edgeSelection = new BooleanProperty (displayGraph);
  edgeSelection->setAllNodeValue (false);
  edgeSelection->setAllEdgeValue (false);
  visibleSelection = new BooleanProperty (hierarchy);
  visibleSelection->setAllNodeValue (false);
  visibleSelection->setAllEdgeValue (false);
}//end constructor

//========================================================================
GrouseSelectionProperty::~GrouseSelectionProperty() {
  delete nodeSelection;
  delete edgeSelection;
  delete visibleSelection;
}

//========================================================================
void GrouseSelectionProperty::clearSelectionNoDraw () {
  //set all node displays to unselected
  DoubleProperty *borderWidth = 
    displayGraph->getProperty<DoubleProperty> (viewBorderWidth);
  borderWidth->setAllNodeValue (META_NODE_BORDER_WIDTH);
  ColorProperty *borderColor =
    displayGraph->getProperty<ColorProperty> (viewBorderColor);
  borderColor->setAllNodeValue (META_NODE_BORDER_COLOUR);

  //set all the edges to unselected
  ColorProperty *color =
    displayGraph->getProperty<ColorProperty> (viewColor);
  color->setAllEdgeValue (EDGE_COLOUR);
  
  //unselect all augmented displays
  Iterator<node> *selectionIterator = nodeSelection->getNodesEqualTo (true);
  while (selectionIterator->hasNext()) {
    GlABoundingCircle *curCircle = 
      openCloseHierarchy->getBoundingCircle (selectionIterator->next());
    if (curCircle != NULL) 
      curCircle->setSelected (false);
  } delete selectionIterator; //end while

  //set all proxies false
 nodeSelection->setAllNodeValue (false);
 nodeSelection->setAllEdgeValue (false);
 edgeSelection->setAllNodeValue (false);
 edgeSelection->setAllEdgeValue (false);
}//end clear selection

//========================================================================
void GrouseSelectionProperty::clearSelection () {
  clearSelectionNoDraw();
  displayWidget->draw();
  node lastSelectedHierarchyNode;
  emit selectionChanged (*nodeSelection, lastSelectedHierarchyNode);
}
//========================================================================
void GrouseSelectionProperty::setNodeSelection (const node &displayNode, 
						bool value) {
  DoubleProperty *borderWidth = 
    displayGraph->getProperty<DoubleProperty> (viewBorderWidth);
  ColorProperty *borderColor =
    displayGraph->getProperty<ColorProperty> (viewBorderColor);
  node hierarchyNode = openCloseHierarchy->getHierarchyNode (displayNode);
  if (value) {
    borderWidth->setNodeValue (displayNode, SELECTED_BORDER_WIDTH);
    borderColor->setNodeValue (displayNode, SELECTION_COLOUR);
  }//end if
  else {
    if (OpenCloseHierarchy::isLeaf (displayNode, hierarchyNode)) {
      borderWidth->setNodeValue (displayNode, LEAF_NODE_BORDER_WIDTH);
      borderColor->setNodeValue (displayNode, LEAF_NODE_BORDER_COLOUR);
    }
    else {
      borderWidth->setNodeValue (displayNode, META_NODE_BORDER_WIDTH);
      borderColor->setNodeValue (displayNode, META_NODE_BORDER_COLOUR);
    }
  }
  openCloseHierarchy->getHierarchyNode (displayNode);
  nodeSelection->setNodeValue (hierarchyNode, value);
}


//========================================================================
void GrouseSelectionProperty::setEdgeSelection (const edge &displayEdge, 
						bool value) {
  ColorProperty *color =
    displayGraph->getProperty<ColorProperty> (viewColor);
  if (value) 
    color->setEdgeValue (displayEdge, SELECTED_EDGE_COLOUR);
  else
    color->setEdgeValue (displayEdge, EDGE_COLOUR);
  edgeSelection->setEdgeValue (displayEdge, value);
}

//========================================================================
void GrouseSelectionProperty::setAugSelection (GlABoundingCircle *aug,
					       bool value) {
  aug->setSelected (value);
  nodeSelection->setNodeValue (aug->getHierarchyNode(), value);
}

//========================================================================
void GrouseSelectionProperty::invertModifySelection (const vector<node> &nodes,
						     vector<GlABoundingCircle *> &augs) {
  for (vector<node>::const_iterator nodeIt = nodes.begin();
       nodeIt != nodes.end(); ++nodeIt) {
    node hierarchyNode = openCloseHierarchy->getHierarchyNode (*nodeIt);
    setNodeSelection (*nodeIt, !nodeSelection->getNodeValue (hierarchyNode));
  }//end for
  for (vector<GlABoundingCircle *>::iterator augsIt = augs.begin();
       augsIt != augs.end(); ++augsIt) {
    node hierarchyNode = (*augsIt)->getHierarchyNode();
    setAugSelection (*augsIt, !nodeSelection->getNodeValue (hierarchyNode));
  }//end for
  displayWidget->draw();
}
//========================================================================
void GrouseSelectionProperty::setSelection (const vector<node> &nodes,
					    vector<GlABoundingCircle *> &augs) {
  clearSelectionNoDraw();
  for (vector<node>::const_iterator nodeIt = nodes.begin();
       nodeIt != nodes.end(); ++nodeIt)
    setNodeSelection (*nodeIt, true);
  for (vector<GlABoundingCircle *>::iterator augsIt = augs.begin();
       augsIt != augs.end(); ++augsIt)
    setAugSelection (*augsIt, true);
  displayWidget->draw();
}


//========================================================================
void GrouseSelectionProperty::setEdgeSelection (const vector<edge> &edges) {
  for (vector<edge>::const_iterator edgeIt = edges.begin();
       edgeIt != edges.end(); ++edgeIt)
    setEdgeSelection (*edgeIt, true);
  displayWidget->draw();
}//end setEdgeSelection

//========================================================================
void 
GrouseSelectionProperty::invertModifyEdgeSelection (const vector<edge> &edges) {
  for (vector<edge>::const_iterator edgeIt = edges.begin();
       edgeIt != edges.end(); ++edgeIt)
    setEdgeSelection (*edgeIt, !edgeSelection->getEdgeValue (*edgeIt));
  displayWidget->draw();
}

//========================================================================
void GrouseSelectionProperty::setMouseSelection (const vector<node> &nodes,
						 vector<GlABoundingCircle *> &augs,
						 const node &lastSelectedHierarchyNode) {
  setSelection (nodes, augs);
  node lastSelected; //send invalid last selection
  emit selectionChanged (*nodeSelection, lastSelectedHierarchyNode);
}//end setMouseSelection

//========================================================================
void GrouseSelectionProperty::invertModifyMouseSelection (const vector<node> &nodes,
							  vector<GlABoundingCircle *> &augs,
							  const node &lastSelectedHierarchyNode) {
  invertModifySelection (nodes, augs);
  emit selectionChanged (*nodeSelection, lastSelectedHierarchyNode);
}//end setMouseSelection

//========================================================================
BooleanProperty &GrouseSelectionProperty::getHierarchySelection () {
  return *nodeSelection;
}

//========================================================================
BooleanProperty &GrouseSelectionProperty::getVisibleSelection () {
  visibleSelection->setAllNodeValue (false);
  visibleSelection->setAllEdgeValue (false);
  node curDisplayNode;
  forEach (curDisplayNode, displayGraph->getNodes()) {
    node curHierarchyNode = 
      openCloseHierarchy->getHierarchyNode (curDisplayNode);
    if (nodeSelection->getNodeValue (curHierarchyNode))
      visibleSelection->setNodeValue (curHierarchyNode, true);
  }//end forEach
  edge curDisplayEdge;
  forEach (curDisplayEdge, displayGraph->getEdges())
    visibleSelection->setEdgeValue (curDisplayEdge,
				    edgeSelection->getEdgeValue (curDisplayEdge));
  return *visibleSelection;
}

//========================================================================
void GrouseSelectionProperty::modifySelection (const vector<node> &nodes,
					       vector<GlABoundingCircle *> 
					       &augs, bool setTrue) {
  for (vector<node>::const_iterator nodeIt = nodes.begin();
       nodeIt != nodes.end(); ++nodeIt)
    setNodeSelection (*nodeIt, setTrue);
  for (vector<GlABoundingCircle *>::iterator augsIt = augs.begin();
       augsIt != augs.end(); ++augsIt)
    setAugSelection (*augsIt, setTrue);
  displayWidget->draw();
}

//========================================================================
void 
GrouseSelectionProperty::modifyEdgeSelection (const vector<edge> &edges,
					      bool setTrue) {
  for (vector<edge>::const_iterator edgeIt = edges.begin();
       edgeIt != edges.end(); ++edgeIt)
    setEdgeSelection (*edgeIt, setTrue);
  displayWidget->draw();
}
