#include <string>
#include <tulip/DoubleProperty.h>
#include <tulip/IntegerProperty.h>
#include <tulip/ColorProperty.h>
#include <tulip/StringProperty.h>
#include <tulip/ConnectedTest.h>
#include <tulip/StableIterator.h>
#include <tulip/ForEach.h>
#include "ColourShapeDefaults.h"
#include "CleanGraph.h"
#include "HierarchyNodeCreation.h"
#include "HierarchyMap.h"
#include "InitializeHierarchyStructure.h"

using namespace tlp;
using namespace std;

const string finalLayoutLabel ("1__Final Layout");
const string hierarchyLabel ("1__Hierarchy");
const string hierarchySubGraphsLabel ("1__HierarchySubGraphs");
const string viewColor ("viewColor");

//================================================================
inline void clear (Graph *clearGraph) {
  StableIterator<node> delIt (clearGraph->getNodes());
  while (delIt.hasNext())
    clearGraph->delNode (delIt.next());
}//end clear

//================================================================
inline string numberToString (unsigned int number) {
  char charNumber[10];
  sprintf (charNumber, "%d", number);
  return charNumber;
} 

//================================================================
/**
 * The function finds the first subgraph with this name in the
 * hierarchy.
 */
Graph *findGraph (const string &graphName, Graph *rootGraph) {
  Graph *foundGraph = NULL, *curGraph;
  Iterator<Graph *> *subgraphs = rootGraph->getSubGraphs();
  while (subgraphs->hasNext()) {
    curGraph = subgraphs->next();
    string name = curGraph->getAttribute <string> ("name");
    if (name == graphName) {
      foundGraph = curGraph;
      break;
    }//end if
  } delete subgraphs; //end while
  return foundGraph;
}

//================================================================
/**
 * If the subgraph below the root is disconnected, the hierarchy
 * is fixed here and noted.
 */
bool initDisconnectedRoot (Graph *hierarchy, Graph *hierarchySubGraphs,
			   GraphProperty *hierarchyMetaGraphs) {
  //get the first level below the root graph
  Graph *belowRoot = 
    hierarchyMetaGraphs->getNodeValue (hierarchy->getOneNode());
  vector<node> createdHierarchyNodes;
  const string errMsg ("In initDisconnectedRoot");

  //if connected do nothing and indicate it is connected
  if (ConnectedTest::isConnected (belowRoot)) return false;

  //if a field of disconnected metanodes, there is nothing to do
  else if (belowRoot->numberOfEdges () == 0) return true;
  
  //otherwise, fix disconnected root by creating metanodes
  else processConnectedComponents (belowRoot, belowRoot, createdHierarchyNodes,
				   hierarchy, hierarchySubGraphs,
				   hierarchyMetaGraphs, errMsg);

  ColorProperty *hierarchyColors =
    hierarchy->getProperty<ColorProperty> (viewColor);

  for (unsigned int i = 0; i < createdHierarchyNodes.size(); ++i) {
    hierarchyColors->setNodeValue (createdHierarchyNodes[i],
				   CONN_COMP_COLOUR);
    Graph *curGraph = 
      hierarchyMetaGraphs->getNodeValue (createdHierarchyNodes[i]);
    curGraph->setAttribute<string> ("name", "Connected " + 
				    numberToString (i));
  }

  //clean up stray subgraphs
  StableIterator<Graph *> conSubGraphs (belowRoot->getSubGraphs());
  while (conSubGraphs.hasNext ()) 
    belowRoot->delAllSubGraphs (conSubGraphs.next ());

  return true;
}

//================================================================
void initHierarchy (Graph *&hierarchy, Graph *&hierarchySubGraphs,
		    node &hierarchyRootNode,
		    const string &hierarchyMetaGraphProperty,
		    Graph *rootGraph,
		    bool &isDisconnected) {
  hierarchy = findGraph (hierarchyLabel, rootGraph);
  if (hierarchy == NULL) {
    hierarchy = rootGraph->getRoot()->addSubGraph();
    hierarchy->setAttribute<string> ("name", hierarchyLabel);
  }//end if
  
  hierarchySubGraphs = findGraph (hierarchySubGraphsLabel, rootGraph);
  if (hierarchySubGraphs == NULL) {
    hierarchySubGraphs = rootGraph->getRoot()->addSubGraph();
    hierarchySubGraphs->setAttribute<string> ("name", hierarchySubGraphsLabel);
  }

  GraphProperty *hierarchyMetaGraphs = 
    hierarchy->getProperty<GraphProperty> (hierarchyMetaGraphProperty);

  //this is the case when there is a pre-existing topolayout hierarchy
  if (hierarchy->numberOfNodes() > 1) {
    Iterator<Graph *> *firstLevel = hierarchy->getSubGraphs();
    Graph *topoHierarchy = firstLevel->next();
    delete firstLevel;
    Graph *newTopoRoot = hierarchySubGraphs->addSubGraph();
    newTopoRoot->setAttribute<string> ("name", 
				       topoHierarchy->getAttribute<string> ("name"));
    copy (topoHierarchy, newTopoRoot);
    clear (hierarchy);
    hierarchyRootNode = hierarchy->addNode();
    hierarchyMetaGraphs->setNodeValue (hierarchyRootNode, newTopoRoot);
    hierarchy->delAllSubGraphs (topoHierarchy);
  }//end if

  //if there is a blank hierarchy or we have just created one.
  else if (hierarchy->numberOfNodes() == 0) {
    Graph *newLeafLevel = hierarchySubGraphs->addSubGraph();
    copy (rootGraph->getRoot(), newLeafLevel);
    cleanGraph (newLeafLevel);
    hierarchyRootNode = hierarchy->addNode();
    hierarchyMetaGraphs->setNodeValue (hierarchyRootNode, newLeafLevel);
    IntegerProperty *shapes = 
      newLeafLevel->getProperty<IntegerProperty> ("viewShape");
    shapes->setAllNodeValue (LEAF_NODE_SHAPE);
    ColorProperty *leafColors =
      newLeafLevel->getProperty<ColorProperty> (viewColor);
    leafColors->setAllNodeValue (LEAF_COLOUR);
    StringProperty *origEdgeID =
      newLeafLevel->getRoot ()->getLocalProperty<StringProperty> (ORIGINAL_EDGE_NAME);
    edge curEdge;
    forEach (curEdge, newLeafLevel->getEdges ())
      origEdgeID->setEdgeValue (curEdge, numberToString (curEdge.id));
    
  }//end else if

  ColorProperty *hierarchyColors =
    hierarchy->getProperty<ColorProperty> (viewColor);
  hierarchyColors->setNodeValue (hierarchyRootNode, ROOT_COLOUR);
  isDisconnected = initDisconnectedRoot (hierarchy, hierarchySubGraphs,
					 hierarchyMetaGraphs);
}
//================================================================
/**
 * Set the rendering window to the final layout
 */
void initFinalLayout (Graph *&finalLayout, Graph *rootGraph) {
  finalLayout = findGraph (finalLayoutLabel, rootGraph);
  if (finalLayout == NULL) {
    finalLayout = rootGraph->getRoot()->addSubGraph();
    finalLayout->setAttribute<string> ("name", finalLayoutLabel);
    Graph *leafLevel = finalLayout->addSubGraph();
    copy (rootGraph->getRoot(), leafLevel);
    cleanGraph (finalLayout);
  }//end if
}//end initFinalLayout
