#include <qinputdialog.h>
#include <tulip/GraphProperty.h>
#include <tulip/StringProperty.h>
#include <tulip/ColorProperty.h>
#include <tulip/ForEach.h>
#include <tulip/GlGraph.h>
#include <tulip/Graph.h>
#include <tulip/GrouseTableWidget.h>
#include "ColouredListViewItem.h"
#include "MetaGraphHierarchyWidget.h"

using namespace std;
using namespace tlp;

const string viewMetaGraph ("viewMetaGraph");
const string viewLabel ("viewLabel");
const QColor defaultColour (255, 255, 255);

#define LABELS_COLUMN (0)
#define NODES_COLUMN (1)
#define EDGES_COLUMN (2)
#define NUM_COLUMNS (3)

//=======================================================
class MetaGraphListViewItem : public ColouredListViewItem {
private:
  tlp::node hierNode;
  MetaGraphHierarchyWidget *metaGraphHierarchyWidget;
public:
  MetaGraphListViewItem (QListView *parent, 
			 const tlp::node &hierNode,
			 MetaGraphHierarchyWidget *
			 metaGraphHierarchyWidget) : 
    ColouredListViewItem(parent), hierNode (hierNode),
    metaGraphHierarchyWidget(metaGraphHierarchyWidget) {}

  MetaGraphListViewItem (QListViewItem *item,
			 const tlp::node &hierNode,
			 MetaGraphHierarchyWidget *
			 metaGraphHierarchyWidget) : 
    ColouredListViewItem(item), hierNode(hierNode),
    metaGraphHierarchyWidget(metaGraphHierarchyWidget) {}

  /**
   * A function to return the hierarchy node associated with this
   * coloured list view item.
   */
  tlp::node getHierarchyNode () { return hierNode; }

  /**
   * A function to compart two list view items.  Used for sorting
   * via various tulip properties.
   * \param otherItem The item this QListViewItem is being compared to.
   * \param col the column being sorted by.
   * \param ascending if we are ascending or descening in order.
   */
  int compare (QListViewItem *otherItem, int col, bool ascending) const {
    if (col >= metaGraphHierarchyWidget->numberOfColumns ()) {
      cerr << "Column " << col << " does not exist." << endl;
      cerr << "Sorting by 0." << endl;
      col = 0;
    }//end if
    QString type = metaGraphHierarchyWidget->getListViewPropertyType (col);
    QString thisString = this->text (col);
    QString otherString = otherItem->text (col);

    //Special sort using node number if col is zero
    if (col == 0) {
      int nodesColumn = metaGraphHierarchyWidget->getNodesColumn ();
      int thisNumNodes = this->text (nodesColumn).toInt();
      int otherNumNodes = otherItem->text (nodesColumn).toInt();
      if (!ascending) {
	if (thisNumNodes < otherNumNodes) return -1;
	else if (thisNumNodes > otherNumNodes) return 1;
	else return QString::compare (this->text (0), otherItem->text (0));
      }//end if
      else {
	if (thisNumNodes > otherNumNodes) return -1;
	else if (thisNumNodes < otherNumNodes) return 1;
	else return QString::compare (this->text (0), otherItem->text (0));
      }//end if
    }//end if

    if (type == "int") {
      if (thisString.toInt () == otherString.toInt ()) return 0;
      if (ascending) {
	if (thisString.toInt () < otherString.toInt ()) return -1;
	else return 1;
      }//end if
      else {
	if (thisString.toInt () > otherString.toInt ()) return -1;
	else return 1;
      }//end else
    }//end if
    else if (type == "double") {
      if (thisString.toDouble () == otherString.toDouble ()) return 0;
      if (ascending) {
	if (thisString.toDouble () < otherString.toDouble ()) return -1;
	else return 1;
      }//end if
      else {
	if (thisString.toDouble () > otherString.toDouble ()) return -1;
	else return 1;
      }//end else
    }//end if
    else if (type == "float") {
      if (thisString.toFloat () == otherString.toFloat ()) return 0;
      if (ascending) {
	if (thisString.toFloat () < otherString.toFloat ()) return -1;
	else return 1;
      }//end if
      else {
	if (thisString.toFloat () > otherString.toFloat ()) return -1;
	else return 1;
      }//end else
    }//end if
    else {
      if (ascending)
	return QString::compare (this->text (col), otherItem->text (col));
      else 
	return !QString::compare (this->text (col), otherItem->text (col));
    }//end else
  }//end compare
};

//=======================================================================
MetaGraphHierarchyWidget::
MetaGraphHierarchyWidget (QWidget *parent,
			  const char *name, WFlags flags, 
			  tlp::Graph *hierarchy,
			  const string &metagraphProperty,
			  tlp::Graph *inputGraph) :
  MetaGraphHierarchyWidgetData (parent, name, flags),
  hierarchy(hierarchy), 
  inputGraph(inputGraph),
  metagraphProperty(metagraphProperty),
  disableSelectionSynch (false)
{
  oldNodeSelection.setAll (false);
  hierNodeToListViewItem.setAll (NULL);
  grouseNodeProperties
    ->resetProperties (hierarchy,
		       hierarchy->getProperty<GraphProperty> (metagraphProperty));
  populate();
}

//=======================================================================
MetaGraphHierarchyWidget::
MetaGraphHierarchyWidget (QWidget* parent, const char *name, Qt::WFlags fl) :
  MetaGraphHierarchyWidgetData (parent, name, fl),
  hierarchy(NULL),
  inputGraph(NULL),
  metagraphProperty(viewMetaGraph),
  disableSelectionSynch (false)
{
  oldNodeSelection.setAll (false);
  hierNodeToListViewItem.setAll (NULL);
}
  
//=======================================================================
QString MetaGraphHierarchyWidget::getListViewPropertyType (int col) {
  return listViewPropertyTypes[col];
}

//=======================================================================
int MetaGraphHierarchyWidget::numberOfColumns () {
  return metaGraphHierarchyListView->columns ();
}

//=======================================================================
int MetaGraphHierarchyWidget::getNodesColumn () {
  return NODES_COLUMN;
}

//=======================================================================
int MetaGraphHierarchyWidget::getEdgesColumn () {
  return EDGES_COLUMN;
}

//=======================================================================
MetaGraphHierarchyWidget::~MetaGraphHierarchyWidget () {
  metaGraphHierarchyListView->clear();
}

//=======================================================================
node MetaGraphHierarchyWidget::getHierarchyNode (QListViewItem *listViewItem) {
  return ((MetaGraphListViewItem *)listViewItem)->getHierarchyNode();
}//end getHierarchyNode

//=======================================================================
string MetaGraphHierarchyWidget::getLabel (const tlp::node &curNode) {
  return hierNodeToListViewItem.get (curNode.id)
    ->text (LABELS_COLUMN).latin1 ();
}//end getLabel

//=======================================================================
Graph *MetaGraphHierarchyWidget::getHierarchyGraph (QListViewItem 
						    *listViewItem) {
  GraphProperty *metaGraphHierarchy = 
    hierarchy->getProperty<GraphProperty> (metagraphProperty);
  node queryNode = getHierarchyNode (listViewItem);
  if (queryNode.isValid())
    return metaGraphHierarchy->getNodeValue (queryNode);
  else if (listViewItem->parent() == NULL)
    return hierarchy;
  else
    return NULL;
}

//=======================================================================
inline void deleteColumns (QListView *metaGraphHierarchyListView) {
  int numCols = metaGraphHierarchyListView->columns ();
  for (int i = 0; i < numCols; ++i)
    metaGraphHierarchyListView->removeColumn (0);
}

//=======================================================================
void MetaGraphHierarchyWidget::
setHierarchyProperties (const string &newMetaProperty,
			Graph *newRootGraph, 
			Graph *newInputGraph) {
  metagraphProperty = newMetaProperty;
  hierarchy = newRootGraph;
  inputGraph = newInputGraph;
  grouseNodeProperties->resetProperties (hierarchy,
					 hierarchy->getProperty<GraphProperty> (metagraphProperty));
  populate();
}

//=======================================================================
void 
MetaGraphHierarchyWidget::setMetaGraphProperty (const string &newProperty) {
  metagraphProperty = newProperty;
  grouseNodeProperties->resetProperties (hierarchy,
					 hierarchy->getProperty<GraphProperty> (metagraphProperty));
  populate();
}

//=======================================================================
void 
MetaGraphHierarchyWidget::setHierarchyGraph (Graph *rootGraph) {
  hierarchy = rootGraph;
  grouseNodeProperties->resetProperties (hierarchy,
					 hierarchy->getProperty<GraphProperty> (metagraphProperty));
  populate();
}

//=======================================================================
void 
MetaGraphHierarchyWidget::setInputGraph (Graph *newInputGraph) {
  inputGraph = newInputGraph;
  grouseNodeProperties->resetProperties (hierarchy,
					 hierarchy->getProperty<GraphProperty> (metagraphProperty));
  populate();
}//end setInputGraph

//=======================================================================
void MetaGraphHierarchyWidget::setSelectedLabel (const node &hierarchyNode) {
  QListViewItem *selectedItem = 
    hierNodeToListViewItem.get (hierarchyNode.id);
  if (selectedItem != NULL) {
    ((MetaGraphListViewItem *)selectedItem)->setSelected (true);
    metaGraphHierarchyListView->repaintItem (selectedItem);
  }//end if
}

//=======================================================================
void 
MetaGraphHierarchyWidget::synchListViewColour (const node &hierarchyNode) {
  ColorProperty *hierarchyColours =
    hierarchy->getProperty<ColorProperty> ("viewColor");
  Color nodeColour = hierarchyColours->getNodeValue (hierarchyNode);
  ColouredListViewItem *clistViewItem =
    ((ColouredListViewItem *)hierNodeToListViewItem.get (hierarchyNode.id));
  QColor qboundColor (nodeColour.getR(), nodeColour.getG(), 
		      nodeColour.getB());
  clistViewItem->setBackground(0, qboundColor);
  metaGraphHierarchyListView->repaintItem (clistViewItem);
}

//=======================================================================
void MetaGraphHierarchyWidget::genListViewProperties () {
  listViewProperties.clear();
  listViewPropertyTypes.clear();
  QString nullString ("");
  Iterator<string> * propIt = hierarchy->getProperties();
  while (propIt->hasNext()) {
    string curPropName = propIt->next();
    PropertyInterface *curProperty = hierarchy->getProperty (curPropName);
    string propertyTypename = PropertyInterface::getTypename (curProperty);
    if ((curPropName.find ("view") == string::npos) && 
	((propertyTypename == "string") || (propertyTypename == "double") ||
	 (propertyTypename == "boolean") || (propertyTypename == "int"))) {
      listViewProperties.push_back (curProperty);
      listViewPropertyTypes.push_back (propertyTypename);
    }//end if
  } delete propIt; //end while
  metaGraphHierarchyListView->addColumn ("Labels", 1);
  metaGraphHierarchyListView->addColumn ("Nodes", 1);
  listViewPropertyTypes.push_back ("int");
  metaGraphHierarchyListView->addColumn ("Edges", 1);
  listViewPropertyTypes.push_back ("int");
}//end getListViewProperties

//=======================================================================
void MetaGraphHierarchyWidget::initListViewWidths () {
  double totalWidth = metaGraphHierarchyListView->width();
  double noColumns = metaGraphHierarchyListView->columns();
  int widthPerColumn = static_cast<int> (floor (totalWidth/noColumns));
  for (unsigned int i = 0; i < noColumns; ++i)
    metaGraphHierarchyListView->setColumnWidth (i, widthPerColumn);
}//end initListViewWidths

//=======================================================================
inline string intToString (int integer) {
  char intToChar [20];
  sprintf (intToChar, "%d", integer);
  return string (intToChar);
}

//=======================================================================
void MetaGraphHierarchyWidget::populateListViewProperties (node curNode,
							   Graph *curGraph,
							   QListViewItem *
							   newItem) {
  if (curGraph != NULL) {
    newItem->setText (NODES_COLUMN, 
		      QString(intToString (curGraph->numberOfNodes())));
    newItem->setText (EDGES_COLUMN, 
		      QString(intToString (curGraph->numberOfEdges())));
  }//end  if
  else {
    string label = "";
    newItem->setText (NODES_COLUMN, QString (intToString (0)));
    newItem->setText (EDGES_COLUMN, QString (intToString (0)));
    for (int i = 0; i < grouseNodeProperties->numCols (); ++i) {
      SelectionTableItem *curSelection =
	static_cast<SelectionTableItem *> (grouseNodeProperties->item (0, i));
      if (curSelection->isChecked ()) {
	int tableRow = grouseNodeProperties->getRow (curNode);
	label += string (" ") + 
	  grouseNodeProperties->item (tableRow, i)->text ().latin1();
      }//end if
    }//end for 
    newItem->setText (LABELS_COLUMN, QString (label));
  }
}

//=======================================================================
void MetaGraphHierarchyWidget::populate () {
  metaGraphHierarchyListView->clear();
  deleteColumns (metaGraphHierarchyListView);
  hierNodeToListViewItem.setAll (NULL);
  genListViewProperties();
  initListViewWidths();
  GraphProperty *viewMetaGraph = 
    hierarchy->getProperty<GraphProperty> (metagraphProperty);
  node rootNode = hierarchy->getOneNode();
  MetaGraphListViewItem *root = 
    new MetaGraphListViewItem (metaGraphHierarchyListView, rootNode,
			       this);

  hierNodeToListViewItem.set (rootNode.id, root);
  root->setOpen (true);
  Graph *firstHierarchyGraph = viewMetaGraph->getNodeValue (rootNode);
  root->setText (0, QString ("Everything"));
  root->setText (NODES_COLUMN, 
		 QString (intToString (firstHierarchyGraph->numberOfNodes())));
  root->setText (EDGES_COLUMN, 
		 QString (intToString (firstHierarchyGraph->numberOfEdges())));
  root->setBackground(0, defaultColour);
  root->setRenameEnabled (0, true);
  populate (firstHierarchyGraph, root);
  metaGraphHierarchyListView->setSorting (0, TRUE);
  emit labelDataChanged ();
}

//=======================================================================
void MetaGraphHierarchyWidget::populate (Graph *curGraph,
					 QListViewItem *inputItem) {
  GraphProperty *viewMetaGraph = 
    curGraph->getProperty<GraphProperty> (metagraphProperty);
  node curNode;
  forEach (curNode, curGraph->getNodes()) {
    Graph *metaGraph = viewMetaGraph->getNodeValue (curNode);
    MetaGraphListViewItem *newItem = 
      new MetaGraphListViewItem (inputItem, curNode, this);
    newItem->setRenameEnabled (0, true);
    hierNodeToListViewItem.set (curNode.id, newItem);
    //if this is a meta node, title it the same as its graph
    if (metaGraph != NULL) {
      newItem->setOpen (true);
      populateListViewProperties (curNode, metaGraph, newItem);
      string name = metaGraph->getAttribute<string> ("name");
      newItem->setText (0, QString (name));
      newItem->setBackground(0, defaultColour);
      populate (metaGraph, newItem);
    }
    //otherwise a leaf of the hierarchy (original node of graph)
    else {
      //if we check for dummy nodes, make sure this is not one
      if ((inputGraph == NULL) ||
	  ((inputGraph != NULL) && inputGraph->isElement (curNode))) {
	newItem->setOpen (true);
	populateListViewProperties (curNode, NULL, newItem);
	newItem->setBackground(0, defaultColour);
      }//end if

      //if it is a dummy node, it should not be in the hierarchy
      else {
	hierNodeToListViewItem.set (curNode.id, NULL);
	delete newItem;
      }
    }
  }
}

//=======================================================================
void MetaGraphHierarchyWidget::clearSelection () {
  metaGraphHierarchyListView->clearSelection();
  metaGraphHierarchyListView->update();
}

//=======================================================================
void MetaGraphHierarchyWidget::selectAll () {
  metaGraphHierarchyListView->clearSelection();
  metaGraphHierarchyListView->invertSelection();
  metaGraphHierarchyListView->update();
}

//=======================================================================
void MetaGraphHierarchyWidget::
selectionChanged (BooleanProperty &newSelection,
		  const node &lastSelected) {
  disableSelectionSynch = true;
  metaGraphHierarchyListView->clearSelection();
  Iterator<node> *newNodeSelection = newSelection.getNodesEqualTo (true);
  while (newNodeSelection->hasNext()) {
    node selectedNode = newNodeSelection->next();
    QListViewItem *selItem = hierNodeToListViewItem.get (selectedNode.id);
    selItem->setSelected (true);
    if (lastSelected == selectedNode)
      metaGraphHierarchyListView->ensureItemVisible (selItem);
  } delete newNodeSelection; //end while
  metaGraphHierarchyListView->update();
  disableSelectionSynch = false;
}//end selectionChanged

//=======================================================================
void MetaGraphHierarchyWidget::selectionUpdate (vector<node> &nodeSelection,
						vector<node> &removed,
						vector<node> &added) {
  QListViewItemIterator listViewIterator (metaGraphHierarchyListView);
  while (listViewIterator.current() != NULL) {
    QListViewItem *curItem = listViewIterator.current();
    node hierarchyNode = getHierarchyNode (curItem);
    if (curItem->isSelected())
      nodeSelection.push_back (hierarchyNode);
    if (curItem->isSelected() && !oldNodeSelection.get (hierarchyNode.id))
      added.push_back (hierarchyNode);
    if (!curItem->isSelected() && oldNodeSelection.get (hierarchyNode.id))
      removed.push_back (hierarchyNode);
    oldNodeSelection.set (hierarchyNode.id, curItem->isSelected());
    ++listViewIterator;
  }//end while
}//end selectionUpdate

//=======================================================================
void MetaGraphHierarchyWidget::newHierarchyPath (QListViewItem *pathOrigin,
						 vector<node> &path) {
  QListViewItem *curItem = pathOrigin;
  while (curItem != NULL) {
    path.push_back (getHierarchyNode (curItem));
    curItem = curItem->parent();
  }//end while
  metaGraphHierarchyListView->clearSelection();
  metaGraphHierarchyListView->update();
}

//=======================================================================
void MetaGraphHierarchyWidget::
metaGraphHierarchyListView_mouseButtonClicked(int button, 
					      QListViewItem *itemClicked, 
					      const QPoint &pointClicked, 
					      int column) {
  if ((button == Qt::RightButton) && (itemClicked != NULL)) {
    vector<node> path;
    newHierarchyPath (itemClicked, path);
    emit newPath (path);
  }
}

//=======================================================================
void MetaGraphHierarchyWidget::
metaGraphHierarchyListView_doubleClicked(QListViewItem *newItem, 
					 const QPoint &pos, int col) {
  if ((col == -1) || (newItem == NULL)) return;
  if (newItem->renameEnabled (col)) newItem->startRename (col);
}

//=======================================================================
void MetaGraphHierarchyWidget::
metaGraphHierarchyListView_itemRenamed(QListViewItem *renameItem, 
				       int col, const QString &text) {
  GraphProperty *subgraphs = 
    hierarchy->getProperty<GraphProperty> (metagraphProperty);
  node curNode = ((MetaGraphListViewItem *) renameItem)->getHierarchyNode ();
  Graph *curHierMetaGraph = subgraphs->getNodeValue (curNode);
  if (curHierMetaGraph != NULL)
    curHierMetaGraph->setAttribute<string> ("name", text);
  emit labelDataChanged ();
}//end itemRenamed

//=======================================================================
void MetaGraphHierarchyWidget::
metaGraphHierarchyListView_selectionChanged() {
  vector<node> nodeSelection;
  vector<node> removed;
  vector<node> added;
  selectionUpdate (nodeSelection, removed, added);
  if (!disableSelectionSynch) 
    emit selectionChanged (nodeSelection, removed, added);
}//end selectionChanged

//=======================================================================
void MetaGraphHierarchyWidget::nullChildren (QListViewItem *curItem) {
  MetaGraphListViewItem *curDel =
    static_cast<MetaGraphListViewItem *> (curItem);
  QListViewItem *delChild = curDel->firstChild();
  while (delChild != NULL) {
    nullChildren (delChild);
    delChild = delChild->nextSibling();
  }//end while
  hierNodeToListViewItem.set (curDel->getHierarchyNode ().id,
			      NULL);
}//end nullChildren

//=======================================================================
void MetaGraphHierarchyWidget::updateTree (const node &curHierarchyNode) {
  MetaGraphListViewItem *item =
    ((MetaGraphListViewItem *)hierNodeToListViewItem.get (curHierarchyNode.id));
  QListViewItem *delChild = item->firstChild();
  while(delChild != NULL) {
    QListViewItem *nextItem = delChild->nextSibling();
    nullChildren (delChild);
    delete delChild;
    delChild = nextItem;
  }//end delete children
  
  GraphProperty *viewMetaGraph = 
    hierarchy->getProperty<GraphProperty> (metagraphProperty);
  Graph *curHierarchyNodeGraph =
    viewMetaGraph->getNodeValue (curHierarchyNode);
  populate (curHierarchyNodeGraph, item);
  metaGraphHierarchyListView->update();
  emit labelDataChanged ();
}//end updateTree

//=======================================================================
void MetaGraphHierarchyWidget::grouseNodeProperties_valueChanged (int row, int col) {
  if (row == 0) this->populate ();
}//end grouseNodeProperties_valueChanged

//=======================================================================
void MetaGraphHierarchyWidget::grouseNodeProperties_propertyReorderSignal (int fromIndex, int toIndex) {
  this->populate ();
}//end grouseNodeProperties_propertyReorderSignal
