
#include <string>
#include <vector>
#include <qtable.h>
#include <qlistview.h>
#include <tulip/PropertyWidget.h>
#include <tulip/Graph.h>
#include <tulip/StringProperty.h>
#include <tulip/DoubleProperty.h>
#include <tulip/BooleanProperty.h>
#include <tulip/MutableContainer.h>
#include "MetaGraphHierarchyWidgetData.h"

#ifndef __META_GRAPH_HIERARCHY_CLASS__
#define __META_GRAPH_HIERARCHY_CLASS__

class MetaGraphHierarchyWidget : public MetaGraphHierarchyWidgetData {
  Q_OBJECT;
 public:
  MetaGraphHierarchyWidget (QWidget *parent,
			    const char *name, WFlags flags, 
			    tlp::Graph *hierarchy,
			    const std::string &metagraphProperty,
			    tlp::Graph *inputGraph = NULL);
  MetaGraphHierarchyWidget (QWidget* parent = 0, const char *name = 0, 
			    Qt::WFlags fl = 0);

  ~MetaGraphHierarchyWidget ();

  /**
   * Returns the label in the widget given the passed node.
   * \param curNode The node we want to know the label for
   */
  std::string getLabel (const tlp::node &curNode); 

  /**
   * Returns the node associated with this listViewItem
   * \param listViewItem MetaGraphHierarchyListView item which we
   * want to retrieve the corresponding node.
   */
  tlp::node getHierarchyNode (QListViewItem *listViewItem);

  /**
   * Returns graph in the hierarchy associated with this listViewItem
   * \param listViewItem MetaGraphHierarchyListView item which we
   * want to retrieve the corresponding graph.
   */
  tlp::Graph *getHierarchyGraph (QListViewItem *listViewItem);

  /**
   * Sets the metagraph, label, and root of the hierarchy for this
   * graph widget.
   * \param newMetaProperty The name of the metagraph property.
   * \param newRootGraph The root graph of the widget.
   * \param newInputGraph The input graph for filtering out dummy nodes.
   * Default is no dummy node filtering or newInputGraph == NULL
   */
  void setHierarchyProperties (const std::string &newMetaProperty,
			       tlp::Graph *newRootGraph,
			       tlp::Graph *newInputGraph = NULL);

  /**
   * Sets the metaGraph property for which this widget represents
   * \param newProperty The new metaGraph property this widget represents.
   */
  void setMetaGraphProperty (const std::string &newProperty);

  /**
   * Sets the root graph which the widget will use to populate its entries.
   * \param rootGraph The new root graph of the metagraph hierarchy
   * this widget represents.
   */
  void setHierarchyGraph (tlp::Graph *rootGraph);

  /**
   * Sets the input graph used for filtering out dummy nodes in the hierarchy.
   * \param newInputGraph The new input graph for the widget
   */
  void setInputGraph (tlp::Graph *newInputGraph);

  /**
   * Given a node in the graph hierarchy.  This node sets the
   * label in the MetaGraphHierarchy to selected.
   * \param hierarchyNode The node we want to select.
   */
  void setSelectedLabel (const tlp::node &hierarchyNode);

  /**
   * Synchronizes the colour of the list view item with the colour of
   * the hierarchy node.
   * \param hierarchyNode The node for which we want to change label colour
   */
  void synchListViewColour (const tlp::node &hierarchyNode);

  /**
   * Function to populate the MetaGraphHierarchyWidget.  It will
   * use the hierarchy graph and proxies alread associated with the
   * widget to do its work.
   */
  void populate ();

  /**
   * Unselects all elements of the listview.
   */
  void clearSelection ();

  /**
   * Selects all elements of the listview.
   */
  void selectAll ();
 
  /**
   * Updates the hierarchy which has changed below the hierarchy
   * node curHierarchyNode.
   * \param curHierarchyNode - the parent of all hierarchy nodes.
   */
  void updateTree (const tlp::node &curHierarchyNode);

  /**
   * Returns the type of the column numbered col.
   *\param col The column which we would like to know the type.
   */
  QString getListViewPropertyType (int col);

  /**
   * Returns the number of columns in the meta graph hierarchy listview.
   */
  int numberOfColumns ();

  /**
   * Returns the nodes column.
   */
  int getNodesColumn ();

  /**
   * Returns the edges column.
   */
  int getEdgesColumn ();
  
 signals:
  /**
   * Signal emitted when selection in the widget has changed.
   * \param newNode The new node in the meta graph hierarchy.
   * \param removed The nodes removed from the selection,
   * \param added The nodes added to the selection
   */
  void selectionChanged (const std::vector<tlp::node> &newNodes,
			 const std::vector<tlp::node> &removed,
			 const std::vector<tlp::node> &added);

  /**
   * Signal emitted when a new right button path through the hierarchy
   * has been specified.
   * \param newPath The new path through the hierarchy.
   */
  void newPath (const std::vector<tlp::node> &nodePath);

  /**
   * Signal emitted when the label data in the graph has changed.
   * \param newPath The new path through the hierarchy.
   */
  void labelDataChanged ();

  public slots:
  /**
   * Slot to update the list view with the new passed selection.
   * \param newSelection the new selection items.
   * \param lastSelected the node which was last selected.
   */
    void selectionChanged (tlp::BooleanProperty &newSelection,
			   const tlp::node &lastSelected);

 private:
  tlp::Graph *hierarchy;
  tlp::Graph *inputGraph;
  std::string metagraphProperty;
  tlp::MutableContainer<QListViewItem *> hierNodeToListViewItem;
  std::vector<tlp::PropertyInterface *> listViewProperties;
  std::vector<QString> listViewPropertyTypes;
  bool disableSelectionSynch;
  tlp::MutableContainer<bool> oldNodeSelection;
  std::vector<int> labelCols;

  /**
   * Sets all children and curDel's hierNodeToListViewItem to null.
   * \param hierNodeToListViewItem Hierarchy list view item.
   */
  void nullChildren (QListViewItem *curDel);

  /**
   * Function to populate the MetaGraphHierarchyWidget. The private
   * method considers a graph in the hierarchy and its associated item.
   */
  void populate (tlp::Graph *curGraph, QListViewItem *inputItem);

  /**
   * Function to handle when the selection has been updated.
   * \param nodeSelection the set of hierarchy nodes in the selection
   * \param removed Nodes removed from the selected set
   * \param added Nodes added from the selected set
   */
  void selectionUpdate (std::vector<tlp::node> &nodeSelection,
			std::vector<tlp::node> &removed,
			std::vector<tlp::node> &added);

  /**
   * Function to handle when a new hierarchy path is requested
   * \param pathOrigin The item where we start to trace back up the path
   * \param path The returned path by the function.
   */
  void newHierarchyPath (QListViewItem *pathOrigin, 
			 std::vector<tlp::node> &path);

  /**
   * Signal handler for whenever a mouse button is clicked on an element of
   * the listview.
   * \param the QT button which was clicked.
   * \param newItem The new item which is selected.
   * \param coord The new point.
   * \param col the column of the listview
   */
  void metaGraphHierarchyListView_mouseButtonClicked(int button,
						     QListViewItem *newItem, 
						     const QPoint &coord, 
						     int col);
  
  /**
   * Signal handler for when someone double clicks on an item.
   * \param newItem The item which was double clicked.
   */
  
  void metaGraphHierarchyListView_doubleClicked(QListViewItem *newItem, 
						const QPoint &pos, int col);

  /**
   * Signal handler for when an item is renamed
   * \param renameItem The item which was double renamed.
   */
  void metaGraphHierarchyListView_itemRenamed(QListViewItem *renameItem, 
					      int col, const QString &text);

  /**
   * Signal handler for when the widget selection has changed.
   */
  void metaGraphHierarchyListView_selectionChanged();

  /**
   * Signal handler for when a value in the properties table has 
   * changed.
   */
  void grouseNodeProperties_valueChanged (int row, int col);

  /**
   * Signal handler for when the columns of the property table
   * have been reordered.
   */
  void grouseNodeProperties_propertyReorderSignal (int fromIndex, int toIndex);

  /**
   * Generates a list of displayable properties.  These properties are
   * either string or numeric and are not "view" properties.
   */
  void genListViewProperties ();
  
  /**
   * Initializes the widths of the columns in the widget so that all
   * are visible.
   */
  void initListViewWidths ();

  /**
   * Populates the metagraph hierarchy widget with listView properties
   * for a particular node.
   * \param curNode The node for which we would add those properties.
   * \param curGraph The metagraph below curNode, if any.
   * \param listViewItem The listView item which represents the node
   * inside the hierarchy list view widget.
   */
  void populateListViewProperties (tlp::node curNode, tlp::Graph *curGraph,
				   QListViewItem *listViewItem);
 
};
#endif
