/**
 * This file implements metanode highlighting using GlGraphWidgets.
 * For each metanode, it looks at the boundary circle and renders it as
 * a gl widget.
 */
#include <cmath>
#include <iostream>
#include <tulip/ForEach.h>
#include <tulip/GraphProperty.h>
#include <tulip/DoubleProperty.h>
#include "HierarchyMap.h"
#include "HSVConversion.h"
#include "MetaNodeBoundingCircle.h"
using namespace std;
using namespace tlp;

//base depth for bouding objects with initialization
double GlABoundingCircle::baseDepth = -1.0;

//================================================================
//Inline to compute geometric depth from hierarchy depths.
#define DEPTH_MULTIPLY (0.01)
double MetaNodeBoundingCircles::computeSceneDepth (unsigned int levelDepth) {
  return DEPTH_MULTIPLY*(static_cast<double> (levelDepth) - 
			 static_cast<double> (maxDepth));
}

//================================================================
#define SATURATION_BASE (0.3)
#define VALUE_KICK (0.1)
Color MetaNodeBoundingCircles::
computeColourSaturation (const tlp::node &hierarchyNode,
			 unsigned int levelDepth) {
  ColorProperty *hierarchyColours = 
    hierarchy->getProperty<ColorProperty> ("viewColor");
  Color baseColour = hierarchyColours->getNodeValue (hierarchyNode);
  HSV circleHSV = rgb_to_hsv (baseColour);
  circleHSV.s = circleHSV.s*(SATURATION_BASE + 
			     (1.0 - SATURATION_BASE)*
			     (((double)levelDepth)/((double)maxDepth)));
  return hsv_to_rgb (circleHSV);
}//end computeColourSaturation

//================================================================
/**
 * Given a bounding circle size object, it draws the rendering widget
 * in the tulip window.
 */
GlABoundingCircle *
MetaNodeBoundingCircles::addBoundingCircle (Graph *displayGraph,
					    const node &hierarchyNode,
					    const Coord &circleCentre, 
					    const Size &circleSize, 
					    unsigned int depthFromRoot) {

  if (hierNodeToCircle.find (hierarchyNode) != hierNodeToCircle.end()) {
    cerr << "Subgraph already has a bounding circle " << endl;
    return hierNodeToCircle[hierarchyNode];
  }
  char charID[10];
  sprintf (charID, "%d", hierarchyNode.id);
  string boundCircleID (charID);

  //recheck the baseDepth
  LayoutProperty *viewLayout = 
    displayGraph->getProperty<LayoutProperty> ("viewLayout");
  edge curEdge;
  forEach (curEdge, displayGraph->getEdges()) {
    const vector<Coord> &bends = viewLayout->getEdgeValue (curEdge);
    for (unsigned int i = 0; i < bends.size(); ++i) {
      if (bends[i].getZ() < GlABoundingCircle::getBaseDepth ())
	GlABoundingCircle::setBaseDepth (bends[i].getZ());
    }//end for
  }//end forEach
  
  if (depthFromRoot + 1 > maxDepth) maxDepth = depthFromRoot + 1;
  
  double depth = computeSceneDepth (depthFromRoot);
  Color circleColour = computeColourSaturation (hierarchyNode, depthFromRoot);
  GlABoundingCircle *newCircle = 
    new GlABoundingCircle (hierarchyNode, circleCentre, circleSize, 
			   circleColour, depth);
  hierNodeToCircle[hierarchyNode] = newCircle;
  showBoundingCircle (hierarchyNode);
  return newCircle;
}
//================================================================
void 
MetaNodeBoundingCircles::updateBoundingCircle (const tlp::node &hierarchyNode,
					       const Coord &newPos,
					       const Size &newSize,
					       unsigned int depthFromRoot) {
  GlABoundingCircle *curCircle = getBoundingCircle (hierarchyNode);
  if (curCircle == NULL) return;
  if (depthFromRoot + 1 > maxDepth) maxDepth = depthFromRoot + 1;
  double depth = computeSceneDepth (depthFromRoot);
  Color circleColour = computeColourSaturation (hierarchyNode, depthFromRoot);
  curCircle->update (newPos, newSize, circleColour, depth);
}//end updateBoundingCircle

//================================================================
void GlABoundingCircle::computePoints () {
  points.clear();
  double alpha = 0;
  const double delta = 2.0*M_PI/30.0;
  const double rad = sqrt (size.getW()*size.getW()/4.0 + 
			   size.getH()*size.getH()/4.0);
  Coord curPoint;
  for(int i = 0; i <= 30; ++i) {
    curPoint.setX (rad*cos(alpha) + pos.getX());
    curPoint.setY (rad*sin(alpha) + pos.getY());
    curPoint.setZ (depth + baseDepth);
    points.push_back (curPoint);
    alpha += delta;
  }//end for
}//end computePoints

//================================================================
 void GlABoundingCircle::draw (GlGraph *graphRenderer) {
   setup();
   glColor4fv(color.getGL());
   glBegin(GL_TRIANGLE_FAN);
   glVertex3d (pos.getX(), pos.getY(), depth);
   for(unsigned int i=0; i < points.size(); ++i) 
     glVertex3d(points[i].getX(), points[i].getY(), depth);
   glEnd();
    if (!selected) {
      glColor4fv (META_NODE_BORDER_COLOUR.getGL());
      glLineWidth (3.0);
    }
    else {
      glColor4fv (SELECTION_COLOUR.getGL());
      glLineWidth (8.0);
    }
    glBegin(GL_LINES);
    for(unsigned int i=0; i < points.size() - 1; ++i) {
      glVertex3d(points[i].getX(), points[i].getY(), depth);
      glVertex3d(points[i + 1].getX(), points[i + 1].getY(), 
		 depth);
    }//end for
    glEnd();
    unsetup();
 }//end draw
