/**
 * This file implements metanode highlighting using GlGraphWidgets.
 * For each metanode, it looks at the boundary circle and renders it as
 * a gl widget.
 */

#ifndef __META_NODE_BOUNDING_CIRCLE__
#define __META_NODE_BOUNDING_CIRCLE__
#include <cstdio>
#include <vector>
#include <string>
#include <tulip/GlAugmentedDisplay.h>
#include <tulip/GlGraphWidget.h>
#include <tulip/GlGraph.h>
#include <tulip/ForEach.h>
#include <tulip/Graph.h>
#include <tulip/SizeProperty.h>
#include <tulip/LayoutProperty.h>
#include <tulip/ColorProperty.h>
#include "ColourShapeDefaults.h"

class GlABoundingCircle : public tlp::GlAugmentedDisplay {
 private:
  tlp::node hierarchyNode;
  tlp::Coord pos;
  tlp::Size size;
  tlp::Color color;
  double depth;
  static double baseDepth;
  bool selected;
  std::vector<Coord> points;

 public:

  void computePoints (); 

  GlABoundingCircle (tlp::node hierarchyNode,
		     Coord pos, Size size, Color color, double depth) :
    hierarchyNode(hierarchyNode),
    pos(pos), size(size), color(color), 
    depth(depth), selected (false) {
    computePoints();
  }

  inline void update (const tlp::Coord &newPos,
		      const tlp::Size &newSize,
		      const tlp::Color &newColor,
		      double newDepth) {
    pos = newPos;
    size = newSize;
    color = newColor;
    depth = newDepth;
    computePoints();
  }

  const std::vector<Coord> &getPoints () {
    return points;
  }

  void setPoints (const tlp::Coord &newPos,
		  const std::vector<Coord> &newPoints) {
    pos = newPos;
    points = newPoints;
  }

 void draw (GlGraph *graphRenderer);
 
  inline bool getSelected () { return selected; }
  inline double getDepth () { return depth; }
  inline static double getBaseDepth () { return baseDepth; } 
  inline const Color &getColor () { return color; }
  inline const Coord &getPos () { return pos; }
  inline const Size &getSize () { return size; }
  inline double getRadius () { return sqrt (size.getW()*size.getW()/4.0 + 
					    size.getH()*size.getH()/4.0); }
  inline const tlp::node &getHierarchyNode () { return hierarchyNode; }
  inline static void setBaseDepth (double newBaseDepth) { 
    baseDepth = newBaseDepth;
  } 
  inline void setSelected (bool newSel) { selected = newSel; }
};

class MetaNodeBoundingCircles {
 private:
  GlGraphWidget *renderingWidget;
  tlp::Graph *hierarchy;
  stdext::hash_map<tlp::node, GlABoundingCircle*> hierNodeToCircle;
  unsigned int maxDepth;

  /**
   * Function to return the string id of a bounding object given its
   * hierarchy node
   */
  inline std::string getID (const tlp::node &hierNode) {
    char charID[10];
    std::sprintf (charID, "%d", hierNode.id);
    return charID;
  }

  /**
   * Computes geometric depth from the level depth.
   */
  double computeSceneDepth (unsigned int levelDepth);

  /**
   * Attenuates colour according to the level depth.
   */
  tlp::Color computeColourSaturation (const tlp::node &hierarchyNode,
				      unsigned int levelDepth);
  
 public:
  MetaNodeBoundingCircles (GlGraphWidget *renderingWidget, 
			   tlp::Graph *hierarchy, unsigned int maxDepth) :
    renderingWidget(renderingWidget), 
    hierarchy(hierarchy), 
    maxDepth(maxDepth) {
  }

  inline ~MetaNodeBoundingCircles () {
    stdext::hash_map<tlp::node, GlABoundingCircle*>::iterator nodeCircleIt =
      hierNodeToCircle.begin();
    for (; nodeCircleIt != hierNodeToCircle.end(); ++nodeCircleIt)
      delete nodeCircleIt->second;
  }//end destructor

  /**
   * Adds a new bounding circle to the set of bounding cirles and
   * returns a pointer to the new circle.
   */
  GlABoundingCircle *addBoundingCircle (tlp::Graph *displayGraph,
					const tlp::node &hierarchyNode,
					const tlp::Coord &circleCentre, 
					const tlp::Size &circleSize, 
					unsigned int depthFromRoot);

  /**
   * Updates attributes about the bounding circle associated with
   * the passed hierarchy node.
   */
  void updateBoundingCircle (const tlp::node &hierarchyNode,
			     const Coord &newPos,
			     const Size &newSize,
			     unsigned int depthFromRoot);
  
  /**
   * Given a hierarchy node, this function returns the bounding circle
   */
  inline GlABoundingCircle *getBoundingCircle (const tlp::node &hierNode) {
    if (hierNodeToCircle.find(hierNode) == hierNodeToCircle.end())
      return NULL;
    else
      return hierNodeToCircle[hierNode];
  }

  /**
   * Given a hierarchy node, this function deletes its bounding circle
   */
  inline void delBoundingCircle (const tlp::node &hierNode) {
    GlABoundingCircle *delCircle = getBoundingCircle (hierNode);
    if (delCircle != NULL) {
      hierNodeToCircle.erase (hierNode);
      delete delCircle;
    }//end if
  }//end delBoundingCircle

  /**
   * Given a hierarchy node, this function hides the bounding circle
   */
  inline void hideBoundingCircle (const tlp::node hierNode) {
    if (hierNodeToCircle.find(hierNode) == hierNodeToCircle.end()) {
      cerr << "cannot hide circle of hierarchy node " << hierNode.id << endl 
	   << "It is not present." << endl;
      return;
    }
    std::string stringID = getID (hierNode);
    renderingWidget->removeGlAugmentedDisplay (stringID);
  }//end remove

  /**
   * Given a hierarchy node, this function shows the bounding circle
   */
  inline void showBoundingCircle (const tlp::node &hierNode) {
    if (hierNodeToCircle.find(hierNode) == hierNodeToCircle.end()) {
      cerr << "cannot show circle of hierarchy node " << hierNode.id << endl 
	   << "It is not created." << endl;
      return;
    } 
    
    std::string stringID = getID (hierNode);
    renderingWidget
      ->addGlAugmentedDisplay (hierNodeToCircle.find(hierNode)->second, 
			       stringID);
  }
  /**
   * Given a hierarchy node, this function returns true if the
   * bouding circle is hidden.
   */
  inline bool isHidden (const tlp::node &hierNode) {
    if (hierNodeToCircle.find(hierNode) == hierNodeToCircle.end()) {
      cerr << "cannot determine if hidden " << hierNode.id << endl 
	   << "It is not created." << endl;
      return false;
    } 
    
    std::string stringID = getID (hierNode);
    return renderingWidget->findGlAugmentedDisplay (stringID) == NULL;
  }//end isHidden

};
#endif
