#include <qcursor.h>
#include <tulip/BooleanProperty.h>
#include "MouseGrouseBaseNavigate.h"
#include "MouseGrouseGeneralOperations.h"

using namespace tlp;
using namespace std;

//===============================================================
//Defines for buttons to be used on each mouse operation
#define QT_FREE_PAN_BUTTON (Qt::LeftButton)
#define QT_FREE_PAN_KEY_DOWN (Qt::NoButton)
#define QT_SELECT_BUTTON (Qt::LeftButton)
#define QT_SELECT_KEY_DOWN (Qt::NoButton)
#define QT_FREE_PAN_DRAG_STATE (QT_FREE_PAN_BUTTON)

//===============================================================
bool MouseGrouseBaseNavigate::inDragState = false;

//===============================================================
/*
 * Default constructors and destructors
 */
MouseGrouseBaseNavigate::MouseGrouseBaseNavigate (OpenCloseHierarchy 
						  *openCloseHierarchy,
						  GrouseSelectionProperty
						  *grouseSelectionProperty) :
  openCloseHierarchy (openCloseHierarchy),
  grouseSelectionProperty (grouseSelectionProperty) {}
MouseGrouseBaseNavigate::~MouseGrouseBaseNavigate() {}

//===============================================================
/**
 * Function to pan the mouse.
 */
void MouseGrouseBaseNavigate::pan (const QPoint &end, GlGraphWidget *graph) {
  if (end.x() != startDrag.x()) 
    graph->translateCamera(end.x() - startDrag.x(), 0, 0);
  if (end.y() != startDrag.y()) 
    graph->translateCamera(0, startDrag.y() - end.y(), 0);
  initStartDrag (end);
}

//===============================================================
/**
 * Function to select a internal node of the hierarchy
 */
void MouseGrouseBaseNavigate::select (const QPoint &mouseClick, 
				  GlGraphWidget *graphWidget) {
  node selectedHierarchy;
  node pickedNode;
  edge pickedEdge;
  vector<node> nodeSelections;
  vector<GlABoundingCircle *> circleSelections;
  GlAugmentedDisplay *pickedAug = NULL;
  pick (mouseClick.x(), mouseClick.y(), graphWidget, true, true, true, 
	false, true, pickedNode, pickedEdge, pickedAug);

  if (pickedNode.isValid()) {
    selectedHierarchy = openCloseHierarchy->getHierarchyNode (pickedNode);
    nodeSelections.push_back (pickedNode);
  }
  else if (pickedAug != NULL) {
    GlABoundingCircle *boundingCircle =
      static_cast<GlABoundingCircle *> (pickedAug);
    selectedHierarchy = boundingCircle->getHierarchyNode();
    circleSelections.push_back (boundingCircle);
  }
  
  if (!nodeSelections.empty() || !circleSelections.empty()) {
    grouseSelectionProperty
      ->invertModifyMouseSelection (nodeSelections, circleSelections,
				    selectedHierarchy);
    graphWidget->draw();
  }//end selection changed
}

//===============================================================
/**
 * Callback to handle mouse down events for grouse.  It also
 * listens to the keybord to see if any keys are down as well.
 * The graph widget is graph, and the Qt mouse event is e.
 */
bool MouseGrouseBaseNavigate::mPressEvent(GlGraphWidget *graph, QMouseEvent *e) {
  if ((e->button() == QT_FREE_PAN_BUTTON) &&
      (e->state() == QT_FREE_PAN_KEY_DOWN))
    initStartDrag (e->pos());
  else {
    e->ignore();
    return false;
  }
  return true;
}

//===============================================================
/**
 * Callback to handle mouse up events for grouse.  It also
 * listens to the keybord to see if any keys are down as well.
 * The graph widget is graph, and the Qt mouse event is e.
 */
bool MouseGrouseBaseNavigate::mReleaseEvent(GlGraphWidget *graph, QMouseEvent *e) {
  if ((e->button() == QT_SELECT_BUTTON) &&
      (e->state() == QT_SELECT_BUTTON) && !inDragState)
    select (e->pos(), graph);
  else {
    e->ignore();
    inDragState = false;
    return false;
  }
  inDragState = false;
  return true; 
}

//===============================================================
/**
 * Callback to handle when the mouse is moving.  It will look also
 * to see if any keys or mouse buttons are pressed.  The graph widget
 * is graph, and the mouse event is e.
 */
bool MouseGrouseBaseNavigate::mMoveEvent(GlGraphWidget *graph, QMouseEvent *e) {
  if (e->state() == QT_FREE_PAN_DRAG_STATE) {
    pan (e->pos(), graph);
    graph->draw();
    inDragState = true;
  }
  else {
    e->ignore();
    return false;
  }
  return true;
}

//===============================================================
/**
 * Callback to handle when a key on the keybord is pressed.  The
 * graph widget is g and the QKeyEvent is e.
 */
bool MouseGrouseBaseNavigate::keyPressEvent(GlGraphWidget *graph, QKeyEvent *e) {
  int delta = (e->isAutoRepeat() ? 3 : 1);
  e->accept();
  switch(e->key()) {
    case Qt::Key_Right: graph->translateCamera(delta * 2,0,0); break;
    case Qt::Key_Left: graph->translateCamera(-1 * delta * 2,0,0); break;
    case Qt::Key_Down: graph->translateCamera(0,-1 * delta * 2,0); break;
    case Qt::Key_Up: graph->translateCamera(0,delta * 2,0); break;
    case Qt::Key_Prior: graph->zoom(delta); break;
    case Qt::Key_Next: graph->zoom(-1 * delta); break;
    case Qt::Key_Home: graph->translateCamera(0,0,-1 * delta * 2); break;
    case Qt::Key_End: graph->translateCamera(0,0,delta * 2); break;
    case Qt::Key_Insert: graph->rotateScene(0,0,-1 * delta * 2); break;
    case Qt::Key_Delete : graph->rotateScene(0,0,delta * 2); break;
    default: 
      e->ignore();
      return false;
  }
  graph->draw();
  return true;
}

//===============================================================
/**
 * Callback to handle when the mouse wheel is used.  The graph
 * widget is g and the wheel event is e.
 */
// according to Qt's doc, this constant has been defined by wheel mouse vendors
// we need it to interpret the value of QWheelEvent->delta()
#define WHEEL_DELTA (120)
bool MouseGrouseBaseNavigate::wheelEvent(GlGraphWidget *graph, QWheelEvent *e) {
  if (e->orientation() == Qt::Vertical) {
    graph->zoomXY(e->delta() / WHEEL_DELTA, e->x(), e->y());
    graph->draw();
  }
  else {
    e->ignore();
    return false;
  }
  return true;
}

//===============================================================
/**
 * Event filter for all the mouse events on the grouse view window.
 * Determines which private navigation function to call.
 */
bool MouseGrouseBaseNavigate::eventFilter (QObject *widget, QEvent *e) {
  GlGraphWidget *glGraphWidget = static_cast<GlGraphWidget *> (widget);
  if (e->type() == QEvent::MouseButtonPress) 
    return mPressEvent(glGraphWidget, static_cast<QMouseEvent *> (e));
  else if (e->type() == QEvent::MouseButtonRelease)
    return mReleaseEvent (glGraphWidget, static_cast<QMouseEvent *> (e));
  else if (e->type() == QEvent::MouseMove)
    return mMoveEvent (glGraphWidget, static_cast<QMouseEvent *> (e));
  else if (e->type() == QEvent::KeyPress)
    return keyPressEvent (glGraphWidget, static_cast<QKeyEvent *> (e));
  else if (e->type() == QEvent::Wheel)
    return wheelEvent (glGraphWidget, static_cast<QWheelEvent *> (e));
  else return false;
  return true;
}
