
#if (__GNUC__ < 3)
#  include <hash_map>
#else
#  include <ext/hash_map>
#endif
#include <tulip/ForEach.h>
#include <tulip/TreeTest.h>
#include <tulip/SortIterator.h>
#include <tulip/DoubleProperty.h>
#include <tulip/BooleanProperty.h>
#include <tulip/ColorProperty.h>
#include <tulip/MutableContainer.h>
#include <tulip/GlGraph.h>
#include "ColourShapeDefaults.h"
#include "HierarchyNodeCreation.h"
#include "NodeDegEdgeContraction.h"

using namespace std;
using namespace tlp;

#define IN_PATTERN_MATCH (string ("In Pattern Match "))
#define OUT_PATTERN_MATCH (string ("Out of Pattern Match "))
#define CATEGORY_MATCH (string ("Category"))
#define DEFAULT_LABEL (string ("Contracted "))
#define CONTRACTION_COLOUR (GROUSE_BROWN_2)

//========================================================================
inline string getContractionNumber (unsigned int &number) {
  char charAlgoNumber[10];
  sprintf (charAlgoNumber, "%d", number);
  number++;
  return charAlgoNumber;
} 

//========================================================================
NodeDegEdgeContraction::NodeDegEdgeContraction (Graph *hierarchy,
						Graph *hierarchySubGraphs,
						const string &metaGraphPropertyName,
						PluginProgress *pluginProgress) :
hierarchy (hierarchy), 
hierarchySubGraphs (hierarchySubGraphs),
pluginProgress (pluginProgress),
numContractionComponents (0)
{
  hierarchyMetaGraphs = 
    hierarchy->getProperty<GraphProperty> (metaGraphPropertyName);
}
//========================================================================
int NodeDegEdgeContraction::getLeavesBelow (const node &curNode) {
  unsigned int numLeaves = 0;
  Graph *curSubGraph = hierarchyMetaGraphs->getNodeValue (curNode);
  if (curSubGraph != NULL) {
    node nextNode;
    forEach (nextNode, curSubGraph->getNodes ()) {
      Graph *nextSubGraph = hierarchyMetaGraphs->getNodeValue (nextNode);
      if (nextSubGraph == NULL) ++numLeaves;
      else numLeaves += getLeavesBelow (nextNode);
    }//end forEach
  }//end if
  return numLeaves;
}//end getLeavesBelow

//========================================================================
void NodeDegEdgeContraction::snipTreesAtNonTreeRoots (Graph *contractionGraph,
						      Graph *removalGraph,
						      unsigned int maxNodeNum,
						      IntegerProperty 
						      *nodeComponent) {
  MutableContainer<node> nodePutInside;
  Graph *treesGraph = removalGraph->getRoot ()->addSubGraph ();
  copy (removalGraph, treesGraph);
  string errMsg;
  DoubleProperty trees (treesGraph);
  DataSet localData;
  localData.set ("Non-Tree Roots", true);
  bool success = 
    treesGraph->computeProperty ("Connected_Tree Component 2",
				 &trees, errMsg, NULL, &localData);
  if (!success) {
    cerr << "Connected_Tree Component 2 returned the following error:" << endl
	 << errMsg << endl << "(Is the plugin installed?)" << endl;
    return;
  }//end if

  BooleanProperty *nonTreeRootSelection =
    treesGraph->getProperty<BooleanProperty> ("non Tree Root");

  DataSet equalValueData;
  equalValueData.set("Property", &trees);
  success = tlp::applyAlgorithm (treesGraph, errMsg, &equalValueData, 
				 "Equal Value");
  if (!success) {
    cerr << "Equal Value clustering returned the following error:\n"
	 << errMsg << "\n(Is the plugin properly installed?)" << endl;
    return;
  }//end if

  StableIterator <Graph *> clusters (treesGraph->getSubGraphs());
  while (clusters.hasNext() && 
	 (removalGraph->numberOfNodes () > maxNodeNum)) {
    Graph *curCluster = clusters.next();
    if (TreeTest::isTree (curCluster)) {
      node root;
      node findRoot;
      forEach (findRoot, curCluster->getNodes ()) {
	if (nonTreeRootSelection->getNodeValue (findRoot)) {
	  root = findRoot;
	  breakForEach;
	}//end if
      }//end forEach
      
      //the id of a tree component is its nontree root.
      StableIterator<node> treeNodeIt (curCluster->getNodes ());
      while (treeNodeIt.hasNext ()) {
	node curNode = treeNodeIt.next ();
	if (curNode != root) {
	  nodePutInside.set (curNode.id, root);
	  removalGraph->delNode (curNode);
	}//end if
      }//end while
    }//end if
  }//end while

  //fix the component ids after removal of nodes on this pass
  node fixNode;
  forEach (fixNode, contractionGraph->getNodes ()) {
    int repNode = nodeComponent->getNodeValue (fixNode);
    node changeRepNode = nodePutInside.get (repNode);
    if (changeRepNode.isValid ())
      nodeComponent->setNodeValue (fixNode, changeRepNode.id);
  }//end forEach
  removalGraph->getRoot ()->delAllSubGraphs (treesGraph);
}//end snipTreesAtNonTreeRoots

//========================================================================
void NodeDegEdgeContraction::contractEdgesMetric (Graph *contractionGraph,
						  unsigned int maxNodeNum,
						  IntegerProperty *nodeComponent) {
  DoubleProperty nodeWeight (contractionGraph);
  node curNode;
  const unsigned int total = contractionGraph->numberOfNodes () - maxNodeNum;
  const double logTotal = log (total);
  forEach (curNode, contractionGraph->getNodes ()) {
    nodeWeight.setNodeValue (curNode, getLeavesBelow (curNode) + 1);
    nodeComponent->setNodeValue (curNode, curNode.id);
  }//end forEach
  
  Graph *removalGraph = contractionGraph->getRoot ()->addSubGraph ();
  removalGraph->setAttribute<string> ("name", "removalGraph");
  copy (contractionGraph, removalGraph);
  vector<edge> contractionEdges;
  while ((removalGraph->numberOfNodes () > maxNodeNum) && 
	 !TreeTest::isFreeTree (removalGraph)) {
    snipTreesAtNonTreeRoots (contractionGraph, removalGraph, 
			     maxNodeNum, nodeComponent);

    SortNodeIterator nodeSort (removalGraph->getNodes (), &nodeWeight);
    BooleanProperty marked (removalGraph);
    marked.setAllNodeValue (false);
    
    while (nodeSort.hasNext () && 
	   (removalGraph->numberOfNodes () - contractionEdges.size () 
	    > maxNodeNum)) {
      node curRemovalNode = nodeSort.next ();
      if (!marked.getNodeValue (curRemovalNode)) {
	node minOpposite;
	double minWeight = DBL_MAX;
	node opposite;
	forEach (opposite, removalGraph->getInOutNodes (curRemovalNode)) {
	  if (marked.getNodeValue (opposite)) continue;
	  if (nodeWeight.getNodeValue (opposite) < minWeight) {
	    minOpposite = opposite;
	    minWeight = nodeWeight.getNodeValue (opposite);
	  }//end if
	}//end forEach
	if (minOpposite.isValid ()) {
	  marked.setNodeValue (minOpposite, true);
	  marked.setNodeValue (curRemovalNode, true);
	  edge contractionEdge = 
	    removalGraph->existEdge (curRemovalNode, minOpposite);
	  if (!contractionEdge.isValid ())
	    contractionEdge = 
	      removalGraph->existEdge (minOpposite, curRemovalNode);
	  contractionEdges.push_back (contractionEdge);
	}//end if
      }//end if
    }//end while
    
    //contract the edges in the removal graph and connect as needed.
    MutableContainer<node> nodePutInside;
    for (unsigned int i = 0; i < contractionEdges.size (); ++i) {
      node source = removalGraph->source (contractionEdges[i]);
      node target = removalGraph->target (contractionEdges[i]);
      node minNode = (source.id < target.id) ? source : target;
      node delNode = (minNode == source) ? target : source;
      nodePutInside.set (delNode.id, minNode);
      nodeWeight.setNodeValue (minNode, 
			       nodeWeight.getNodeValue (minNode) + 
			       nodeWeight.getNodeValue (delNode));
      StableIterator<node> oppositeIt (removalGraph->getInOutNodes (delNode));
      while (oppositeIt.hasNext()) {
	node opposite = oppositeIt.next ();
	if (opposite == minNode) continue;
	if (!removalGraph->existEdge (minNode, opposite).isValid () &&
	    !removalGraph->existEdge (opposite, minNode).isValid ()) {
	  removalGraph->addEdge (minNode, opposite);
	}//end if
      }//end forEach
      removalGraph->delNode (delNode);
    }//end for
    
    //fix the component ids after removal of nodes on this pass
    node fixNode;
    forEach (fixNode, contractionGraph->getNodes ()) {
      int repNode = nodeComponent->getNodeValue (fixNode);
      node changeRepNode = nodePutInside.get (repNode);
      if (changeRepNode.isValid ())
	nodeComponent->setNodeValue (fixNode, changeRepNode.id);
    }//end forEach
    contractionEdges.clear ();
    unsigned int current = removalGraph->numberOfNodes () - maxNodeNum;
    double logCurrent = log (current);
    pluginProgress->progress (static_cast<int> (floor (logTotal)) - 
			      static_cast<int> (floor (logCurrent)), 
			      static_cast<int> (floor (logTotal)));
  }//end while
  contractionGraph->getRoot ()->delSubGraph (removalGraph);
  pluginProgress->progress (static_cast<int> (floor (logTotal)), 
			    static_cast<int> (floor (logTotal)));
}//end contractEdges

//========================================================================
void NodeDegEdgeContraction::setLabel (Graph *curGraph) {
  node curHierNode;
  stdext::hash_map <string, unsigned int> done;
  forEach (curHierNode, curGraph->getNodes ()) {
    Graph *curHierGraph = hierarchyMetaGraphs->getNodeValue (curHierNode);
    if (curHierGraph == NULL) continue;
    string curLabel = curHierGraph->getAttribute<string> ("name");
    unsigned int labelLen = curLabel.length ();
    string labelAddition;
    if (curLabel.find (IN_PATTERN_MATCH) != string::npos) {
      unsigned int pos = curLabel.find (CATEGORY_MATCH) +
	IN_PATTERN_MATCH.length ();
      labelAddition = "In " + curLabel.substr (pos, labelLen - pos);
    }//end if
    else if (curLabel.find (OUT_PATTERN_MATCH) != string::npos) {
      unsigned int pos = curLabel.find (CATEGORY_MATCH) +
	OUT_PATTERN_MATCH.length ();
      labelAddition = "Out "  + curLabel.substr (pos, labelLen - pos);
    }//end else if
    else if (curLabel.find (CATEGORY_MATCH) != string::npos) {
      unsigned int pos = curLabel.find (CATEGORY_MATCH) + 
	CATEGORY_MATCH.length ();
      labelAddition = curLabel.substr (pos, labelLen - pos);
    }//end else if
    if (done.find (labelAddition) == done.end ())
      done[labelAddition] = 1;
    else
      done[labelAddition]++;
  }//end forEach
  stdext::hash_map <string, unsigned int>::iterator doneIt = done.begin();
  string curGraphLabel = "";
  for (; doneIt != done.end(); ++doneIt)
    curGraphLabel += doneIt->first + ", ";

  curGraphLabel = curGraphLabel.substr (0, curGraphLabel.length() - 2); 
  curGraph->setAttribute<string> ("name", curGraphLabel);
}//end setLabel

//========================================================================
bool NodeDegEdgeContraction::makeOneMoreLevel (Graph *hierLevel, 
					       IntegerProperty *components) {
  Graph *cloneHierarchy = hierLevel->getRoot()->addSubGraph();
  copy (hierLevel, cloneHierarchy);
  string errMsg;
  DataSet equalValueData;
  equalValueData.set("Property", components);
  bool success = tlp::applyAlgorithm (cloneHierarchy, errMsg, &equalValueData, 
				      "Equal Value");
  if (!success) {
    cerr << "Equal Value clustering returned the following error:\n"
	 << errMsg << "\n(Is the plugin properly installed?)" << endl;
    return false;
  }//end if

  //Go through and fix subgraphs
  ColorProperty *viewColor = 
    hierLevel->getProperty<ColorProperty> ("viewColor");
  StableIterator <Graph *>  contractionGraphs (cloneHierarchy->getSubGraphs());
  while (contractionGraphs.hasNext()) {
    Graph *curSubGraph = contractionGraphs.next ();
    if (curSubGraph->numberOfNodes () > 1) {
      Graph *newHierarchyGraph = hierarchySubGraphs->addSubGraph();
      copy (curSubGraph, newHierarchyGraph);
      setLabel (newHierarchyGraph);
      if (newHierarchyGraph->getAttribute<string> ("name").length () == 0)
	newHierarchyGraph
	  ->setAttribute<string> ("name", DEFAULT_LABEL + 
				  getContractionNumber (numContractionComponents));
      node newSubGraphNode = hierLevel->addNode();
      viewColor->setNodeValue (newSubGraphNode, CONTRACTION_COLOUR);
      hierarchyMetaGraphs->setNodeValue (newSubGraphNode, newHierarchyGraph);
      setNewConnections (hierLevel, newHierarchyGraph, newSubGraphNode);
      delFeature (newHierarchyGraph, hierLevel);
    }//end if
  }//end while

  hierLevel->getRoot()->delAllSubGraphs (cloneHierarchy);
  return true;
}//end makeOneMoreLevel

//========================================================================
void NodeDegEdgeContraction::contractHierarchy (const node &hierNodeBelow,
						unsigned int maxNodeNum) {
  Graph *hierLevel = hierarchyMetaGraphs->getNodeValue (hierNodeBelow);
  IntegerProperty components (hierLevel);
  contractEdgesMetric (hierLevel, maxNodeNum, &components);
  makeOneMoreLevel (hierLevel, &components);
}//end contractHierarchy
