#ifndef __NODEDEG_EDGE_CONTRACTION__
#define __NODEDEG_EDGE_CONTRACTION__

#include <string>
#include <tulip/IntegerProperty.h>
#include <tulip/GraphProperty.h>
#include <tulip/PluginProgress.h>
#include <tulip/QtProgressWidget.h>

class NodeDegEdgeContraction {
 public:
  /**
   * Only constructor which takes the name of the property storing the hierarchy.
   */
  NodeDegEdgeContraction (tlp::Graph *hierarchy,
			  tlp::Graph *hierarchySubGraphs,
			  const std::string &metaGraphPropertyName,
			  tlp::PluginProgress *pluginProgress);

  inline ~NodeDegEdgeContraction () {}

  /**
   * Performs edge contraction of the graph below hierNodeBelow until less than maxNodeNum 
   * elements are in the graph.
   * \param hierNodeBelow The graph to coarsen is below this node.
   * \param maxNodeNum The maximum number of node elements.
   */
  void contractHierarchy (const tlp::node &hierNodeBelow, unsigned int maxNodeNum);

 private:
  //stores the hierarchy
  tlp::Graph *hierarchy;
  tlp::Graph *hierarchySubGraphs;
  tlp::GraphProperty *hierarchyMetaGraphs;
  tlp::PluginProgress *pluginProgress;
  unsigned int numContractionComponents;
  
  /**
   * Get the number of leaves below a node in the hierarchy.
   * \param curNode The node we want to count the number of leaves below.
   */
  int getLeavesBelow (const tlp::node &curNode);

  /**
   * The following function tries to set an intellegent label based on
   * the nodes under it.
   * \param curGraph The graph we are trying to label.
   */
  void setLabel (tlp::Graph *curGraph);

  /**
   * Function to prune trees away from graph before edge contraction.
   * \param contractionGraph The original graph passed into contractEdgesMetric.
   * \param removalGraph The graph where pruning takes place
   * \param maxNodeNum The maximum number of nodes.
   * \param nodeComponent The component numbers of the nodes.
   */
  void snipTreesAtNonTreeRoots (tlp::Graph *contractionGraph,
				tlp::Graph *removalGraph,
				unsigned int maxNodeNum,
				tlp::IntegerProperty *nodeComponent);
  
  /**
   * The following function labels nodes into groups of nodes based on
   * the number of leaves below each of them.  Contraction continues until
   * there are less than maxNodeNum nodes in the graph.
   * \param contractionGraph The graph to contract the nodes.
   * \param maxNodeNum The maximum number of nodes in the graph allowed after
   * the contraction process.
   * \param nodeComponent The components after contraction of nodes.
   */
  void contractEdgesMetric (tlp::Graph *contractionGraph,
			    unsigned int maxNodeNum,
			    tlp::IntegerProperty *nodeComponent);
  /**
   * The following method actually transforms the hierarchy based on
   * the groupings provided in components.  Returns true if successful.
   * \param hierLevel The level of the hierarchy where contraction is happening.
   * \param components The groups of components proveded by contractEdgeMetric
   */
  bool makeOneMoreLevel (tlp::Graph *hierLevel, 
			 tlp::IntegerProperty *components);
};

#endif
