#include <algorithm>
#include <tulip/ForEach.h>
#include <tulip/ColorProperty.h>
#include <tulip/StableIterator.h>
#include <tulip/DoubleProperty.h>
#include <tulip/TreeTest.h>
#include "OpenCloseHierarchy.h"

using namespace std;
using namespace tlp;
const Coord noPos (0, 0, 0);
const Size noSize (1, 1, 1);
const node invalidNode (NO_PARENT);
#define DEFAULT_MAX_NODE_NUM (200)

/**
 * Loads all nodes in the src subgraph into the dest subgraph
 */
inline void copy (tlp::Graph *src, tlp::Graph *dest) {
  tlp::node curNode;
  forEach (curNode, src->getNodes()) dest->addNode (curNode);
  tlp::edge curEdge;
  forEach (curEdge, src->getEdges()) dest->addEdge (curEdge);
}

//===============================================================
inline void startTime (clock_t &start, time_t &startLongTime) {
  time (&startLongTime);
  start = clock ();
}//end startPullOutTime

//===============================================================
inline void endTime (clock_t &start, time_t &startLongTime) {
  clock_t end = clock();
  time_t endLongTime;
  time (&endLongTime);
  double diff, diffLong;
  diff = ((double) (end - start) / (double) CLOCKS_PER_SEC);
  diffLong = difftime (endLongTime, startLongTime);
  cout << "-------------------------------------------------------" << endl;
  cout << "Action took:" << endl;
  cout << "In short time " << diff << endl;
  cout << "In long time " <<  diffLong << endl; 
  cout << "-------------------------------------------------------" << endl;
}//end endPullOutTime

//================================================================
/**
 * The function initializes the rendering widgets to their initial
 * startup state after a file has been loaded
 */
OpenCloseHierarchy::OpenCloseHierarchy (GlGraph *glGraph,
					Graph *hierarchy,
					const node &hierarchyRoot,
					Graph *hierarchySubGraphs,
					Graph *newDisplayGraph,
					Graph *inputGraph,
					MetaNodeBoundingCircles 
					*boundingCircles,
					MetaGraphHierarchyWidget 
					*metaGraphHierarchyWidget,
					PluginProgress *pluginProgress,
					bool scaleMode, 
					bool useDefaultAlgorithms) :
  hierarchyChange (false), 
  glGraph (glGraph),
  hierarchy (hierarchy), hierarchyRoot (hierarchyRoot),
  hierarchySubGraphs (hierarchySubGraphs),
  inputGraph (inputGraph),
  metaGraphHierarchyWidget (metaGraphHierarchyWidget),
  pluginProgress (pluginProgress),
  boundingCircles (boundingCircles), scaleMode (scaleMode),
  useDefaultAlgorithms (useDefaultAlgorithms),
  hasEdgeBundling (false),
  maxNodeNum (DEFAULT_MAX_NODE_NUM),
  onLogSize (false)
{
  if (newDisplayGraph == NULL) {
    displayGraph = inputGraph->getRoot()->addSubGraph();
    hierNodeToDisplayMetaNode = 
      hierarchy->getRoot ()
      ->getLocalProperty<IntegerProperty> (hierToDisplayProperty);
    displayMetaNodeToHierNode =
      displayGraph->getRoot ()
      ->getLocalProperty<IntegerProperty> (displayToHierProperty);
    hierNodeToDisplayMetaNode->setAllNodeValue (invalidNode.id);
    displayMetaNodeToHierNode->setAllNodeValue (invalidNode.id);
    copy (inputGraph, displayGraph);
  }
  else {
    displayGraph = newDisplayGraph;
    hierNodeToDisplayMetaNode = 
      hierarchy->getProperty<IntegerProperty> (hierToDisplayProperty);
    displayMetaNodeToHierNode =
      displayGraph->getProperty<IntegerProperty> (displayToHierProperty);
  }//end else
  displayGraphRenderParams = new DisplayGraphRendering (displayGraph,
							hierarchy,
							inputGraph);
  edgeBundling = new EdgeBundling (displayGraph, this);
  edgeContraction = new NodeDegEdgeContraction (hierarchy, hierarchySubGraphs,
					       "viewMetaGraph",
						pluginProgress);
  containedByNode.setAll (node (invalidNode));

  //layout for the hierarchy
  displayLayout = 
    displayGraph->getLocalProperty<LayoutProperty> ("viewLayout");
  displaySizes = 
    displayGraph->getLocalProperty<SizeProperty> ("viewSize");
  hierarchyMetaGraphs = 
    hierarchy->getProperty<GraphProperty> ("viewMetaGraph");
  
  //add an node to represent the top level of the hierarchy and initialize
  //the hierarchy
  initializeOpenCloseHierarchy (hierarchyRoot);
  openClose = new OpenClose (displayGraph);
  closeAll (hierarchyRoot);
  LayoutProperty *hierarchyLayout =
    hierarchy->getProperty<LayoutProperty> ("viewLayout");
  hierarchyLayout->setNodeValue (hierarchyRoot, noPos);
  vector<GlABoundingCircle *> boundCircles;
  cleanupDisplay (hierarchyRoot, boundCircles);
}//end OpenCloseHierarchy

//================================================================
void OpenCloseHierarchy::deleteDataStructures (Graph *curHierarchy) {
  EdgeOrderTable *curTable = 
    curHierarchy->getAttribute<EdgeOrderTable *> (EDGE_ORDER_ATTRIBUTE);
  if (curTable != NULL) delete curTable;
  node curHierNode;
  forEach (curHierNode, curHierarchy->getNodes()) {
    Graph *subGraph = hierarchyMetaGraphs->getNodeValue (curHierNode);
    if (subGraph != NULL)
      deleteDataStructures (subGraph);
  }
}

//================================================================
OpenCloseHierarchy::~OpenCloseHierarchy () {
  delete openClose;
  delete displayGraphRenderParams;
  delete edgeContraction;
  deleteDataStructures (hierarchy);
}//end ~OpenCloseHierarchy

//================================================================
void OpenCloseHierarchy::initNewSubGraph (const node &hierarchyNode,
					  bool keepLayoutParams) {
  Graph *curHierarchy = hierarchyMetaGraphs->getNodeValue (hierarchyNode);
  curHierarchy->setAttribute<EdgeOrderTable *> (EDGE_ORDER_ATTRIBUTE, NULL);
  curHierarchy->setAttribute<bool> (HAS_LAYOUT_ATTRIBUTE, false);

  if (!keepLayoutParams) {
    LayoutProperty *curHierarchyLayout =
      curHierarchy->getLocalProperty<LayoutProperty> ("viewLayout");
    curHierarchyLayout->setAllNodeValue (noPos);
    DoubleProperty *curHierarchyRotations =
      curHierarchy->getLocalProperty<DoubleProperty> ("viewRotation");
    curHierarchyRotations->setAllNodeValue (0);
  }
}//end initNewSubgraph

//================================================================
void
OpenCloseHierarchy::initializeOpenCloseHierarchy (const node &hierarchyNode,
						  bool keepLayoutParams) {
  initNewSubGraph (hierarchyNode, keepLayoutParams);
  Graph *curHierarchy = hierarchyMetaGraphs->getNodeValue (hierarchyNode);
  node curNode;
  forEach (curNode, curHierarchy->getNodes()) {
    containedByNode.set (curNode.id, hierarchyNode);
    Graph *nodeMetaGraph = hierarchyMetaGraphs->getNodeValue (curNode);
    //add only nodes in the original graph
    if (nodeMetaGraph == NULL) {
      if (inputGraph->isElement (curNode)) {
	hierNodeToDisplayMetaNode->setNodeValue (curNode, curNode.id);
	displayMetaNodeToHierNode->setNodeValue (curNode, curNode.id);
	displayGraph->addNode (curNode);
      }//end if
    }//end if
    //recursively go through the hierarchy
    else initializeOpenCloseHierarchy (curNode, keepLayoutParams);
  }//end forEach
}//end initialize hierarchy

//================================================================
unsigned int OpenCloseHierarchy::
getSizeEstimateSquared (const node &hierarchyNode) {
  Graph *subgraph = hierarchyMetaGraphs->getNodeValue (hierarchyNode);
  if (subgraph == NULL) return 0;
  unsigned int numLeaf = 0;
  node curNode;
  forEach (curNode, subgraph->getNodes ()) {
    Graph *subsub = hierarchyMetaGraphs->getNodeValue (curNode);
    if (subsub == NULL) ++numLeaf;
    else numLeaf += getSizeEstimateSquared (curNode);
  }//end forEach
  return numLeaf;
}//end getSizeEstimateSquared

//================================================================
Size OpenCloseHierarchy::setSizeEstimate (const node &hierarchyNode,
					  double numberOfNodes) {
  Graph *containingHierarchyGraph = 
    getContainingHierarchyGraph (hierarchyNode);
  SizeProperty *hierarchySizes =
    containingHierarchyGraph->getProperty<SizeProperty> ("viewSize");
  SizeProperty *hierarchyBoundingBox =
    containingHierarchyGraph->getLocalProperty<SizeProperty> ("boundingBox");
  double theSize = sqrt (numberOfNodes);
  if (onLogSize) theSize = max (log (theSize), 1.0);
  Size nodeSize (theSize, theSize, 1.0);
  Size boundingBox (M_SQRT2*theSize, M_SQRT2*theSize);
  hierarchySizes->setNodeValue (hierarchyNode, nodeSize);
  hierarchyBoundingBox->setNodeValue (hierarchyNode, boundingBox);
  return nodeSize;
}//end setContainingSizeEstimate

//================================================================
unsigned int OpenCloseHierarchy::closeAll (const node &curHierarchyNode) {
  Graph *curHierarchy = hierarchyMetaGraphs->getNodeValue (curHierarchyNode);
  unsigned int numLeaf = 0;
  node curNode;
  forEach (curNode, curHierarchy->getNodes()) {
    Graph *curSubHierarchy = hierarchyMetaGraphs->getNodeValue (curNode);
    if (curSubHierarchy != NULL) numLeaf += closeAll (curNode);
    else ++numLeaf;
  }//end forEach
  close (curHierarchyNode, true, numLeaf, true);
  return numLeaf;
}//end closeAll

//================================================================
void OpenCloseHierarchy::
getDisplayGraphSelection (const node &hierarchyNode,
			  BooleanProperty &subGraphSelection) {
  node curDisplayNode (hierNodeToDisplayMetaNode
		       ->getNodeValue (hierarchyNode));
  if (curDisplayNode.isValid () && displayGraph->isElement (curDisplayNode)) {
    Graph *curHierarchyMeta = 
      hierarchyMetaGraphs->getNodeValue (hierarchyNode);
      subGraphSelection.setNodeValue (curDisplayNode, true);
    return;
  }//end if
  Graph *hierarchySubGraph = hierarchyMetaGraphs->getNodeValue (hierarchyNode);
  if (hierarchySubGraph == NULL) return;
  node curHierarchyNode;
  forEach (curHierarchyNode, hierarchySubGraph->getNodes ())
    getDisplayGraphSelection (curHierarchyNode, subGraphSelection);
}//end getDisplayGraphSelection

//================================================================
void OpenCloseHierarchy::close (const node &hierarchyNode,
				const bool layoutReset,
				double sizeEstimateSquared,
				bool initialization) { 
  Graph *metaNodeHierarchyGraph = 
    hierarchyMetaGraphs->getNodeValue(hierarchyNode);
  bool hasLayout = 
    metaNodeHierarchyGraph->getAttribute<bool> (HAS_LAYOUT_ATTRIBUTE);
  Graph *hierarchyFather = getContainingHierarchyGraph (hierarchyNode);
  LayoutProperty *hierarchyLayout = 
    hierarchyFather->getProperty<LayoutProperty> ("viewLayout");
  Coord hierPos = hierarchyLayout->getNodeValue (hierarchyNode);
  ColorProperty *hierarchyColours =
    hierarchyFather->getProperty<ColorProperty> ("viewColor");
  Color hierColor = hierarchyColours->getNodeValue (hierarchyNode);
  SizeProperty *hierarchySizes =
    hierarchyFather->getProperty<SizeProperty> ("viewSize");
 
  string label = metaNodeHierarchyGraph->getAttribute<string> ("name");
  Size hierSize;
  if (hasLayout)
    hierSize = hierarchySizes->getNodeValue (hierarchyNode);
  else {
    if (sizeEstimateSquared < 1e-6)
      sizeEstimateSquared = getSizeEstimateSquared (hierarchyNode);
    hierSize = setSizeEstimate (hierarchyNode, sizeEstimateSquared);
  }//end else

  BooleanProperty subGraphSelection (displayGraph);
  subGraphSelection.setAllNodeValue (false);
  subGraphSelection.setAllEdgeValue (false);
  getDisplayGraphSelection (hierarchyNode, subGraphSelection);
  node addedNode = openClose->close (displayGraph, subGraphSelection, 
				     hierPos, META_NODE_SHAPE, 
				     hierColor, META_NODE_BORDER_WIDTH,
				     hierSize, label);
  hierNodeToDisplayMetaNode->setNodeValue (hierarchyNode, addedNode.id);
  displayMetaNodeToHierNode->setNodeValue (addedNode, hierarchyNode.id);
  if (!initialization) {
    vector<GlABoundingCircle *> augsAdded;
    updateCutLayout (hierarchy, noPos, 0, 1.0);
    updateBoundingCircles (hierarchy, noPos, 0, 1.0, 0, false, augsAdded);
    if (layoutReset) {
      clearLayoutParameters (metaNodeHierarchyGraph);
      initNewSubGraph (hierarchyNode, false);
    }//end if
  }//end if
}//end close

//================================================================
node OpenCloseHierarchy::getHierarchyNode (const node &groupNode) {
  if (!groupNode.isValid()) return invalidNode;
  node hierarchyNode (displayMetaNodeToHierNode->getNodeValue (groupNode));
  return hierarchyNode;
}//end getHierarchyNode

//================================================================
node OpenCloseHierarchy::getDisplayNode (const node &hierarchyNode) {
  if (!hierarchyNode.isValid()) return invalidNode;
  node displayNode (hierNodeToDisplayMetaNode->getNodeValue (hierarchyNode));
  return displayNode;
}//end getDisplayNode

//================================================================
void OpenCloseHierarchy::getPath (const node &hierarchyNode,
				  vector<node> &path) {
  node superNode = hierarchyNode;
  while (superNode.isValid ()) {
    path.push_back (superNode);
    superNode = containedByNode.get (superNode.id);
  }//end while
  reverse (path.begin (), path.end ());
}//end getPath

//================================================================
GlABoundingCircle *OpenCloseHierarchy::
getBoundingCircle (const node &hierarchyNode) {
  return boundingCircles->getBoundingCircle (hierarchyNode);
}//end getBoundingCircle

//================================================================
Graph *OpenCloseHierarchy::getHierarchyGraph (const node &hierarchyNode) {
  return hierarchyMetaGraphs->getNodeValue (hierarchyNode);
}//end getHierarchyGraph

//================================================================
void 
OpenCloseHierarchy::getDisplaySelection (const node &hierarchyNode,
					 BooleanProperty &displaySelection) {
  displaySelection.setAllNodeValue (false);
  displaySelection.setAllEdgeValue (false);
  getDisplayGraphSelection (hierarchyNode, displaySelection);
}//end getDisplaySelection

//================================================================
Graph *
OpenCloseHierarchy::getContainingHierarchyGraph (const node &hierarchyNode) {
  node containedBy = containedByNode.get (hierarchyNode.id);
  if (!containedBy.isValid() && hierarchy->isElement (hierarchyNode)) 
    return hierarchy; 
  else return hierarchyMetaGraphs->getNodeValue (containedBy);
}//end getContainingHierarchyGraph

//================================================================
node
OpenCloseHierarchy::getContainingHierarchyNode (const node &hierarchyNode) {
  return containedByNode.get (hierarchyNode.id);
}

//================================================================
void OpenCloseHierarchy::findAugsBelow (const node &curHierarchyNode,
					std::vector<GlABoundingCircle *> 
					&augsBelow) {
  GlABoundingCircle *curAug = 
    boundingCircles->getBoundingCircle (curHierarchyNode);
  if (curAug != NULL)
    augsBelow.push_back (curAug);

  Graph *subGraph = hierarchyMetaGraphs->getNodeValue (curHierarchyNode);
  if (subGraph != NULL) {
    node curSubNode;
    forEach (curSubNode, subGraph->getNodes())
      findAugsBelow (curSubNode, augsBelow);
  }//end if
}//end findAugsBelow

//================================================================
bool OpenCloseHierarchy::open (node oldMetaNode,
			       vector<node> &nodesAdded,
			       vector<edge> &edgesAdded,
			       vector<GlABoundingCircle *> &augsAdded,
			       const bool layoutReset,
			       bool synchronizing) {
  if (openClose->getGroupedGraph (oldMetaNode) == NULL) return false;
  node hierarchyNode = getHierarchyNode (oldMetaNode);
  bool hasLayout = false;

  if (!synchronizing) {
    Graph *metaNodeHierarchyGraph = 
      hierarchyMetaGraphs->getNodeValue (hierarchyNode);
    if (layoutReset) {
      clearLayoutParameters (metaNodeHierarchyGraph);
      initNewSubGraph (hierarchyNode, false);
    }//end if
    hasLayout =
      metaNodeHierarchyGraph->getAttribute<bool> (HAS_LAYOUT_ATTRIBUTE);
    if (!hasLayout) {
      edgeContractIfNeeded (hierarchyNode);
      
      //if the hierarchy has changed, synch metagraph and openclose data.
      if (hierarchyChange) {
	synchOpenCloseAndHierarchy (hierarchyNode);
	hierarchyChange = false;
      }// end if

      Graph *hierarchyFather = getContainingHierarchyGraph (hierarchyNode);
      bool success = drawSteeredNode (metaNodeHierarchyGraph, hierarchyFather,
				      useDefaultAlgorithms, pluginProgress);
      if (!success) return false;
    }//end if no layout
  }//end !syncronizing

  //record added nodes and edges
  node metaNode = getDisplayNode (hierarchyNode);
  Graph *groupGraph = openClose->getGroupedGraph (metaNode);
  node curNode;
  forEach (curNode, groupGraph->getNodes())
    nodesAdded.push_back (curNode);
  edge curEdge;
  forEach (curEdge, groupGraph->getEdges())
    edgesAdded.push_back (curEdge);
 
  openClose->open (displayGraph, metaNode);
  hierNodeToDisplayMetaNode->setNodeValue (hierarchyNode, invalidNode.id);
  displayMetaNodeToHierNode->setNodeValue (metaNode, invalidNode.id);
  if (!synchronizing) {
    if (!hasLayout) {
      if (!scaleMode)
	cleanupDisplay (hierarchyNode, augsAdded);
      else
	scaleDown (hierarchyNode, augsAdded);
    }// end if
    else {
      updateCutLayout (hierarchy, noPos, 0, 1.0);
      updateBoundingCircles (hierarchy, noPos, 0, 1.0, 0, false, augsAdded);
    }//end else
  }//end if  
  return true;
}//end open

//================================================================
void OpenCloseHierarchy::updateDisplayGraphRendering () {
  displayGraphRenderParams->ensureNodeColours ();
  displayGraphRenderParams->ensureEdgeColours ();
  //displayGraphRenderParams->ensureNodeRenderOrder ();
  //GlGraphRenderingParameters param = 
  //  glGraph->getRenderingParameters();
  //param.setElementOrdered (true);
  //glGraph->setRenderingParameters(param);
  displayGraphRenderParams->ensureEdgeSizes ();
  edgeBundling->toggleEdgeBundling (hasEdgeBundling);
}

//================================================================
void OpenCloseHierarchy::openAll (const node &displayNode) {
  vector<node> nodesAdded;
  vector<edge> edgesAdded;
  vector<GlABoundingCircle *> augsAdded;
  if (openClose->getGroupedGraph (displayNode) != NULL)
    open (displayNode, nodesAdded, edgesAdded, augsAdded, false, true);

  for (unsigned int i = 0; i < nodesAdded.size(); ++i) {
    if (openClose->getGroupedGraph (nodesAdded[i]) != NULL)
      openAll (nodesAdded[i]);
  }//end for
}//end closeAllAndUnsetLayout

//================================================================
void OpenCloseHierarchy::synchOpenCloseData (const node &hierarchyNode,
					     bool keepLayoutParams) {
  node displayNode = getDisplayNode (hierarchyNode);
  openAll (displayNode);
  initializeOpenCloseHierarchy (hierarchyNode, keepLayoutParams);
  closeAll (hierarchyNode);
}//end synchDisplayGraph

//================================================================
void OpenCloseHierarchy::initHierarchyNodeLevel (const node &hierarchyNode,
						 const node &superNode) {
  initNewSubGraph (hierarchyNode, true);
  Graph *hierarchySubGraph = hierarchyMetaGraphs->getNodeValue (hierarchyNode);
  node newContain;
  forEach (newContain, hierarchySubGraph->getNodes())
    containedByNode.set (newContain.id, hierarchyNode);
  containedByNode.set (hierarchyNode.id, superNode);
}//end initHierarchyNodeLevel

//================================================================
void OpenCloseHierarchy::redrawPath (const node &startNode) {
  vector<GlABoundingCircle *> augsAdded;
  if (!scaleMode) cleanupDisplay (startNode, augsAdded);
  else scaleDown (startNode, augsAdded);
}//end redrawPath

//================================================================
void OpenCloseHierarchy::addOneLevel (const node &hierarchyNode,
				      const node &superNode) {
  Graph *hierarchySubGraph = hierarchyMetaGraphs->getNodeValue (hierarchyNode);
  initHierarchyNodeLevel (hierarchyNode, superNode);
}//end addOneLevel

//================================================================
void OpenCloseHierarchy::synchHierarchyBelow (const node &hierarchyNode) {
  initNewSubGraph (hierarchyNode, true);
  Graph *hierarchySubGraph = hierarchyMetaGraphs->getNodeValue (hierarchyNode);
  node newContain;
  forEach (newContain, hierarchySubGraph->getNodes())
    containedByNode.set (newContain.id, hierarchyNode);
}//end fixLevelBelow

//================================================================
void OpenCloseHierarchy::edgeContractIfNeeded (const node &hierarchyNode) {
  if (isContractionNeeded (hierarchyNode)) {
    cout << "Edge contraction" << endl;
    clock_t start;
    time_t startLong;
    startTime (start, startLong);
    edgeContraction->contractHierarchy (hierarchyNode, maxNodeNum);
    endTime (start, startLong);
    setHierarchyChange ();
  }//end if
}//end edgeContractIfNeeded

//================================================================
bool OpenCloseHierarchy::isContractionNeeded (const node &hierarchyNode) {
  Graph *hierarchyGraph = hierarchyMetaGraphs->getNodeValue (hierarchyNode);
  return (hierarchyGraph->numberOfNodes () > maxNodeNum) &&
    !TreeTest::isFreeTree (hierarchyGraph);
}//end contractionNeeded

//================================================================
void OpenCloseHierarchy::synchOpenCloseAndHierarchy (const node &hierarchyNode) {
  synchOpenCloseData (hierarchyNode);
  metaGraphHierarchyWidget->updateTree (hierarchyNode);
}
