
/**
 * Keeps track of a subgraph hierarchy for opening and closing nodes.
 */
#ifndef __OPEN_CLOSE_HIERARCHY_H__
#define __OPEN_CLOSE_HIERARCHY_H__
#include <string>
#if (__GNUC__ < 3)
#  include <hash_map>
#else
#  include <ext/hash_map>
#endif
#include <tulip/TlpTools.h>
#include <tulip/ForEach.h>
#include <tulip/Graph.h>
#include <tulip/GlGraph.h>
#include <tulip/LayoutProperty.h>
#include <tulip/StringProperty.h>
#include <tulip/IntegerProperty.h>
#include <tulip/BooleanProperty.h>
#include <tulip/SizeProperty.h>
#include <tulip/GraphProperty.h>
#include <tulip/MutableContainer.h>
#include <tulip/MetaGraphHierarchyWidget.h>
#include <tulip/PluginProgress.h>
#include <tulip/QtProgressWidget.h>
#include "MetaNodeBoundingCircle.h"
#include "ColourShapeDefaults.h"
#include "OpenClose.h"
#include "SteeredLayout.h"
#include "HierarchyMap.h"
#include "DisplayGraphRendering.h"
#include "NodeDegEdgeContraction.h"
#include "EdgeBundling.h"

#define OPEN_CLOSE_HIERARCHY_NAME ("display graph")
#define hierToDisplayProperty (std::string ("view Hier Node to Display"))
#define displayToHierProperty (std::string ("view Display Node to Hier"))

class OpenCloseHierarchy {
 private:
  bool hierarchyChange;
  bool scaleMode;
  bool useDefaultAlgorithms;
  bool hasEdgeBundling;
  unsigned int maxNodeNum;
  tlp::GlGraph *glGraph;
  bool onLogSize;
  tlp::Graph *displayGraph; //the display graph of the hierarchy
  tlp::Graph *inputGraph; //pointer to orignal input graph
  tlp::Graph *hierarchy; //pointer to meta-graph hierarchy
  tlp::Graph *hierarchySubGraphs;
  tlp::PluginProgress *pluginProgress;

  OpenClose *openClose;
  DisplayGraphRendering *displayGraphRenderParams;
  EdgeBundling *edgeBundling;
  NodeDegEdgeContraction *edgeContraction;
  tlp::LayoutProperty *displayLayout; //hierarchy node positions
  tlp::SizeProperty *displaySizes; //hierarchy node sizes
  tlp::GraphProperty *hierarchyMetaGraphs; //property for hierarchy meta graphs

  MetaNodeBoundingCircles *boundingCircles;
  tlp::IntegerProperty *hierNodeToDisplayMetaNode;
  tlp::IntegerProperty *displayMetaNodeToHierNode;
  MetaGraphHierarchyWidget *metaGraphHierarchyWidget;
  tlp::MutableContainer<tlp::node> containedByNode;
  tlp::node hierarchyRoot;

  /**
   * When a new hierarchy subgraph has been created, initNewSubGraph
   * initializes this subgraph
   * \param hierarchyNode The hierarchy node containing this level
   * \param keepLayoutParams if set to true, view layout and rotation not
   * reset.
   */
  void initNewSubGraph (const tlp::node &hierarchyNode,
			bool keepLayoutParams);
  
  /**
   * A function to set up initial display nodes in the openCloseHierarchy
   * \param hierarchyNode The hierarchy node containing this level
   */
  void initializeOpenCloseHierarchy (const tlp::node &hierarchyNode,
				     bool keepLayoutParams = false);

  /**
   * Private function to delete all data structures.  Used by destructor only.
   * \param curHierarchy A subgraph of the graph hierarchy.
   */
  void deleteDataStructures (Graph *curHierarchy);

  /**
   * Function to update the bounding circle positions, size, and rotation.
   * \param curHierGraph name of the hierarchy meta graph
   * \param pos New position of the sub graph
   * \param rotation New rotation of the subgraph
   * \param augsAdded Bounding circles added along the way (if any)
   */
  void updateBoundingCircles (tlp::Graph *curHierGraph, 
			      const tlp::Coord &pos, 
			      double rotation, float scale,
			      unsigned int depthFromRoot,
			      bool hideBelow,
			      vector<GlABoundingCircle *> &augsAdded);

  /**
   * Function to update the layouts along a particular chain.
   * \param hierarchyNode The hierarchy node currently opened
   */
  void redrawChain (const tlp::node &hierarchyNode);

  /**
   * Function to update the positions of the nodes along the cut.
   * \param curHierarchySubGraph - The current hierarchy subgraph being
   * adjusted.
   * \param cumulativeTrans - Cumulative translation to this point
   * \param cumulativeRotation - Cumulative rotation to this point
   * \param cumulativeScale - Cumulative scale to this point.
   */
  void updateCutLayout (tlp::Graph *curHierarchySubGraph,
			const tlp::Coord &cumulativeTrans,
			double cumulativeRotation,
			float cumulativeScale);

  /**
   * Function to scale subgraph just opened to the metanode size.
   * \param hierarchyNode Subgraph to scale to required size.
   */
  void scaleDown (const tlp::node &hierarchyNode,
		  vector<GlABoundingCircle *> &augsAdded);

  /**
   * returns a size estimate based on the number of leaves below
   * a node in the hierarchy.
   */
  unsigned int getSizeEstimateSquared (const node &hierarchyNode);

  /**
   * Function to compute a size estimate based on the number of nodes
   * when no layout has been computed thus far.
   * \param hierarchyNode - The graph which the size estimate is being computed
   * \param numberOfNodes - The number of nodes or the square of the
   * size requested.
   */
  tlp::Size setSizeEstimate (const tlp::node &hierarchyNode,
			     double numberOfNodes);

  /**
   * Funtion to return the nodes in the cut display below a passed
   * hierarchy node.
   * \param hierarchyNode The node below which we want the display selection
   * \param displayGraph The displayGraph
   * \param subGraphSelection The set of display nodes below hierarchyNode.
   */
  void getDisplayGraphSelection (const node &hierarchyNode,
				 BooleanProperty &subGraphSelection);
  /**
   * Function to close all meta nodes in the hierarchy below a hierarchy node.
   * \param hierarchyNode Node below which everything should be closed.
   */
  unsigned int closeAll (const tlp::node &hierarchyNode);

  /**
   * Function to open all meta nodes in the display graph contained by
   * the passed display node.
   * \param displayNode Display node to be opened all the way down.
   */
  void openAll (const tlp::node &displayNode);

  /**
   * Function to perform edge contraction if the number of nodes in the subgraph is
   * too large.
   * \param hierarchyNode The node below which edge contraction is performed if needed.
   */
  void edgeContractIfNeeded (const tlp::node &hierarchyNode);

  /**
   * Given a hierarchy node, this function synchronizes the
   * open close data below it.
   * \param hierarchyNode The node which has a new set of hierarchy sub
   * graphs below it.
   */
  void synchOpenCloseData (const tlp::node &hierarchyNode,
			   bool keepLayoutParams = false);

 public:
  OpenCloseHierarchy (GlGraph *glGraph, tlp::Graph *hierarchy, 
		      const tlp::node &hierarchyRoot,
		      tlp::Graph *hierarchySubGraphs,
		      tlp::Graph *displayGraph,
		      tlp::Graph *inputGraph,
		      MetaNodeBoundingCircles *boundingCircles,
		      MetaGraphHierarchyWidget 
		      *metaGraphHierarchyWidget,
		      tlp::PluginProgress *pluginProgress,
		      bool scaleMode, bool useDefaultAlgorithms);
  
  ~OpenCloseHierarchy (); 
  inline static bool isLeaf (const tlp::node &displayNode,
			     const tlp::node &hierarchyNode) {
    return displayNode == hierarchyNode;
  }
  inline void setLogSize (bool newLogSize) { onLogSize = newLogSize; }
  inline void setScaleMode (bool newScaleMode) { scaleMode = newScaleMode; }
  inline void setUseDefaultAlgorithms (bool newUseDefaultAlgorithms) { 
    useDefaultAlgorithms = newUseDefaultAlgorithms; 
  }
  inline void setEdgeBundling (bool on) { hasEdgeBundling = on; }

  inline bool getUseDefaultAlgorithms () { return useDefaultAlgorithms; }
  inline bool getEdgeBundling () { return hasEdgeBundling; }
  inline void setMaxNodeNum (unsigned int newMaxNodeNum) {
    maxNodeNum = newMaxNodeNum;
  }//end setMaxNodeNum

  inline unsigned int getMaxNodeNum () {
    return maxNodeNum;
  }

  /**
   * Variable to indicate that the hierarchy has changed and should
   * be fixed on opening a metanode
   */
  inline void setHierarchyChange () {
    hierarchyChange = true;
  }

  /**
   * Returns a subgraph pointer to the hierarchy
   */
  inline Graph *getDisplayGraph () {
    return displayGraph;
  }//end getOpenCloseHierarchy

  /**
   * Function to get the hierarchy node given a group/display node.  Returns
   * the all closed node.
   * \param groupNode the display node. 
   */
  tlp::node getHierarchyNode (const tlp::node &displayNode);

  /**
   * Function to get a display node given a hierarchy node.
   * \param hierarchyNode The hierarchy node.
   */
  tlp::node getDisplayNode (const tlp::node &hierarchyNode);

  /**
   * Function to get the hierarchy subgraph given the hierarchy node.
   * \param hierarchyNode The hierarchy node.
   */
  tlp::Graph *getHierarchyGraph (const tlp::node &hierarchyNode);

  /**
   * Function to get the selection of the display graph below this hierarchy
   * node.
   * \param hierarchyNode The hierarchy node.
   * \param displaySelection The display graph below
   */
  void getDisplaySelection (const tlp::node &hierarchyNode,
			    tlp::BooleanProperty &displaySelection);

  /**
   * Function to get containing graph in the graph hierarchy given
   * the hierarchy node.  Returns null if the hierarchy root node is
   * passed in.
   * \param hierarchyNode The hierarchy node.
   */
  tlp::Graph *getContainingHierarchyGraph (const tlp::node &hierarchyNode);

  /**
   * Function to get containing node of the passed hierarchy node.
   * \param hierarchyNode The hierarchy node.
   */
  tlp::node getContainingHierarchyNode (const tlp::node &hierarchyNode);

  /**
   * Returns the global position of the passed hierarchy metanode.
   * \param globalNode Hierarchy metanode where position is found.
   */
  tlp::Coord getGlobalPos (const tlp::node &globalNode);

  /**
   * Returns the path through the hierarchy to the passed metanode
   * \param hierarchyNode Hierarchy metanode where the path is found.
   */
  void getPath (const tlp::node &hierarchyNode,
		std::vector<tlp::node> &path);

  /**
   * Function to get a bouding circle given a hierarchy node
   * \param hierarchyNode The hierarchy node.
   */
  GlABoundingCircle *getBoundingCircle (const tlp::node &hierarchyNode);

  /**
   * Provides a list of all the bounding circles below this
   * hierarchy node that are open.
   * \param hierarchyNode The node below which we want to find augs.
   * \param augsBelow The list of augmented displays below
   */
  void findAugsBelow (const tlp::node &hierarchyNode,
		      std::vector<GlABoundingCircle *> &augsBelow);

  /**
   * Function to close a subgraph into a meta-node.
   * \param hierarchyNode - The node in the graph hierarchy
   * \param layoutReset - The layout of the subgraph is reset after the
   * close event if true
   * \param sizeEstimateSquared - A size estimate for the closed node
   * a value of zero sets the size estimate proportional to the number of 
   * nodes.
   * \param initialization states if this was called using close all.
   */
  void close (const tlp::node &hierarchyNode,
	      const bool layoutReset,
	      double sizeEstimateSquared = 0,
	      bool initialization = false);

  /**
   * Function to open a meta-node in the hierarchy.  Returns false if
   * the node does not open.
   * \param metaNode The meta-node to be ungrouped.
   * \param nodesAdded The nodes added due to the open event
   * \param edgesAdded The edges added due to the open event
   * \param augsAdded The Augemented displays added due to the event
   * \param layoutReset The layout of the subgraph is reset before drawing
   * if this parameter is set to true.
   * \param synchonizing If this is a synchonizing event.  If true,
   * graphs are not layed out below the node and no augmented displays
   * are added.
   */
  bool open (tlp::node metaNode,
	     std::vector<tlp::node> &nodesAdded,
	     std::vector<tlp::edge> &edgesAdded,
	     std::vector<GlABoundingCircle *> &augsAdded,
	     const bool layoutReset,
	     bool synchronizing = false);

  /**
   * Function to clean up the display.  It starts at the passed hierarchy node
   * and redraws all meta graphs up the branch to the root.
   * \param hierarchyNode The starting node of the chain redraw.
   */
  void cleanupDisplay (const tlp::node &hierarchyNode,
		       vector<GlABoundingCircle *> &augsAdded);

  /**
   * The following function synchs the display graph with the hierarchy
   * graph and applies user options.
   */
  void updateDisplayGraphRendering ();

   /**
   * Given a hierarchy node, this function synchronizes the
   * open close data and the graph hierarchy.
   * \param hierarchyNode The node which has a new set of hierarchy sub
   * graphs below it.
   */
  void synchOpenCloseAndHierarchy (const node &hierarchyNode);

  /**
   * Function to initialize a new hierarchy level which occurs between
   * two other hierarchy levels in the graph hierarchy
   * \param hierarchyNode The node below which one level has been added
   * \param superNode The node which contains the new hierarchyNode
   */
  void initHierarchyNodeLevel (const tlp::node &hierarchyNode,
			       const tlp::node &superNode);
  /**
   * Given a hierarchy subgraph, this funching synchronizes the one
   * level which has been added below hierarchyNode
   * \param hierarchyNode The node containing the new level.
   * \param superNode The node which contains the newly added hierarchy node
   */
  void addOneLevel (const tlp::node &hierarchyNode,
		    const tlp::node &superNode);

  /**
   * Updates data structures so that the level below is updated to
   * make hierarchy consistant.
   */
  void synchHierarchyBelow (const tlp::node &hierarchyNode);

  /**
   * Redraws the path from startNode up to the root.  
   * \param startNode The node where the path redraw starts.
   */
  void redrawPath (const node &startNode);

  /**
   * Function to reset the node and edge positions of the display
   * graph so that they are the same as the hierarchy.
   */
  void resynchDisplayGraph ();

  /**
   * Function to determine if edge contraction is needed.  The graph
   * is not modified.  It returns true if contraction is needed.
   * \param hierarchyNode The node below which the graph is checked.
   */
  bool isContractionNeeded (const tlp::node &hierarchyNode);
};

/**
 * Returns the centre (by area) of the subgraph.
 * \param subGraph Hierarchy subgraph we want the centre for.
 */
tlp::Coord computeCentre (tlp::Graph *subGraph);

/**
 * centres the hierarchy subgraph by area.
 * \param subGraph Hierarchy subgraph we want the centre for.
 * \param subGraphLayout Layout proxy which is centred.
 */
void centreByArea (tlp::Graph *subGraph, tlp::LayoutProperty &subGraphLayout);

/**
 * Computes and sets the bounding circles and boxes of the appropriate
 * node in fatherGraph for the passed childGraph and node.
 * \param fatherGraph The graph which contains the childNodeInFather
 * \param childGraph The graph contained by the node childNodeInFather
 * \param childNodeInFather The node of fatherGraph which represents
 * childGraph
 */
void setBoundingBoxAndCircle (Graph *fatherGraph, Graph *childGraph,
			      const node &childNodeInFather);
#endif
