#include <string>
#include <tulip/ForEach.h>
#include <tulip/DrawingTools.h>
#include <tulip/LayoutProperty.h>
#include <tulip/SizeProperty.h>
#include <tulip/DoubleProperty.h>
#include "HierarchyMap.h"
#include "OpenCloseHierarchy.h"
#include "SteeredLayout.h"
using namespace tlp;
using namespace std;

const Coord noPos (0, 0, 0);
const Size noSize (1, 1, 1);

//a function to convert degrees to radians
inline double toRad (double deg) {
  return deg*(M_PI/180.0);
}//end toRad

//a fundtion to convert radians to degrees
inline double toDeg (double rad) {
  return rad*(180.0/M_PI);
}//end toDeg

//a function to rotate by theta in degreesabout a (0, 0) origin
inline Coord rotate (double theta_deg, const Coord pos) {
  Coord result;
  double theta = toRad (theta_deg);
  result.setX (pos.getX()*std::cos(theta) - pos.getY()*std::sin (theta));
  result.setY (pos.getX()*std::sin(theta) + pos.getY()*std::cos (theta));
  result.setZ (0);
  return result;
}//end rotate

//=====================================================================
Coord computeCentre (Graph *subGraph) {
  SizeProperty *sizes = subGraph->getProperty<SizeProperty> ("viewSize");
  LayoutProperty *layout = 
    subGraph->getProperty<LayoutProperty> ("viewLayout");

  //compute the centre of the localGraph by area
  Coord max (-DBL_MAX, -DBL_MAX, 0), min (DBL_MAX, DBL_MAX, 0);
  node curNode;
  forEach (curNode, subGraph->getNodes()) {
    Coord pos = layout->getNodeValue (curNode);
    Size size = sizes->getNodeValue (curNode) / 2.0;
    double rad = sqrt (size.getW()*size.getW() + 
		       size.getH()*size.getH());
    if (max.getX() < pos.getX() + rad)
      max.setX (pos.getX() + rad);
    if (min.getX() > pos.getX() - rad)
      min.setX (pos.getX() - rad);
    if (max.getY() < pos.getY() + rad)
      max.setY (pos.getY() + rad);
    if (min.getY() > pos.getY() - rad)
      min.setY (pos.getY() - rad);
  } //end for each node

  //update edge bends
  edge curEdge;
  forEach (curEdge, subGraph->getEdges()) {
    vector<Coord> bends = layout->getEdgeValue (curEdge);
    vector<Coord>::iterator bendsIt = bends.begin();
    for (; bendsIt != bends.end(); ++bendsIt) {
      Coord pos = *bendsIt;
      if (max.getX() < pos.getX())
	max.setX (pos.getX());
      if (min.getX() > pos.getX())
	min.setX (pos.getX());
      if (max.getY() < pos.getY())
	max.setY (pos.getY());
      if (min.getY() > pos.getY())
	min.setY (pos.getY());
    }
  } 
  return (max + min)/2.0;
}//end computeCentre

//=====================================================================
void centreByArea (Graph *subGraph, LayoutProperty &subGraphLayout) {
  Coord centre = computeCentre (subGraph);

  //centre the localGraph by area
  node curNode;
  forEach (curNode, subGraph->getNodes())
    subGraphLayout.setNodeValue (curNode,
				 subGraphLayout.getNodeValue (curNode)
				 - centre);

  //update edge bends
  Iterator<edge> *edges = subGraph->getEdges();
  while (edges->hasNext()) {
    edge curEdge = edges->next();
    vector<Coord> bends = subGraphLayout.getEdgeValue (curEdge);
    vector<Coord>::iterator bendsIt = bends.begin();
    for (; bendsIt != bends.end(); ++bendsIt) 
      *bendsIt = *bendsIt - centre;
    subGraphLayout.setEdgeValue (curEdge, bends);
  } delete edges;
}//end centreByArea

//================================================================
void setBoundingBoxAndCircle (Graph *fatherGraph, Graph *childGraph,
			      const node &childNodeInFather) {
  LayoutProperty *viewLayout =
    fatherGraph->getProperty<LayoutProperty> ("viewLayout");
  SizeProperty *viewSize =
    fatherGraph->getProperty<SizeProperty> ("viewSize");
  SizeProperty *boundingBoxes =
    fatherGraph->getLocalProperty<SizeProperty> ("boundingBox");

  pair<Coord, Coord> metaGraphBoundCircle = 
    tlp::computeBoundingRadius (childGraph, 
				childGraph
				->getProperty<LayoutProperty> ("viewLayout"),
				childGraph
				->getProperty<SizeProperty> ("viewSize"),
				childGraph
				->getProperty<DoubleProperty> ("viewRotation"));
  double radius = (metaGraphBoundCircle.second - 
		   metaGraphBoundCircle.first).norm();
  double diamLength = M_SQRT2*radius;
  Size size (diamLength, diamLength, 1.0);
  viewSize->setNodeValue (childNodeInFather, size);
  Size boxSize (2*radius, 2*radius, 1.0);
  boundingBoxes->setNodeValue (childNodeInFather, boxSize);
}//end setBoundingBoxAndCircle

//================================================================
//Local function to apply torques to the passed curGraph
void applyTorques (Graph *torqueGraph, DoubleProperty *viewRotation,
		   Graph *inputGraph) {
  DoubleProperty rotations (torqueGraph);
  string errMsg;
  DataSet torqueData;
  torqueData.set ("origGraph", inputGraph);


  bool success = 
    torqueGraph->computeProperty ("Torques", &rotations, errMsg,
				  NULL, &torqueData);
  if (!success) {
    cerr << "Error calling torques:\n" << endl << errMsg << endl;
    return;
  }//end if

  node curNode;
  forEach (curNode, torqueGraph->getNodes())
    viewRotation->setNodeValue (curNode, 
				rotations.getNodeValue (curNode));
}//end applyTorques

//================================================================
void 
OpenCloseHierarchy::updateBoundingCircles (Graph *curHierarchySubGraph,
					   const Coord &cumulativeTrans,
					   double cumulativeRotation,
					   float cumulativeScale,
					   unsigned int depthFromRoot,
					   bool hideBelow,
					   vector<GlABoundingCircle *> 
					   &augsAdded) {

  LayoutProperty *hierLayout =
    curHierarchySubGraph->getProperty<LayoutProperty> ("viewLayout");
  SizeProperty *hierSize =
    curHierarchySubGraph->getProperty<SizeProperty> ("viewSize");
  DoubleProperty *hierRotation =
    curHierarchySubGraph->getProperty<DoubleProperty> ("viewRotation");
  GraphProperty *hierGraph =
    curHierarchySubGraph->getProperty<GraphProperty> ("viewMetaGraph");
  bool scaling = curHierarchySubGraph->existProperty ("viewScale");

  node curNode;
  forEach (curNode, curHierarchySubGraph->getNodes()) {
    Graph *nodeSubGraph = hierGraph->getNodeValue (curNode);
    if (nodeSubGraph == NULL) continue;
    node displayMetaNode = getDisplayNode (curNode);
    GlABoundingCircle *metaGraphBoundCircle =
      boundingCircles->getBoundingCircle (curNode);

    Coord curPos = hierLayout->getNodeValue (curNode);
    Size curSize = hierSize->getNodeValue (curNode);
    double curRotation = hierRotation->getNodeValue (curNode);
    if (cumulativeScale > 1e-6) curPos *= cumulativeScale;
    Coord rotCoord = rotate (cumulativeRotation, curPos);
    Coord globalPos = rotCoord + cumulativeTrans;
    curSize *= cumulativeScale;
    //No bounding circle but no display metanode means we need to
    //add a bounding circle to the graph
    if ((metaGraphBoundCircle == NULL) && 
	!displayMetaNode.isValid() && !hideBelow) {
      metaGraphBoundCircle = 
	boundingCircles->addBoundingCircle (displayGraph,
					    curNode, globalPos, curSize,
					    depthFromRoot);
      augsAdded.push_back (metaGraphBoundCircle);
    }

    //a bounding circle exists and it may need to be displayed
    else if (metaGraphBoundCircle != NULL) {
      boundingCircles->updateBoundingCircle (curNode, globalPos, curSize, 
					     depthFromRoot);
      bool nextHideBelow = hideBelow;
      if (displayMetaNode.isValid() || hideBelow) {
	if (!boundingCircles->isHidden (curNode))
	  boundingCircles->hideBoundingCircle (curNode);
	nextHideBelow = true;
      }//end if
      else {
	if (boundingCircles->isHidden (curNode)) {
	  boundingCircles->showBoundingCircle (curNode);
	  augsAdded.push_back (metaGraphBoundCircle);
	}//end if
      }//end else

      float scaleFactor = 1.0;
      if (scaling) {
	DoubleProperty *viewScale = 
	  curHierarchySubGraph->getProperty<DoubleProperty> ("viewScale");
	scaleFactor = viewScale->getNodeValue (curNode);
      }
      updateBoundingCircles (nodeSubGraph, globalPos,
			     curRotation + cumulativeRotation,
			     scaleFactor*cumulativeScale, depthFromRoot + 1,
			     nextHideBelow, augsAdded);
    }//end if
  }//end forEach
}//end updateBoundingCircle

//================================================================
inline Coord transformPosition (const Coord &curPos,
				double rot,
				const Coord cumulativeTrans,
				float cumulativeScale,
				double cumulativeRotation,
				bool scaling) {
  Coord globalPos = curPos;
  if (scaling && (cumulativeScale > 1e-6)) 
    globalPos *= cumulativeScale;
  Coord rotCoord = rotate (cumulativeRotation, globalPos);
  globalPos = rotCoord + cumulativeTrans;
  return globalPos;
}

//=====================================================================
Coord OpenCloseHierarchy::getGlobalPos (const node &globalNode) {

  vector<node> path;
  Coord globalPos (0, 0, 0);
  float globalScale = 1.0;
  double globalRotation = 0.0;
  getPath (globalNode, path);

  for (unsigned int i = 0; i < path.size() - 1; ++i) {
    Graph *curHierarchySubGraph = hierarchyMetaGraphs->getNodeValue (path[i]);
    node curNode = path[i + 1];
    LayoutProperty *hierLayout =
      curHierarchySubGraph->getProperty<LayoutProperty> ("viewLayout");
    SizeProperty *hierSize =
      curHierarchySubGraph->getProperty<SizeProperty> ("viewSize");
    DoubleProperty *hierRotation =
      curHierarchySubGraph->getProperty<DoubleProperty> ("viewRotation");
    bool scaling = curHierarchySubGraph->existProperty ("viewScale");
    Coord curPos = hierLayout->getNodeValue (curNode);
    Size curSize = hierSize->getNodeValue (curNode);
    double curRotation = hierRotation->getNodeValue (curNode);
    globalPos =
      transformPosition (curPos, curRotation,
			 globalPos, globalScale, globalRotation,
			 scaling);
  }//end for
  return globalPos;
}//end getGlobalPos

//================================================================
void 
OpenCloseHierarchy::updateCutLayout (Graph *curHierarchySubGraph,
				     const Coord &cumulativeTrans,
				     double cumulativeRotation,
				     float cumulativeScale) {
  LayoutProperty *hierLayout =
    curHierarchySubGraph->getProperty<LayoutProperty> ("viewLayout");
  SizeProperty *hierSize =
    curHierarchySubGraph->getProperty<SizeProperty> ("viewSize");
  DoubleProperty *hierRotation =
    curHierarchySubGraph->getProperty<DoubleProperty> ("viewRotation");
  bool scaling = curHierarchySubGraph->existProperty ("viewScale");

  edge curEdge;
  forEach (curEdge, curHierarchySubGraph->getEdges()) {
    node displaySource = 
      getDisplayNode (curHierarchySubGraph->source (curEdge));
    node displayTarget =
      getDisplayNode (curHierarchySubGraph->target (curEdge));
    edge displayEdge; 
    if (displaySource.isValid () && displayTarget.isValid ())
      displayEdge = displayGraph->existEdge (displaySource, displayTarget);
    if (displayEdge.isValid()) {
      vector<Coord> bends = hierLayout->getEdgeValue (curEdge);
      vector<Coord>::iterator bendsIt = bends.begin();
      for (; bendsIt != bends.end(); ++bendsIt) {
	*bendsIt = transformPosition (*bendsIt, 0,
				      cumulativeTrans, cumulativeScale,
				      cumulativeRotation, scaling);
      }//end for
      displayLayout->setEdgeValue (displayEdge, bends);
    }//end if
  }//end forEach
  
  node curNode;
  forEach (curNode, curHierarchySubGraph->getNodes()) {
    Coord curPos = hierLayout->getNodeValue (curNode);
    Size curSize = hierSize->getNodeValue (curNode);
    double curRotation = hierRotation->getNodeValue (curNode);
    Coord globalPos =
      transformPosition (curPos, curRotation,
			 cumulativeTrans, cumulativeScale, cumulativeRotation,
			 scaling);
    Graph *nodeSubGraph = hierarchyMetaGraphs->getNodeValue (curNode);
    if (nodeSubGraph == NULL) {
      displayLayout->setNodeValue (curNode, globalPos);
      if (scaling) {
	Size curSize = hierSize->getNodeValue(curNode);
	curSize *= cumulativeScale;
	displaySizes->setNodeValue (curNode, curSize);
      }
      continue;
    }//end if

    node groupedNode = getDisplayNode (curNode);
    if (groupedNode.isValid()) {
      displayLayout->setNodeValue (groupedNode, globalPos);
      if (scaling) {
	Size curSize = hierSize->getNodeValue(curNode);
	curSize *= cumulativeScale;
	displaySizes->setNodeValue (groupedNode, curSize);
      }
      continue;
    }//end if
    else {
      float scaleFactor = 1.0;
      if (scaling) {
	DoubleProperty *viewScale = 
	  curHierarchySubGraph->getProperty<DoubleProperty> ("viewScale");
	scaleFactor = viewScale->getNodeValue (curNode);
      }
      updateCutLayout (nodeSubGraph, globalPos,
		       curRotation + cumulativeRotation, 
		       scaleFactor*cumulativeScale);
    }
  }//end forEach
}//end updateCutLayout


//================================================================
void OpenCloseHierarchy::redrawChain (const node &hierarchyNode) {
  Graph *metaGraphHierarchyGraph =
    hierarchyMetaGraphs->getNodeValue (hierarchyNode);
  Graph *fatherGraphInHierarchy = 
    getContainingHierarchyGraph (hierarchyNode);
  node fatherNode = containedByNode.get (hierarchyNode.id);
  Graph *fatherFatherInHierarchy = NULL;
  fatherFatherInHierarchy = getContainingHierarchyGraph (fatherNode);
  
  setBoundingBoxAndCircle (fatherGraphInHierarchy, metaGraphHierarchyGraph,
			   hierarchyNode);
  LayoutProperty *viewLayout =
    fatherGraphInHierarchy->getProperty<LayoutProperty> ("viewLayout");
  DoubleProperty *viewRotation =
    fatherGraphInHierarchy->getProperty<DoubleProperty> ("viewRotation");
  if (fatherGraphInHierarchy != hierarchy) {
    drawSteeredNode (fatherGraphInHierarchy, fatherFatherInHierarchy,
		     useDefaultAlgorithms, pluginProgress);
    applyTorques (fatherGraphInHierarchy, viewRotation, inputGraph);
    centreByArea (fatherGraphInHierarchy, *viewLayout);
    redrawChain (fatherNode);
  }//end if
}//cleanupDisplay

//================================================================
float scaleFactor (double layoutRad, double nodeRad) {
  return (float)nodeRad/(float)layoutRad;
}
//================================================================
void OpenCloseHierarchy::scaleDown (const node &hierarchyNode,
				    vector<GlABoundingCircle *> &augsAdded) {
  Graph *metaGraphHierarchyGraph = 
    hierarchyMetaGraphs->getNodeValue (hierarchyNode);
  Graph *fatherGraphInHierarchy = 
    getContainingHierarchyGraph (hierarchyNode);
  node fatherNode = containedByNode.get (hierarchyNode.id);

  DoubleProperty *metaGraphHierarchyGraphScale =
    metaGraphHierarchyGraph->getProperty<DoubleProperty> ("viewScale");
  metaGraphHierarchyGraphScale->setAllNodeValue (1.0);
 
  LayoutProperty *viewLayout =
    fatherGraphInHierarchy->getProperty<LayoutProperty> ("viewLayout");
  DoubleProperty *viewScale =
    fatherGraphInHierarchy->getProperty<DoubleProperty> ("viewScale");
  SizeProperty *viewSize =
    fatherGraphInHierarchy->getProperty<SizeProperty> ("viewSize");
  DoubleProperty *viewRotation =
    fatherGraphInHierarchy->getProperty<DoubleProperty> ("viewRotation");
  SizeProperty *boundingBoxes =
    fatherGraphInHierarchy->getLocalProperty<SizeProperty> ("boundingBox");
  
  pair<Coord, Coord> metaGraphBoundCircle =
    tlp::computeBoundingRadius (metaGraphHierarchyGraph,
                                metaGraphHierarchyGraph
                                ->getProperty<LayoutProperty> ("viewLayout"),
                                metaGraphHierarchyGraph
                                ->getProperty<SizeProperty> ("viewSize"),
                                metaGraphHierarchyGraph
                                ->getProperty<DoubleProperty> ("viewRotation"));  double radius = (metaGraphBoundCircle.second -
		   metaGraphBoundCircle.first).norm();
  Size nodeSizeEst = viewSize->getNodeValue (hierarchyNode) / 2.0;
  double nodeRad = sqrt (nodeSizeEst.getW()*nodeSizeEst.getW() +
                         nodeSizeEst.getH()*nodeSizeEst.getH());
  Size boxSize (2*nodeRad, 2*nodeRad, 1.0);
  boundingBoxes->setNodeValue (hierarchyNode, boxSize);
  viewScale->setNodeValue (hierarchyNode, 
			   scaleFactor (radius, nodeRad));

  applyTorques (fatherGraphInHierarchy, viewRotation, inputGraph);
  centreByArea (fatherGraphInHierarchy, *viewLayout);
  updateCutLayout (hierarchy, noPos, 0, 1.0);
  updateBoundingCircles (hierarchy, noPos, 0, 1.0, 0, false, augsAdded);
}

//================================================================
void OpenCloseHierarchy::cleanupDisplay (const node &hierarchyNode,
					 vector<GlABoundingCircle *> 
					 &augsAdded) {
  Graph *hierarchySubGraph = 
    hierarchyMetaGraphs->getNodeValue (hierarchyNode);
  LayoutProperty *hierarchySubGraphLayout =
    hierarchySubGraph->getProperty<LayoutProperty> ("viewLayout");
  centreByArea (hierarchySubGraph, *hierarchySubGraphLayout);
  redrawChain (hierarchyNode);
  updateCutLayout (hierarchy, noPos, 0, 1.0);
  updateBoundingCircles (hierarchy, noPos, 0, 1.0, 0, false, augsAdded);
}

//================================================================
void OpenCloseHierarchy::resynchDisplayGraph () {
  vector<GlABoundingCircle *> augsAdded;
  updateCutLayout (hierarchy, noPos, 0, 1.0);
  updateBoundingCircles (hierarchy, noPos, 0, 1.0, 0, false, augsAdded);
}
