#include <tulip/DoubleProperty.h>
#include "HierarchyNodeCreation.h"
#include "SteeredLayout.h"
#include "PullOut.h"

using namespace std;
using namespace tlp;
using namespace stdext;

#define PULL_OUT_LABEL (string ("Adj Nodes"))
const string pullOutErrMsg ("In pull out");

//================================================================
PullOut::PullOut (Graph *hierarchy,
		  Graph *hierarchySubGraphs,
		  Graph *inputGraph,
		  const string &metaGraphPropertyName,
		  OpenCloseHierarchy *openCloseHierarchy,
		  RotatingColours *rotatingColours,
		  DestroyHierarchy *destroyHierarchy,
		  MetaGraphHierarchyWidget *metaGraphHierarchyWidget) :
  hierarchy(hierarchy), hierarchySubGraphs(hierarchySubGraphs),
  inputGraph(inputGraph),
  openCloseHierarchy(openCloseHierarchy),
  rotatingColours(rotatingColours),
  destroyHierarchy(destroyHierarchy), 
  metaGraphHierarchyWidget(metaGraphHierarchyWidget),
  hierarchyBackup(NULL), hierarchySubGraphsBackup(NULL),
  adjPullOutCompNumber (0), otherPullOutNumber (0),
  onRotatingColours(false) {
  hierarchyMetaGraphs = 
    hierarchy->getProperty<GraphProperty> (metaGraphPropertyName);
}

//================================================================
inline string getStringAlgoNumber (unsigned int number) {
  char charAlgoNumber[10];
  sprintf (charAlgoNumber, "%d", number);
  number++;
  return charAlgoNumber;
} 

//================================================================
void PullOut::setColours (vector<node> &addedHierarchyNodes,
			  hash_map<node, vector<node> > 
			  &hierarchyNodeToComponent) {
  //currently if colour pairs aren't on, there is only uniform colour.
  if (!onRotatingColours) return;

  ColorProperty *viewColor = 
    hierarchy->getProperty<ColorProperty> ("viewColor");

  //set all nodes present to unmatch
  hash_map<node, vector<node> >::iterator pullOutComponentsIt =
    hierarchyNodeToComponent.begin ();
  for (; pullOutComponentsIt != hierarchyNodeToComponent.end(); 
       ++pullOutComponentsIt) {
    node curHierarchyNode = pullOutComponentsIt->first;
    Graph *curHierarchySubGraph =
      hierarchyMetaGraphs->getNodeValue (curHierarchyNode);
    if (curHierarchySubGraph != NULL) {
      node curSubNode;
      forEach (curSubNode, curHierarchySubGraph->getNodes ())
	viewColor->setNodeValue (curSubNode,
				 rotatingColours->getCurUnmatchColour ());
    }//end if
  }//end for

  //set the pullout component nodes to matched
  for (unsigned int i = 0; i < addedHierarchyNodes.size(); ++i)
    viewColor->setNodeValue (addedHierarchyNodes[i],
			     rotatingColours->getCurMatchColour ());

  rotatingColours->incrementColour ();
}//end setColours

//================================================================
void PullOut::extractLeafSubGraph (const node &hierarchyNode,
				   BooleanProperty &leaves) {
  Graph *hierarchyNodeMetaGraph = 
    hierarchyMetaGraphs->getNodeValue (hierarchyNode);
  if (hierarchyNodeMetaGraph == NULL)
    leaves.setNodeValue (hierarchyNode, true);
  else {
    node curNode;
    forEach (curNode, hierarchyNodeMetaGraph->getNodes ())
      extractLeafSubGraph (curNode, leaves);
  }//end else
}//end extractLeafSubGraph

//================================================================
bool PullOut::getLeavesNHopsAway (unsigned int nHops,
				  BooleanProperty &start,
				  BooleanProperty &result) {
  const int undirected = 2;
  string errMsg;
  BooleanProperty reachableSubGraph (inputGraph);
  DataSet reachableDataSet;

  result.setAllNodeValue (false);
  result.setAllEdgeValue (false);
  reachableDataSet.set ("distance", nHops);
  reachableDataSet.set ("direction", undirected);
  reachableDataSet.set ("startingnodes", &start);
  bool success = inputGraph->computeProperty ("ReachableSubGraph",
					      &reachableSubGraph, errMsg, NULL,
					      &reachableDataSet);
  if (!success) {
    cerr << "Reachable subgraph returned the following error " << endl 
	 << errMsg << endl << "(Is the plugin installed?)" << endl;
    return false;
  }//end if

  node curSelNode;
  forEach (curSelNode, reachableSubGraph.getNodesEqualTo (true)) {
    if (!start.getNodeValue (curSelNode))
      result.setNodeValue (curSelNode, true);
  }//end forEach

  edge curSelEdge;
  forEach (curSelEdge, reachableSubGraph.getEdgesEqualTo (true)) {
    node source = inputGraph->source (curSelEdge);
    node target = inputGraph->target (curSelEdge);
    if (!start.getNodeValue (source) && !start.getNodeValue (target))
      result.setEdgeValue (curSelEdge, true);
  }//end forEach
  return true;
}//end getLeavesNHopsAway

//================================================================
node PullOut::findCutDisplayNode (const node &hierarchyNode) {
  node displayNode = openCloseHierarchy->getDisplayNode (hierarchyNode);
  Graph *displayGraph = openCloseHierarchy->getDisplayGraph();
  if (displayGraph->isElement (displayNode)) return displayNode;
  
  node superNode = 
    openCloseHierarchy->getContainingHierarchyNode (hierarchyNode);
  return findCutDisplayNode (superNode);
}//end findCutDisplayNode

//================================================================
void PullOut::setUpComponentsTable (hash_map<node, 
				    vector<node> > &hierarchyNodeToComponent) {
  node componentNode;
  forEach (componentNode, pullOutComponents->getNodes()) {
    node displayNode = findCutDisplayNode (componentNode);
    node hierarchyNode = openCloseHierarchy->getHierarchyNode (displayNode);
    hierarchyNodeToComponent[hierarchyNode].push_back (componentNode);
  }//end forEach
}//end setUpComponentsTable

//================================================================
bool PullOut::createPullOutComponents (hash_map<node, vector<node> > 
				       &hierarchyNodeToComponent,
				       MutableContainer<bool> &noTouchLeaves,
				       vector<node> &addedHierarchyNodes) {
  bool success = true;
  noTouchLeaves.setAll (false);
  hash_map<node, vector<node> >::iterator pullOutComponentsIt =
    hierarchyNodeToComponent.begin ();
  for (; pullOutComponentsIt != hierarchyNodeToComponent.end(); 
       ++pullOutComponentsIt) {
    node curHierarchyNode = pullOutComponentsIt->first;
    Graph *curHierarchyGraph =
      hierarchyMetaGraphs->getNodeValue (curHierarchyNode);
    if (curHierarchyGraph != NULL) {
      Graph *curSubGraph = pullOutComponents->addSubGraph ();
      for (unsigned int i = 0; i < pullOutComponentsIt->second.size (); ++i) {
	node curCompNode = pullOutComponentsIt->second[i];
	curSubGraph->addNode (curCompNode); 
	edge curEdge;
	forEach (curEdge, curHierarchyGraph->getInOutEdges (curCompNode)) {
	  node source = curHierarchyGraph->source (curEdge);
	  node target = curHierarchyGraph->target (curEdge);
	  if (curSubGraph->isElement (source) && 
	      curSubGraph->isElement (target))
	    curSubGraph->addEdge (curEdge);
	}//end forEach
      }//end for
      
      success &= processConnectedComponents (curHierarchyGraph, curSubGraph,
					     addedHierarchyNodes,
					     hierarchy, hierarchySubGraphs,
					     hierarchyMetaGraphs,
					     pullOutErrMsg);
     
      for (unsigned int i = 0; i < addedHierarchyNodes.size(); ++i) {
	Graph *curHierMetaGraph = 
	  hierarchyMetaGraphs->getNodeValue (addedHierarchyNodes[i]);
	if (curHierMetaGraph == NULL)
	  noTouchLeaves.set (addedHierarchyNodes[i].id, true);
	else {
	  string strPullOutNumber = getStringAlgoNumber (adjPullOutCompNumber);
	  ++adjPullOutCompNumber;
	  string label = strPullOutNumber + " " + PULL_OUT_LABEL;
	  curHierMetaGraph->setAttribute<string> ("name", label);
	}//end else 
      }//end for
    }//end if
  }//end for
  return success;
}//end createPullOutHierarchyComponents

//================================================================
bool PullOut::rebuildHierarchy (Graph *curHierarchyGraph,
				Graph *backupHierarchyGraph,
				const node &continaingHierarchyNode,
				MutableContainer<vector <node> > &oldToNew,
				MutableContainer<bool> &noTouchLeaves) {
  GraphProperty *backupHierarchyProperty =
    hierarchySubGraphsBackup->getProperty<GraphProperty> ("viewMetaGraph");
  node curBackupNode;
  forEach (curBackupNode, backupHierarchyGraph->getNodes()) {
    Graph *nextLevel = backupHierarchyProperty->getNodeValue (curBackupNode);
    if (nextLevel != NULL) {
      bool success = rebuildHierarchy (curHierarchyGraph, nextLevel,
				       curBackupNode, oldToNew,
				       noTouchLeaves);
      if (!success) return false;
    }//end if
  }//end forEach

  //group all connected sets that were connected sets previously here.
  BooleanProperty component (curHierarchyGraph);
  bool selectedOne = false;
  forEach (curBackupNode, backupHierarchyGraph->getNodes()) {
    if ((backupHierarchyProperty->getNodeValue (curBackupNode) == NULL) &&
	(curHierarchyGraph->isElement (curBackupNode))) {
      if (!noTouchLeaves.get (curBackupNode.id)) {
	component.setNodeValue (curBackupNode, true);
	selectedOne = true;
      }//end if
    }//end if
    else {
      vector<node> newMetaNodes = oldToNew.get (curBackupNode.id);
      for (unsigned int i = 0; i < newMetaNodes.size(); ++i)
	component.setNodeValue (newMetaNodes[i], true);
      selectedOne = true;
    }//end else
  }//end forEach
  if (!selectedOne) return true;

  BooleanProperty newComponentSubGraphs (curHierarchyGraph);
  DataSet leafData;
  leafData.set ("Nodes", &component);
  string errMsg;
  bool success = 
    curHierarchyGraph->computeProperty ("Induced sub-graph",
					&newComponentSubGraphs, errMsg,
					NULL, &leafData);
  if (!success) {
    cerr << "Induced sub-graph returned the following error\n" + errMsg 
      + "\n(Is the plugin installed?)" << endl;
    return false;
  }//end if
  Graph *conComps = curHierarchyGraph->addSubGraph (&newComponentSubGraphs);

  vector<node> addedHierarchyNodes;
  success =
    processConnectedComponents (curHierarchyGraph, conComps, 
				addedHierarchyNodes,
				hierarchy, hierarchySubGraphs,
				hierarchyMetaGraphs,
				pullOutErrMsg);
  if (!success) return false;
  oldToNew.set (continaingHierarchyNode.id, addedHierarchyNodes);
  curHierarchyGraph->delAllSubGraphs (conComps);
  return true;
}//end rebuildHierarchy

//================================================================
void PullOut::
restoreOldHierarchyRespectingConnected (hash_map<node, vector<node> > 
					&hierarchyNodeToComponent,
					MutableContainer<bool> &noTouchLeaves) {
  hash_map<node, vector<node> >::iterator pullOutComponentsIt =
    hierarchyNodeToComponent.begin ();
  for (; pullOutComponentsIt != hierarchyNodeToComponent.end(); 
       ++pullOutComponentsIt) {
    node curHierarchyNode = pullOutComponentsIt->first;
    Graph *curHierarchyGraph =
      hierarchyMetaGraphs->getNodeValue (curHierarchyNode);
    if (curHierarchyGraph != NULL) {
      GraphProperty *backupHierarchyProperty;
      if (hierarchy->isElement (curHierarchyNode))
	backupHierarchyProperty = 
	  hierarchyBackup->getProperty<GraphProperty> ("viewMetaGraph");
      else
	backupHierarchyProperty = 
	  hierarchySubGraphsBackup->getProperty<GraphProperty> ("viewMetaGraph");
      Graph *backupHierarchyGraph =
	backupHierarchyProperty->getNodeValue (curHierarchyNode);
      MutableContainer<vector <node> > oldToNew;
      oldToNew.setAll (vector<node> ());
      rebuildHierarchy (curHierarchyGraph, backupHierarchyGraph,
			curHierarchyNode, oldToNew, noTouchLeaves);
    }//end if
  }//end for
}//end restorOldHierarchyRespectingConnected
//================================================================
void PullOut::backupHierarchy (Graph *curHierarchy, Graph *curBackup) {
  LayoutProperty *bkpLayout = 
    curBackup->getLocalProperty<LayoutProperty> ("viewLayout");
  DoubleProperty *bkpRotation =
    curBackup->getLocalProperty<DoubleProperty> ("viewRotation");
  GraphProperty *bkpViewGraph =
    curBackup->getProperty<GraphProperty> ("viewMetaGraph");
  LayoutProperty *curHierLayout =
    curHierarchy->getProperty<LayoutProperty> ("viewLayout");
  DoubleProperty *curHierRotation =
    curHierarchy->getProperty<DoubleProperty> ("viewRotation");
  GraphProperty *curHierViewGraph =
    curHierarchy->getProperty<GraphProperty> ("viewMetaGraph");

  node curNode;
  forEach (curNode, curHierarchy->getNodes()) {
    curBackup->addNode (curNode);
    bkpLayout->setNodeValue (curNode, 
			     curHierLayout->getNodeValue (curNode));
    bkpRotation->setNodeValue (curNode,
			       curHierRotation->getNodeValue (curNode));
  }//end forEach

  edge curEdge;
  forEach (curEdge, curHierarchy->getEdges()) {
    curBackup->addEdge (curEdge);
    bkpLayout->setEdgeValue (curEdge,
			     curHierLayout->getEdgeValue (curEdge));
  }//end forEach

  forEach (curNode, curHierarchy->getNodes()) {
    Graph *nextHierGraph = curHierViewGraph->getNodeValue (curNode);
    if (nextHierGraph != NULL) {
      Graph *nextHierBkp = hierarchySubGraphsBackup->addSubGraph();
      bkpViewGraph->setNodeValue (curNode, nextHierBkp);
      copyOverLayoutParms (nextHierGraph, nextHierBkp);
      backupHierarchy (nextHierGraph, nextHierBkp);
    }//end if
  }//end forEach
}//end backupHierarchy

//================================================================
void PullOut::createHierarchyBackup () {
  if (hierarchyBackup != NULL) {
    hierarchy->getRoot()->delAllSubGraphs (hierarchyBackup);
    hierarchy->getRoot()->delAllSubGraphs (hierarchySubGraphsBackup);
  }
  hierarchyBackup = hierarchy->getRoot()->addSubGraph();
  hierarchySubGraphsBackup = hierarchy->getRoot()->addSubGraph();
  hierarchyBackup->getLocalProperty<GraphProperty> ("viewMetaGraph");
  hierarchySubGraphsBackup->getLocalProperty<GraphProperty> ("viewMetaGraph");
  backupHierarchy (hierarchy, hierarchyBackup);
  hierarchyBackup->setAttribute<string> ("name", "1__Hierarchy Backup");
  hierarchySubGraphsBackup->setAttribute<string> ("name", "1__Hierarchy Sub Graphs Backup");
}//end createHierarchyBackup

//================================================================
bool PullOut::destroyBelowCutHierarchyNodes (hash_map<node, vector<node> > 
					     &hierarchyNodeToComponent) {
  BooleanProperty hierarchyToDestroy (hierarchy);
  hierarchyToDestroy.setAllNodeValue (false);
  hierarchyToDestroy.setAllEdgeValue (false);
  hash_map<node, vector<node> >::iterator pullOutComponentsIt =
    hierarchyNodeToComponent.begin ();
  for (; pullOutComponentsIt != hierarchyNodeToComponent.end(); 
       ++pullOutComponentsIt) {
    hierarchyToDestroy.setNodeValue (pullOutComponentsIt->first, true);
  }//end for
  return destroyHierarchy->destroyLevelsBelow (hierarchyToDestroy);
}//end destroyBelowCutHierarchyNodes

//================================================================
void PullOut::openDisplays (hash_map<node, vector<node> > 
			     &hierarchyNodeToComponent,
			    vector<node> &pullOutDisplays) {
  hash_map<node, vector<node> >::iterator pullOutComponentsIt =
    hierarchyNodeToComponent.begin ();
  for (; pullOutComponentsIt != hierarchyNodeToComponent.end(); 
       ++pullOutComponentsIt) {
    node openDisplayNode = 
      openCloseHierarchy->getDisplayNode (pullOutComponentsIt->first);
    pullOutDisplays.push_back (openDisplayNode);
  }//end for
}//end synchDisplays

//================================================================
bool PullOut::performPullOut (const node &displayNode,
			      unsigned int numHops,
			      vector<node> &pullOutDisplays) {
  hash_map<node, vector<node> > hierarchyNodeToComponent;
  node hierarchyNode = openCloseHierarchy->getHierarchyNode (displayNode);
  
  //extract leafs below the selected node
  BooleanProperty leafSelection (inputGraph);
  leafSelection.setAllNodeValue (false);
  leafSelection.setAllEdgeValue (false);
  extractLeafSubGraph (hierarchyNode, leafSelection);

  //get subgraph n hops away and set up components table
  BooleanProperty reachableSelection (inputGraph);
  bool success = getLeavesNHopsAway (numHops,
				     leafSelection, reachableSelection);
  if (!success) return false;

  pullOutComponents = 
    hierarchy->getRoot ()->addSubGraph (&reachableSelection);
  pullOutComponents->setAttribute<string> ("name", "pullOutComponents");
  setUpComponentsTable (hierarchyNodeToComponent);

  //backup entire hierarchy and destroy hierarchy below affected nodes
  MutableContainer<bool> noTouchLeaves;
  vector<node> addedHierarchyNodes;
  createHierarchyBackup ();
  destroyBelowCutHierarchyNodes (hierarchyNodeToComponent);
  createPullOutComponents (hierarchyNodeToComponent, noTouchLeaves,
			   addedHierarchyNodes);
  restoreOldHierarchyRespectingConnected (hierarchyNodeToComponent, 
					  noTouchLeaves);
  setColours (addedHierarchyNodes, hierarchyNodeToComponent);
  openDisplays (hierarchyNodeToComponent, pullOutDisplays);
  hierarchy->getRoot ()->delAllSubGraphs (pullOutComponents);
  return true;
}//end performPullOut
