/**
 * The following class provides functions to destroy hierarchical structure
 * below a comunity and make new structure.
 */
#ifndef __PULL_OUT_H__
#define __PULL_OUT_H__
#if (__GNUC__ < 3)
#  include <hash_map>
#else
#  include <ext/hash_map>
#endif
#include <string>
#include <vector>
#include <tulip/BooleanProperty.h>
#include <tulip/GraphProperty.h>
#include <tulip/MutableContainer.h>
#include <tulip/MetaGraphHierarchyWidget.h>
#include "DestroyHierarchy.h"
#include "OpenCloseHierarchy.h"
#include "RotatingColours.h"

class PullOut {
 public:
  /**
   * Constructor for a pull out object.
   */
  PullOut (tlp::Graph *hierarchy,
	   tlp::Graph *hierarchySubGraphs,
	   tlp::Graph *inputGraph,
	   const std::string &metaGraphPropertyName,
	   OpenCloseHierarchy *openCloseHierarchy,
	   RotatingColours *rotatingColours,
	   DestroyHierarchy *destroyHierarchy,
	   MetaGraphHierarchyWidget *metaGraphHierarchyWidget);

  /**
   * Destructor 
   */
  inline ~PullOut () {}

 public:

  /**
   * Preforms pull out from the passed display node the specified number
   * of hops.  The hashmap converts each display node with the set of leaves
   * beneath it.
   * \param displayNode The display node for which pull out is executed on.
   * \param numHops The number of hops for the pull out.
   * \param displayToComponent list of display nodes that the pull out
   * has been executed on.
   */
  bool performPullOut (const tlp::node &displayNode,
		       unsigned int numHops,
		       std::vector<tlp::node> &pullOutDisplays);

   /**
    * An inline function to set whether or not rotating colours is on.
    * \param turnOn If true, rotating colours is on.  Otherwise, it is off.
    */
   inline void turnOnRotatingColours (bool turnOn) {
     onRotatingColours = turnOn;
   }//end turnOnRotatingColours
   
 private:
  tlp::Graph *hierarchy;
  tlp::Graph *hierarchySubGraphs;
  tlp::Graph *inputGraph;
  tlp::Graph *pullOutComponents;
  tlp::Graph *hierarchyBackup;
  tlp::Graph *hierarchySubGraphsBackup;
  tlp::GraphProperty *hierarchyMetaGraphs;
  OpenCloseHierarchy *openCloseHierarchy;
  RotatingColours *rotatingColours;
  DestroyHierarchy *destroyHierarchy;
  MetaGraphHierarchyWidget *metaGraphHierarchyWidget;
  unsigned int adjPullOutCompNumber, otherPullOutNumber;
  bool onRotatingColours;

   /**
    * Extracts all leaf nodes of the hierarchy below the passed hierarchy
    * node
    * \param hierarchyNode Metanode below which leaves are extracted.
    * \param leaves A selection of the leaves below hierarchyNode
    */
   void extractLeafSubGraph (const tlp::node &hierarchyNode,
			     tlp::BooleanProperty &leaves);

   /**
    * Extracts the set of leaf nodes n hops away from a selected leaf set.
    * Returns true if the function executed correctly.
    * \param nHops The number of hops away to select
    * \param start The starting selected set
    * \param result The selected set of nodes n hops away 
    */
   bool getLeavesNHopsAway (unsigned int nHops,
			    tlp::BooleanProperty &start,
			    tlp::BooleanProperty &result);

   /**
    * Finds the cut display node visible which contains this hierarchy node
    * and returns it.
    * \param hierarchyNode The hierarchy node which is contained by a
    * cut display node.
    */
   tlp::node findCutDisplayNode (const tlp::node &hierarchyNode);

   /**
    * Sets up the component table.  The component table provides a hash 
    * between the display
    * nodes visible on the cut and leaf nodes in the graph
    */
   void setUpComponentsTable (stdext::hash_map<tlp::node, 
			      std::vector<tlp::node> > &hierarchyNodeToComponent);

   /**
    * Creates the pull out components themselves
    * \param hierarchyNodeToComponent maps the new hierarchy node to 
    * pull out component
    * \param noTouchLeaves Leaves which are hierarchy components and should
    * not be touched when restructuring hierarchy
    * \param addedHierarchyNodes A vector of hierarchy nodes.
    */
   bool createPullOutComponents (stdext::hash_map<tlp::node, std::vector<tlp::node> > 
				 &hierarchyNodeToComponent,
				 tlp::MutableContainer<bool> &noTouchLeaves,
				 vector<tlp::node> &addedHierarchyNodes);

   /**
    * Restores the old hierarchy after pullout from the backup respecting
    * connected components
    */
   bool rebuildHierarchy (tlp::Graph *curHierarchyGraph,
			  tlp::Graph *backupHierarchyGraph,
			  const tlp::node &continaingHierarchyNode,
			  tlp::MutableContainer<std::vector <tlp::node> > 
			  &oldToNew,
			  tlp::MutableContainer<bool> &noTouchLeaves);

   /**
    * Restores the old hierarchy after pullout from the backup respecting
    * connected components
    * \param hierarchyNodeToComponent Maps hierarchy nodes to pullOutCompontent
    * \param noTouchLeaves Leaves in hierarchy which do not belong in metanodes
    */
   void restoreOldHierarchyRespectingConnected (stdext::hash_map<tlp::node, std::vector<tlp::node> > &hierarchyNodeToComponent, tlp::MutableContainer<bool> &noTouchLeaves);

   
   /**
    * Recursive helper to create hierarchy backup.
    * \param curHierarchy The current hierarchy
    * \param curBackup The current backup hierarchy.
    */
   void backupHierarchy (tlp::Graph *curHierarchy, tlp::Graph *curBackup);

   /**
    * Creates a backup of the current graph hierarchy.
    */
   void createHierarchyBackup ();

    /**
     * Destroys hierarchies below the nodes involved in the pull out.
     * \param hierarchyNodeToComponent The hierarchy nodes below which
     * the hierarchy is destroyed.
     */
   bool destroyBelowCutHierarchyNodes (stdext::hash_map<tlp::node, std::vector<tlp::node> > 
				       &hierarchyNodeToComponent);

   /**
    * Method which sets the colours of pull out and other components.
    * \param addedHierarchyNodes the new pull out metanodes.
    * \param hierarchyNodeToComponent A table of hierarchy nodes which
    * contain pull out components
    */
   void setColours (std::vector<tlp::node> &addedHierarchyNodes,
		    stdext::hash_map<tlp::node, std::vector<tlp::node> > 
		    &hierarchyNodeToComponent);

   /**
    * Records the display nodes which need to be opened as a result
    * of this pull out operation.
    * \param hierarchyNodeToComponent The hierarchy nodes affected by pull out
    * \param pullOutDisplays The display nodes which should be opened.
    */
   void openDisplays (stdext::hash_map<tlp::node, std::vector<tlp::node> > 
		       &hierarchyNodeToComponent,
		       std::vector<tlp::node> &pullOutDisplays);
};
#endif
