#include "HierarchyNodeCreation.h"
#include "Split.h"

using namespace std;
using namespace tlp;
using namespace stdext;

//==================================================================
Split::Split (Graph *displayGraph, Graph *hierarchy, 
	      Graph *hierarchySubGraphs, 
	      OpenCloseHierarchy *openCloseHierarchy,
	      const string &metaGraphPropertyName) :
  displayGraph (displayGraph),
  hierarchy (hierarchy),
  hierarchySubGraphs (hierarchySubGraphs),
  openCloseHierarchy (openCloseHierarchy) {
  hierarchyMetaGraphs = 
    hierarchy->getProperty<GraphProperty> (metaGraphPropertyName);
}
//==================================================================
node Split::findHierarchyNodeAtGraph (const node &hierarchyNode,
				      Graph *stopGraph) {
  node curNode = hierarchyNode;
  while (curNode.isValid () && !stopGraph->isElement (curNode)) {
    node nextNode = openCloseHierarchy->getContainingHierarchyNode (curNode);
    curNode = nextNode;
  }//end while
  return curNode;
}//end findHierarchyNodeAtGraph

//==================================================================
void Split::addSplitNodeToGraph (const node &splitNode,
				 const node &containingHierarchyNode,
				 Graph *containingHierarchyGraph) {
  containingHierarchyGraph->addNode (splitNode);
  containingHierarchyGraph->addEdge (containingHierarchyNode,
				     splitNode);
  node splitDisplayNode =
    openCloseHierarchy->getDisplayNode (splitNode);
  node adjacentCutNode;
  forEach (adjacentCutNode, displayGraph->getInOutNodes (splitDisplayNode)) {
    node adjacentHierarchyNode =
      openCloseHierarchy->getHierarchyNode (adjacentCutNode);
    node atHierarchy = findHierarchyNodeAtGraph (adjacentHierarchyNode,
						 containingHierarchyGraph);
    if (atHierarchy == containingHierarchyNode) continue;
    if (atHierarchy.isValid () &&
	!containingHierarchyGraph->existEdge (atHierarchy, splitNode).isValid()
	&&
	!containingHierarchyGraph->existEdge (splitNode, atHierarchy).isValid())
      containingHierarchyGraph->addEdge (atHierarchy, splitNode);
  }//end forEach
}//end addSplitNodeToGraph

//==================================================================
void Split::fixHierarchyEdges (const node &splitNode,
			       Graph *containingSplitNodeGraph,
			       const node containingHierarchyNode,
			       Graph *containingHierarchyGraph) {
  MutableContainer<bool> edgeWitness;
  edgeWitness.setAll (false);
  node curHierarchyNode;
  forEach (curHierarchyNode, containingSplitNodeGraph->getNodes ()) {
    if (curHierarchyNode == splitNode) continue;
    node curDisplayNode = 
      openCloseHierarchy->getDisplayNode (curHierarchyNode);
    vector<node> toCheck;
    if (curDisplayNode.isValid ()) 
      toCheck.push_back (curDisplayNode);
    else {
      node subNode;
      Graph *newNodeGraph = hierarchyMetaGraphs->getNodeValue (curHierarchyNode);
      forEach (subNode, newNodeGraph->getNodes ())
	toCheck.push_back (subNode);
    }//end else
    for (unsigned int i = 0; i < toCheck.size(); ++i) {
      node oppositeDisplay;
      forEach (oppositeDisplay, displayGraph->getInOutNodes (toCheck[i])) {
	node oppositeHierarchy =
	  openCloseHierarchy->getHierarchyNode (oppositeDisplay);
	node atHierarchy = findHierarchyNodeAtGraph (oppositeHierarchy,
						     containingHierarchyGraph);
	if (atHierarchy.isValid ()) {
	  edge metaEdge1 = containingHierarchyGraph->existEdge (atHierarchy,
								containingHierarchyNode);
	  edge metaEdge2 = containingHierarchyGraph->existEdge (containingHierarchyNode,
								atHierarchy);
	  if (metaEdge1.isValid ()) edgeWitness.set (metaEdge1.id, true);
	  if (metaEdge2.isValid ()) edgeWitness.set (metaEdge2.id, true);
	}//end if
      }//end forEach
    }//end for
  }//end forEach
  StableIterator<edge> 
    delEdges (containingHierarchyGraph->getInOutEdges (containingHierarchyNode));
  while (delEdges.hasNext ()) {
    edge curEdge = delEdges.next ();
    if (!edgeWitness.get (curEdge.id)) 
      containingHierarchyGraph->delEdge (curEdge);
  }//end while
}//end fixHierarchyEdges

//==================================================================
bool Split::removeSplitNodeFromGraph (const node &splitNode,
				      Graph *splitGraph) {
  splitGraph->delNode (splitNode);

  //regroup connected components
  Graph *groupConnected = splitGraph->getRoot ()->addSubGraph ();
  copy (splitGraph, groupConnected);
  DoubleProperty connectedComponents (groupConnected);
  string errMsg;
  bool success = groupConnected->computeProperty ("Connected Component", 
					      &connectedComponents, errMsg);
  if (!success) {
    cerr << "Connected components returned the following error " << endl;
    cerr << errMsg << endl << " is the plugin installed." << endl;
    return false;
  }//end if
  
  DataSet localData;
  localData.set("Property", &connectedComponents);
  success = tlp::applyAlgorithm (groupConnected, errMsg, &localData, 
				 "Equal Value");
  if (!success) {
    cerr << "Equal Value clustering returned the following error:\n"
	 << errMsg << "\n(Is the plugin properly installed?)" << endl;
    return false;
  }//end if

  unsigned int numComponents = 0;
  Graph *countComp;
  Iterator<Graph *> *graphIterator = groupConnected->getSubGraphs ();
  while (graphIterator->hasNext()) {
    ++numComponents;
    countComp = graphIterator->next();
  }//end while

  if (numComponents > 1) {
    StableIterator<Graph *> concomps (groupConnected->getSubGraphs());
    while (concomps.hasNext()) {
      Graph *curConComp = concomps.next();
      if (curConComp->numberOfNodes() > 1) {
	Graph *newHierarchyGraph = hierarchySubGraphs->addSubGraph();
	newHierarchyGraph->setAttribute<string> ("name", "Split ");
	copy (curConComp, newHierarchyGraph);
	node newHierarchyNode = splitGraph->addNode();
	hierarchyMetaGraphs->setNodeValue (newHierarchyNode, 
					   newHierarchyGraph);
	setNewConnections (splitGraph, newHierarchyGraph, newHierarchyNode);
	delFeature (newHierarchyGraph, splitGraph);
	openCloseHierarchy->synchHierarchyBelow (newHierarchyNode);
      }//end if
    }//end while
  }//end if
  splitGraph->getRoot ()->delAllSubGraphs (groupConnected);
  return true;
}//end removeSplitNodeFromGraph

//==================================================================
void Split::cullEmpties (Graph *containingSplitNodeGraph) {
  StableIterator<node> delEmpty (containingSplitNodeGraph->getNodes ());
  while (delEmpty.hasNext ()) {
    node nextEmptyCheck = delEmpty.next ();
    Graph *nextEmptySubGraph = 
      hierarchyMetaGraphs->getNodeValue (nextEmptyCheck);
    if (nextEmptySubGraph != NULL) {
      if (nextEmptySubGraph->numberOfNodes () == 0)
	containingSplitNodeGraph->delNode (nextEmptyCheck);
    }//end if
  }//end while
}//end cullEmpties

//==================================================================
void Split::splitOneLevel (const node &splitNode,
			   Graph *containingSplitNodeGraph,
			   const node containingHierarchyNode,
			   Graph *containingHierarchyGraph) {
  addSplitNodeToGraph (splitNode, containingHierarchyNode,
		       containingHierarchyGraph);
  fixHierarchyEdges (splitNode, containingSplitNodeGraph,
		     containingHierarchyNode, containingHierarchyGraph);
  removeSplitNodeFromGraph (splitNode, containingSplitNodeGraph);
  cullEmpties (containingSplitNodeGraph);
}//end split

//==================================================================
bool Split::splitToGraph (const node &hierarchyNode, Graph *stopGraph) {
  Graph *curContainingGraph = 
    openCloseHierarchy->getContainingHierarchyGraph (hierarchyNode);
  node containingHierarchyNode =
    openCloseHierarchy->getContainingHierarchyNode (hierarchyNode);
  Graph *containingHierarchyNodeGraph =
    openCloseHierarchy->getContainingHierarchyGraph (containingHierarchyNode);

  while ((curContainingGraph != stopGraph) && 
	 containingHierarchyNode.isValid ()) {
    splitOneLevel (hierarchyNode, curContainingGraph,
		   containingHierarchyNode,
		   containingHierarchyNodeGraph);
    openCloseHierarchy->synchHierarchyBelow (containingHierarchyNode);
    curContainingGraph = containingHierarchyNodeGraph;
    node nextNode =
      openCloseHierarchy->getContainingHierarchyNode (containingHierarchyNode);
    containingHierarchyNode = nextNode;
    if (nextNode.isValid ())
      containingHierarchyNodeGraph =
	openCloseHierarchy->getContainingHierarchyGraph (nextNode);
  }//end while

  if (containingHierarchyNode.isValid ())
    openCloseHierarchy->synchHierarchyBelow (containingHierarchyNode);
  
  return containingHierarchyNode.isValid ();
}//splitToGraph
