/**
 * The following class provides functions to destroy hierarchical structure
 * below a comunity and make new structure.
 */
#ifndef __SPLIT_H__
#define __SPLIT_H__
#include <string>
#include <vector>
#include <tulip/BooleanProperty.h>
#include <tulip/GraphProperty.h>
#include <tulip/MutableContainer.h>
#include <tulip/MetaGraphHierarchyWidget.h>
#include "DestroyHierarchy.h"
#include "OpenCloseHierarchy.h"

class Split {
 public:
  /**
   * Constructor for a pull out object.
   */
  Split (tlp::Graph *displayGraph,
	 tlp::Graph *hierarchy,
	 tlp::Graph *hierarchySubGraphs,
	 OpenCloseHierarchy *openCloseHierarchy,
	 const std::string &metaGraphPropertyName);
  
  /**
   * Destructor 
   */
  inline ~Split () {}

 public:

  /**
   * Recursively splits the hierarchy node out of its
   * containing graph until it is placed in stop graph.
   * \param hierarchyNode The node being split out.
   * \param stopGraph The graph where splitting stops.
   */
  bool splitToGraph (const node &hierarchyNode, Graph *stopGraph);
   
 private:
  tlp::Graph *displayGraph;
  tlp::Graph *hierarchy;
  tlp::Graph *hierarchySubGraphs;
  OpenCloseHierarchy *openCloseHierarchy;
  tlp::GraphProperty *hierarchyMetaGraphs;

  /**
   * Finds the hierarchy node containing hierarchyNode at 
   * stopGraph.
   * \param hierarchyNode The node we want to trace up from
   * \param stopGraph The graph where we stop tracing up
   */
  tlp::node findHierarchyNodeAtGraph (const tlp::node &hierarchyNode,
				      tlp::Graph *stopGraph);
  /**
   * Removes the splitNode from splitGraph and re assembles any
   * disconnection in components.
   * \param splitNode Node to be removed from splitGraph
   * \param graph which contains splitNode.
   */
  bool removeSplitNodeFromGraph (const tlp::node &splitNode,
				 tlp::Graph *splitGraph);

  /**
   * Function to add a the split node to the containing hierarchy
   * graph.
   * \param splitNode The split node being taken out of containingHierarchyNode
   * \param containingHierarchyNode node which contains the split node
   * \param containingHierarchyGraph graph containing containingHierarchyNode
   */
  void addSplitNodeToGraph (const tlp::node &splitNode,
			    const tlp::node &containingHierarchyNode,
			    tlp::Graph *containingHierarchyGraph);

  /**
   * Function to fix the hierarchy edges after a split at one level.
   * \param splitNode Node to be removed from splitGraph
   * \param graph which contains splitNode.
   * \param containingHierarchyNode node which contains the split node
   * \param containingHierarchyGraph graph containing containingHierarchyNode
   */
  void fixHierarchyEdges (const tlp::node &splitNode,
			  tlp::Graph *containingSplitNodeGraph,
			  const tlp::node containingHierarchyNode,
			  tlp::Graph *containingHierarchyNodeGraph);
  
  /**
   * Splits the split node out one level into containingHierarchyGraph.
   * \param splitNode Node to be removed from splitGraph
   * \param graph which contains splitNode.
   * \param containingHierarchyNode node which contains the split node
   * \param containingHierarchyGraph graph containing containingHierarchyNode
   */
  void splitOneLevel (const tlp::node &splitNode,
		      tlp::Graph *containingSplitNodeGraph,
		      const tlp::node containingHierarchyNode,
		      tlp::Graph *containingHierarchyNodeGraph);

  /**
   * Determines if any metanodes are empty after the split.  If so, it deletes them
   */
  void cullEmpties (Graph *containingSplitNodeGraph);
};
#endif
