#include <string>
#include <qinputdialog.h>
#include <tulip/TreeTest.h>
#include <tulip/PlanarityTest.h>
#include <tulip/ForEach.h>
#include <tulip/DrawingTools.h>
#include <tulip/LayoutProperty.h>
#include <tulip/SizeProperty.h>
#include <tulip/DoubleProperty.h>
#include <tulip/StableIterator.h>
#include "HierarchyMap.h"
#include "SteeredLayout.h"
using namespace tlp;
using namespace std;

#define GEM ("GEM")
#define LGL ("LGL Area")
#define MIXED_MODEL ("Mixed Model")
#define RT_EXTENDED ("Hierarchical Tree (R-T Extended)")
#define CONNECTED_COMPONENT ("Connected Componnent Packing")

#define CONNECTED_COMPONENT_ALGORITHMS ("Connected Componnent Packing;Circular")
#define TREE_LAYOUT_ALGORITHMS ("Bubble Tree;Hierarchical Tree (R-T Extended);GEM")
#define EMBEDDER_ALGORITHMS ("High Dimensional Embedder;GEM;")
#define NEAR_COMPLETE_ALGORITHMS ("Circular;GEM;")
#define PLANAR_ALGORITHMS ("GEM;Mixed Model;")
#define UNKNOWN_ALGORITHMS ("GEM;")
#define LARGE_UNKNOWN_ALGORITHMS ("GEM;GEM SPF;")
#define REQUIRES_OVERLAP_REMOVAL ("GEM;High Dimensional Embedder;LGL;LGL Area;GEM SPF;")
#define FAST_OVERLAP_REMOVAL ("Fast Overlap Removal VPSC")
#define TOO_LARGE_THRESH (1000)

const string noAlgorithmSelected ("<no selection>");
const string previousSelectedAlg ("previousSelectedAlgorithm");
const string previousDataSet ("previousDataSet");

//================================================================
//Private helper function to create a new edge order table
void setEdgeOrderTable (Graph *curSubGraph) {
  EdgeOrderTable *newEdgeOrderTable = new EdgeOrderTable();
  node curNode;
  forEach (curNode, curSubGraph->getNodes()) {
    vector<edge> curEdgeList;
    edge curEdge;
    forEach (curEdge, curSubGraph->getInOutEdges(curNode))
      curEdgeList.push_back (curEdge);
    newEdgeOrderTable->set (curNode.id, curEdgeList);
  }//end forEach
  curSubGraph->setAttribute<EdgeOrderTable *> (EDGE_ORDER_ATTRIBUTE,
					       newEdgeOrderTable);
}//end setEdgeOrderTable

//================================================================
//private helper function to set the order of the edges around each
//node consistantly.
void setEdgeOrder (Graph *curSubGraph) {
  EdgeOrderTable *curEdgeOrderTable =
    curSubGraph->getAttribute<EdgeOrderTable *>(EDGE_ORDER_ATTRIBUTE);
  StableIterator<node> curSubGraphNodes (curSubGraph->getNodes());
  while (curSubGraphNodes.hasNext ()) {
    node curNode = curSubGraphNodes.next ();
    const vector<edge> &edgeOrder = curEdgeOrderTable->get (curNode.id);
    curSubGraph->setEdgeOrder (curNode, edgeOrder);
  }//end forEach
}//end setEdgeOrder

//================================================================
//Copies a edge order table given one as input
void copyEdgeOrder (EdgeOrderTable *inputEdgeOrder,
		    EdgeOrderTable *outputEdgeOrder,
		    Graph *inputGraph) {
  EdgeOrderTable *curEdgeOrderTable =
    inputGraph->getAttribute<EdgeOrderTable *>(EDGE_ORDER_ATTRIBUTE);
  node curNode;
  forEach (curNode, inputGraph->getNodes()) {
    const vector<edge> &edgeOrder = curEdgeOrderTable->get (curNode.id);
    outputEdgeOrder->set (curNode.id, edgeOrder);
  }//end foreach
}//end copyEdgeOrder

//================================================================
void copyOverLayoutParms (tlp::Graph *src, tlp::Graph *dest) {
  dest->setAttribute<string> ("name", src->getAttribute<string> ("name"));
  dest->setAttribute<bool> (HAS_LAYOUT_ATTRIBUTE,
			    src->getAttribute<bool> (HAS_LAYOUT_ATTRIBUTE));
  dest->setAttribute<node> (HDE_ROOT_ATTRIBUTE,
			    src->getAttribute<tlp::node> (HDE_ROOT_ATTRIBUTE));
  EdgeOrderTable *curTable =
    src->getAttribute<EdgeOrderTable *> (EDGE_ORDER_ATTRIBUTE);
  if (curTable != NULL) {
    EdgeOrderTable *destTable = new EdgeOrderTable ();
    copyEdgeOrder (curTable, destTable, src);
    dest->setAttribute<EdgeOrderTable *> (EDGE_ORDER_ATTRIBUTE, destTable);
  }//end if
  else dest->setAttribute<EdgeOrderTable *> (EDGE_ORDER_ATTRIBUTE, NULL);
}

//================================================================
inline double computeRad (const Size &curSize) {
  return sqrt (curSize.getW()*curSize.getW()/4.0 + 
	       curSize.getH()*curSize.getH()/4.0);
}//end computeRad

//================================================================
//set up the GEM dataset edge length attributes
void gemEdgeLength(Graph *curSubGraph, DoubleProperty &edgeWeight) {
  SizeProperty *viewSize =
    curSubGraph->getProperty<SizeProperty> ("viewSize");
  edge curEdge;
  forEach (curEdge, curSubGraph->getEdges()) {
    node source = curSubGraph->source (curEdge);
    node target = curSubGraph->target (curEdge);
    Size sourceSize = viewSize->getNodeValue (source);
    Size targetSize = viewSize->getNodeValue (target);
    double weight = computeRad (sourceSize) + computeRad (targetSize);
    edgeWeight.setEdgeValue (curEdge, weight);
  }//end forEach
}

//================================================================
//set up the initial layout
void initializeLayoutAndSize (Graph *curSubGraph, LayoutProperty &initLayout,
			      SizeProperty &initSizes) {
  LayoutProperty *viewLayout =
    curSubGraph->getProperty<LayoutProperty> ("viewLayout");
  SizeProperty *viewSize = 
    curSubGraph->getProperty<SizeProperty> ("viewSize");
  node curNode;
  forEach (curNode, curSubGraph->getNodes()) {
    initSizes.setNodeValue (curNode,
			    viewSize->getNodeValue (curNode));
    initLayout.setNodeValue (curNode, 
			     viewLayout->getNodeValue (curNode));
  }//end forEach
}//end initializeLayoutAndSize

//================================================================
void rootBiconnected (Graph *curSubGraph) {
  node curNode, maxDegNode;
  unsigned int maxDeg = 0;
  forEach (curNode, curSubGraph->getNodes()) {
    if (curSubGraph->deg (curNode) > maxDeg) {
      maxDeg = curSubGraph->deg (curNode);
      maxDegNode = curNode;
    }//end if
  }//end forEach
  TreeTest::makeRootedTree (curSubGraph, maxDegNode);
}

//================================================================
void rootTree (Graph *curSubGraph, Graph *fatherSubGraph) {
  
  if (fatherSubGraph == NULL) {
    rootBiconnected (curSubGraph);
    return;
  }//end if no father
  node curNode, rootNode;
  Graph *root = curSubGraph->getRoot();
  bool done = false;
  Iterator<node> *subgraphNodes = curSubGraph->getNodes();
  while (subgraphNodes->hasNext ()) {
    node curNode = subgraphNodes->next ();
    Iterator<node> *oppositeNodes = root->getInOutNodes (curNode);
    while (oppositeNodes->hasNext ()) {
      node oppositeNode = oppositeNodes->next ();
      if (fatherSubGraph->isElement (oppositeNode)) {
	rootNode = curNode;
	done = true;
	break;
      }//end if
    } delete oppositeNodes; //end while
    if (done) break;
  } delete subgraphNodes; //end while
  if (!rootNode.isValid()) rootBiconnected (curSubGraph);
  else TreeTest::makeRootedTree (curSubGraph, rootNode);
}

//========================================================================
#define CLOSE_ENOUGH_VAR (0.05)
#define MAX_PERCENT_VAR (0.60)
#define MIN_PERCENT_VAR (0.10)
#define VAR_THRESH (1000)

//========================================================================
inline bool enoughVar (double lambda, unsigned int numNodes) {
  return lambda/static_cast<double> (numNodes) > VAR_THRESH; 
}//end enoughVar

//========================================================================
inline void normalizeLambdas (vector<double> &lambdas) {
  double sumLambda = 0;
  vector<double>::iterator lambdasIt = lambdas.begin();
  for (; lambdasIt != lambdas.end(); ++lambdasIt)
    sumLambda += *lambdasIt;
  lambdasIt = lambdas.begin();
  for (; lambdasIt != lambdas.end(); ++lambdasIt)
    (*lambdasIt) = (*lambdasIt)/sumLambda;
}//end normalizeLambdas

//========================================================================
inline bool goodVariance (vector<double> &lambdas,
			  vector<vector<double> > &evecs) {
  //if close to unit basis, probably not a good layout
  if (!enoughVar (lambdas[0], evecs[0].size())) return false;
  normalizeLambdas (lambdas);
 
  //if the percent variance in the two axes is about the same, 
  //layout is good
  if (lambdas[0] - lambdas[1] < CLOSE_ENOUGH_VAR) return true; 
  
  //could be good in three dimensions
  double sumFirstThree = lambdas[0] + lambdas[1] + lambdas[2];
  return (sumFirstThree > (1 - CLOSE_ENOUGH_VAR)) && 
    (lambdas[2] > MIN_PERCENT_VAR) && 
    (lambdas[2] < MIN_PERCENT_VAR + CLOSE_ENOUGH_VAR) && 
    (lambdas[0] < MAX_PERCENT_VAR);
}//end goodVariance

//========================================================================
bool embedderTest (Graph *workingGraph) {
  string errMsg; //for error messages
  
  LayoutProperty localEmbedder (workingGraph);
  DataSet localData;
  localData.set<bool> ("Return Eigen", true);
  for (short i = 0; i < 3; ++i) {
    bool success = 
      workingGraph->computeProperty ("High Dimensional Embedder",
				     &localEmbedder, errMsg, NULL, 
				     &localData);
    if (!success) {
      cerr << "Error encountered while trying to call High Dimensional "
	   << "Embedder " << endl << errMsg << endl;
      return false;
    }//end if
    
    vector<double> lambdas;
    vector<vector<double> > evecs;
    if (!localData.get ("lambdas", lambdas)) {
      cerr << "Error:  eigenvalues not returned." << endl;
      return false;
    }//end if
    if (!localData.get ("evecs", evecs)) {
      cerr << "Error:  eigenvectors not returned." << endl;
      return false;
    }//end if
    
    //if we have good variance, use HDE
    if (goodVariance (lambdas, evecs)) return true;
  }//end for
  return false;
}//end embedder

//================================================================
bool nearCompleteTest (Graph *curSubGraph) {
  return highNodeEdgeRatio (curSubGraph->numberOfNodes(),
			    curSubGraph->numberOfEdges(), 1.0);
}//end nearCompleteTest

//================================================================
string queryAlgorithm (const string layoutAlgoSet, int suggestion) {
  QStringList algChoices;
  StringCollection stringChoices (layoutAlgoSet);
  for (unsigned int i = 0; i < stringChoices.size(); ++i)
    algChoices << stringChoices[i];
  bool ok = false;
  QString algorithm = QInputDialog::getItem ("Grouse",
					     "Select an algorithm:",
					     algChoices, suggestion, 
					     FALSE, &ok);
  if (!ok) return noAlgorithmSelected;
  return algorithm;
}//end querryAlgorithm

//================================================================
void setRTDataset (DataSet &layoutAlgoDataSet) {
  layoutAlgoDataSet.set ("layer spacing", 1.0);
  layoutAlgoDataSet.set ("node spacing", 1.0);
  layoutAlgoDataSet.set ("bounding circles", true);
  layoutAlgoDataSet.set ("discard size adjustment", true);
  StringCollection tmp ("vertical;");
  layoutAlgoDataSet.set ("orientation",tmp);
}

//================================================================
void setMixedModelDataset (Graph *curSubGraph, DataSet &layoutAlgoDataSet) {
  layoutAlgoDataSet.set ("x node-node and edge-node spacing", 1.0f);
  layoutAlgoDataSet.set ("y node-node spacing", 1.0f);
  layoutAlgoDataSet.set ("bounding circles", true);
  StringCollection tmp ("vertical;");
  layoutAlgoDataSet.set ("orientation",tmp);
}

//================================================================
void setGEMDataset (Graph *curSubGraph, DataSet &layoutAlgoDataSet) {
  LayoutProperty *initLayout =
    curSubGraph->getLocalProperty<LayoutProperty> ("view init layout");
  SizeProperty *initSize =
    curSubGraph->getLocalProperty<SizeProperty> ("view init size");
  initializeLayoutAndSize (curSubGraph, *initLayout, *initSize);
  layoutAlgoDataSet.set ("Node Size", initSize);
}

//=====================================================================
void resetGEMDataset (Graph *curSubGraph, DataSet &layoutAlgoDataSet) {
  SizeProperty *initSize = 
    curSubGraph->getProperty<SizeProperty> ("view init size");
  LayoutProperty *initLayout =
    curSubGraph->getProperty<LayoutProperty> ("view init layout");
  initializeLayoutAndSize (curSubGraph, *initLayout, *initSize);
  if (!layoutAlgoDataSet.exist ("initial layout"))
    layoutAlgoDataSet.set ("initial layout", initLayout);
}//end resetGEMDataset


//================================================================
//set up the initial layout
void initPackLayoutAndSize (Graph *curSubGraph, LayoutProperty &initLayout,
			    SizeProperty &initSizes) {
  LayoutProperty *viewLayout =
    curSubGraph->getProperty<LayoutProperty> ("viewLayout");
  SizeProperty *boundingBox = 
    curSubGraph->getProperty<SizeProperty> ("boundingBox");
  node curNode;
  forEach (curNode, curSubGraph->getNodes()) {
    initSizes.setNodeValue (curNode,
			    boundingBox->getNodeValue (curNode));
    initLayout.setNodeValue (curNode, 
			     viewLayout->getNodeValue (curNode));
  }//end forEach
}//end initializeLayoutAndSize

//=====================================================================
void setPackDataset (Graph *curSubGraph, DataSet &layoutAlgoDataSet) {
  LayoutProperty *initLayout =
    curSubGraph->getLocalProperty<LayoutProperty> ("view init layout");
  SizeProperty *initSize =
    curSubGraph->getLocalProperty<SizeProperty> ("view init size");
  initPackLayoutAndSize (curSubGraph, *initLayout, *initSize);
  layoutAlgoDataSet.set ("complexity", StringCollection ("auto;"));
  layoutAlgoDataSet.set ("size", initSize);
}//end setPackDataset

//=====================================================================
void resetPackDataset (Graph *curSubGraph, DataSet &layoutAlgoDataSet) {
  SizeProperty *initSize = 
    curSubGraph->getProperty<SizeProperty> ("view init size");
  LayoutProperty *initLayout =
    curSubGraph->getProperty<LayoutProperty> ("view init layout");
  initPackLayoutAndSize (curSubGraph, *initLayout, *initSize);
  if (!layoutAlgoDataSet.exist ("coordinates"))
    layoutAlgoDataSet.set ("coordinates", initLayout);
}//end resetPackDataset

//=====================================================================
//determines if a tree is bushy (degree > depth)
bool bushy (node n, node cameFrom, Graph *featureGraph,
	    unsigned int curDepth, 
	    unsigned int &maxDepth, unsigned int &maxDegree,
	    MutableContainer<bool> &visited) {
  if (visited.get (n.id)) return maxDepth < maxDegree;
  visited.set (n.id, true);

  //see if we have a new max depth or max degree
  if (featureGraph->deg (n) > maxDegree) maxDegree = featureGraph->deg (n);
  if (curDepth > maxDepth) maxDepth = curDepth;
  
  //compute the max degree and always compare it to the depth
  Iterator <node> *it = featureGraph->getInOutNodes(n);
  while(it->hasNext()) {
    node curNode = it->next();
    if (curNode != cameFrom) 
      bushy (curNode, n, featureGraph, curDepth + 1, maxDepth, maxDegree,
	     visited);
  } delete it; //end while
  return maxDepth < maxDegree;
}//end bushy

//=====================================================================
//A test to determine if a tree is bushy or not
bool isBushy (Graph *tree) {
  node root;
  tlp::getSource (tree, root);
  unsigned int maxDegree = 0, maxDepth = 0;
  MutableContainer<bool> visited;
  visited.setAll (false);
  return bushy (root, root, tree, 0, maxDepth, maxDegree, visited);
}//end isBushy

//================================================================
void setDataSet (Graph *curSubGraph, 
		 string &algorithm, DataSet &layoutAlgoDataSet) {
  if (algorithm == RT_EXTENDED) 
    setRTDataset (layoutAlgoDataSet);
  else if (algorithm == MIXED_MODEL)
    setMixedModelDataset (curSubGraph, layoutAlgoDataSet);
  else if (algorithm == GEM)
    setGEMDataset (curSubGraph, layoutAlgoDataSet);
  else if (algorithm == CONNECTED_COMPONENT)
    setPackDataset (curSubGraph, layoutAlgoDataSet);
}//end setDataSet

//================================================================
string selectAlgorithm (Graph *curSubGraph, Graph *fatherSubGraph,
			bool useDefaultAlgorithms) {
  string layoutAlgoSet = UNKNOWN_ALGORITHMS;
  int suggestion = 0;

  //sets of disconnected nodes are packed
  if ((curSubGraph->numberOfEdges () == 0) &&
      (curSubGraph->numberOfNodes () > 1)) {
    layoutAlgoSet = CONNECTED_COMPONENT_ALGORITHMS;
  }//end if

  else if (TreeTest::isFreeTree (curSubGraph)) {
    rootTree (curSubGraph, fatherSubGraph);
    layoutAlgoSet = TREE_LAYOUT_ALGORITHMS; 
    suggestion = isBushy (curSubGraph) ? 0 : 1; //bubble first rt second
  }//end if

  else if (embedderTest (curSubGraph)) {
    const DataSet &attributes = curSubGraph->getAttributes();  
    if (!attributes.exist (HDE_ROOT_ATTRIBUTE))
      curSubGraph->setAttribute<node> (HDE_ROOT_ATTRIBUTE, 
				       curSubGraph->getOneNode());
    layoutAlgoSet = EMBEDDER_ALGORITHMS;
  }//end else if
  
  else if (nearCompleteTest (curSubGraph)) {
    layoutAlgoSet = NEAR_COMPLETE_ALGORITHMS;
  }//end else if
  
  else if (curSubGraph->numberOfNodes () < TOO_LARGE_THRESH) {
    if (PlanarityTest::isPlanar (curSubGraph))
      layoutAlgoSet = PLANAR_ALGORITHMS;
    //otherwise we are using default as set at top
  }//end else if

  else if (curSubGraph->numberOfNodes () >= TOO_LARGE_THRESH)
    layoutAlgoSet = LARGE_UNKNOWN_ALGORITHMS;
  
  if (useDefaultAlgorithms) 
    return StringCollection (layoutAlgoSet) [suggestion];
  else 
    return queryAlgorithm (layoutAlgoSet, suggestion);
}//end selectAlgorithm

//================================================================
bool requiresOverlapRemoval (const string &algorithm) {
  StringCollection overlap (REQUIRES_OVERLAP_REMOVAL);
  for (unsigned int i = 0; i < overlap.size(); ++i)
    if (overlap[i] == algorithm) return true;
  return false;
}//end requires overlap removal

//================================================================
bool drawSteeredNode (Graph *curSubGraph, Graph *fatherSubGraph,
		      bool useDefaultAlgorithms, 
		      PluginProgress *pluginProgress,
		      string useThisAlgorithm) {
  LayoutProperty *curSubGraphLayout =
    curSubGraph->getProperty<LayoutProperty> ("viewLayout");

  bool hasLayout = curSubGraph->getAttribute<bool> (HAS_LAYOUT_ATTRIBUTE);
  string algorithm; 
  DataSet layoutAlgoDataSet;
  if (!hasLayout) {
    if (useThisAlgorithm == "") {
      algorithm = selectAlgorithm (curSubGraph, fatherSubGraph,
				   useDefaultAlgorithms);
      if (algorithm == noAlgorithmSelected) return false;
    }//end if
    else algorithm = useThisAlgorithm;
    setDataSet (curSubGraph, algorithm, layoutAlgoDataSet);

    setEdgeOrderTable (curSubGraph);
    curSubGraph->setAttribute<string> (previousSelectedAlg, algorithm);
    curSubGraph->setAttribute<DataSet> (previousDataSet, layoutAlgoDataSet);
  }//end if
  else {
    algorithm = curSubGraph->getAttribute<string> (previousSelectedAlg);
    layoutAlgoDataSet = curSubGraph->getAttribute<DataSet> (previousDataSet);
    if (algorithm == GEM) resetGEMDataset (curSubGraph, layoutAlgoDataSet);
    if (algorithm == CONNECTED_COMPONENT)
      resetPackDataset (curSubGraph, layoutAlgoDataSet);
  }//end else
  setEdgeOrder (curSubGraph);

  string errMsg; 
  LayoutProperty hierLayout (curSubGraph);
  pluginProgress->reset_continue();
  bool success = 
    curSubGraph->computeProperty (algorithm, &hierLayout, errMsg,
				  pluginProgress, &layoutAlgoDataSet);
  if (!success) {
    cerr << "Error in calling: " << algorithm << endl << errMsg << endl;
    return false;
  }//end if

  if (pluginProgress->state() == TLP_CANCEL)
    return false;

  node curNode;
  forEach (curNode, curSubGraph->getNodes()) 
    curSubGraphLayout->setNodeValue (curNode, 
				     hierLayout.getNodeValue (curNode));
  
  edge curEdge;
  forEach (curEdge, curSubGraph->getEdges())
    curSubGraphLayout->setEdgeValue (curEdge,
				     hierLayout.getEdgeValue (curEdge));

  if (pluginProgress->state() == TLP_STOP)
    return true;

  if (requiresOverlapRemoval (algorithm)) {
    LayoutProperty overlapsLayout (curSubGraph);
    SizeProperty *boundingBox = 
      curSubGraph->getProperty<SizeProperty> ("boundingBox");
    layoutAlgoDataSet.set ("boundingBox", boundingBox);
    curSubGraph->computeProperty (FAST_OVERLAP_REMOVAL,
				  &overlapsLayout, errMsg,
				  NULL, &layoutAlgoDataSet);
    forEach (curNode, curSubGraph->getNodes()) 
      curSubGraphLayout->setNodeValue (curNode, 
				       overlapsLayout.getNodeValue (curNode));
  }
  
  forEach (curEdge, curSubGraph->getEdges())
    curSubGraphLayout->setEdgeValue (curEdge,
				     hierLayout.getEdgeValue (curEdge));

  curSubGraph->setAttribute<bool> (HAS_LAYOUT_ATTRIBUTE, true);
  return success;
}

//================================================================
void clearLayoutParameters (Graph *delGraph) {
  delGraph->setAttribute<bool> (HAS_LAYOUT_ATTRIBUTE, false);
  node dummy;
  EdgeOrderTable *curTable;
  DataSet &attributes = delGraph->getAttributes();
  if (attributes.exist (HDE_ROOT_ATTRIBUTE))
    attributes.getAndFree (HDE_ROOT_ATTRIBUTE, dummy);

  if (attributes.exist (EDGE_ORDER_ATTRIBUTE)) {
    bool hasTable = attributes.get (EDGE_ORDER_ATTRIBUTE, curTable);
    if (hasTable && (curTable != NULL)) delete curTable;
  }//end if
  delGraph->setAttribute<EdgeOrderTable *> (EDGE_ORDER_ATTRIBUTE, NULL);
}//end clearLayoutParameters
