
#include <tulip/FindSelection.h>
#include <tulip/ForEach.h>
#include "AttributeMetric.h"

using namespace std;
using namespace tlp;

//==================================================================
AttributeMetric::AttributeMetric (const PropertyContext &context):
  tlp::DoubleAlgorithm(context)  {}

AttributeMetric::~AttributeMetric(){}

//==================================================================
Graph *AttributeMetric::trueComponents (BooleanProperty *selection) {
  string errMsg;
  BooleanProperty inducedSubGraph (graph);
  DataSet localData;
  localData.set ("Nodes", selection);
  bool success = graph->computeProperty ("Induced sub-graph",
					 &inducedSubGraph, errMsg, NULL,
					 &localData);
  if (!success) {
    cerr << "Induced sub-graph returned the following error " << endl 
	 << errMsg << endl << "(Is the plugin installed?)" << endl;
    return NULL;
  }//end if
  return graph->getRoot()->addSubGraph(&inducedSubGraph);
}//end trueComponents

//==================================================================
Graph *AttributeMetric::falseComponents (BooleanProperty *selection) {
  BooleanProperty selectionComplement (graph);
  selectionComplement.setAllEdgeValue (false);
  node curNode;
  forEach (curNode, graph->getNodes())
    selectionComplement.setNodeValue (curNode, 
				      !selection->getNodeValue (curNode));
  DataSet localData;
  string errMsg;
  localData.set ("Nodes", &selectionComplement);
  BooleanProperty inducedComplement (graph);
  bool success = graph->computeProperty ("Induced sub-graph",
					 &inducedComplement, errMsg, NULL,
					 &localData);
  if (!success) {
    cerr << "Induced sub-graph returned the following error " << endl 
	 << errMsg << endl << "(Is the plugin installed?)" << endl;
    return NULL;
  }//end if
  return graph->getRoot()->addSubGraph(&inducedComplement);
}//end falseComponents

//==================================================================
bool AttributeMetric::makeConnected (Graph *disconnectedGraph,
				     DoubleProperty &connectedMetric) {
  string errMsg;
  bool success = disconnectedGraph->computeProperty ("Connected Component",
						     &connectedMetric,
						     errMsg);
  if (!success) {
    cerr << "Connected components returned the following error " << endl 
	 << errMsg << endl << "(Is the plugin installed?)" << endl;
    return false;
  }//end if
  return true;
}//end makeConnected

//==================================================================
void AttributeMetric::setupFinalMetric (Graph *inducedComponents,
					DoubleProperty &trueComponents,
					Graph *otherComponents,
					DoubleProperty &falseComponents) {
  double maxVal = 0;
  DoubleAlgorithm::doubleResult->setAllNodeValue (0);
  node curNode;
  forEach (curNode, inducedComponents->getNodes()) {
    double curVal = 1 + trueComponents.getNodeValue (curNode);
    DoubleAlgorithm::doubleResult->setNodeValue (curNode, curVal);
    if (maxVal < curVal) maxVal = curVal;
  }//end forEach
  
  forEach (curNode, otherComponents->getNodes()) {
    double curVal = 1 + maxVal + falseComponents.getNodeValue (curNode);
    DoubleAlgorithm::doubleResult->setNodeValue (curNode, curVal);
  }//end forEach
}//end setupFinalMetric

//==================================================================
bool AttributeMetric::run () {
  bool hasMatches = false;
  bool hasSelection = false;
  BooleanProperty *matches = NULL;
  BooleanProperty *selection = NULL;
  if (dataSet != NULL) {
    hasMatches = dataSet->get ("matches", matches);
    hasSelection = dataSet->get ("selection", selection);
  }

  if (!hasSelection) {
    SelectionWidget *sel = new SelectionWidget(graph);
    Observable::holdObservers();
    sel->exec();
    Observable::unholdObservers();
    delete sel;
    selection =
      graph->getProperty<BooleanProperty> ("viewSelection");
  }//end if

  if (hasMatches) {
    node curNode;
    forEach (curNode, graph->getNodes())
      matches->setNodeValue (curNode, selection->getNodeValue (curNode));
  }//end if
  
  Graph *inducedComponents = trueComponents (selection);
  Graph *otherComponents = falseComponents (selection);
  if ((inducedComponents == NULL) || (otherComponents == NULL))
    return false;

  DoubleProperty trueComponents (inducedComponents);
  DoubleProperty falseComponents (otherComponents);
  bool successInduced = makeConnected (inducedComponents, trueComponents);
  bool successConnected = makeConnected (otherComponents, falseComponents);
  if (!successInduced || !successConnected)
    return false;
  setupFinalMetric (inducedComponents, trueComponents, 
		    otherComponents, falseComponents);

  //cleanup created graphs/selections
  if (!hasSelection) {
    selection->setAllNodeValue (false);
    selection->setAllEdgeValue (false);
  }//end if
  graph->getRoot()->delAllSubGraphs (inducedComponents);
  graph->getRoot()->delAllSubGraphs (otherComponents);
  return true;
}//end run

DOUBLEPLUGIN(AttributeMetric,"Find Attribute Clustering","Daniel Archambault","02/12/2007","Alpha","0","1");
